@@ -24,6 +24,10 @@
 
 #include <netdb.h>
 #include <arpa/inet.h>
+#include <sys/time.h>
+#include <sys/types.h>
+#include <poll.h>
+#include <unistd.h>
 #include "coordinatorapi.h"
 #include "dmtcp.h"
 #include "util.h"
@@ -34,6 +38,12 @@
 #include  "../jalib/jconvert.h"
 #include  "../jalib/jfilesystem.h"
 #include <fcntl.h>
+#include <semaphore.h> // for sem_post(&sem_launch)
+
+// sem_launch is used in threadlist.cpp
+// sem_launch_first_time will be set just before pthread_create(checkpointhread)
+LIB_PRIVATE bool sem_launch_first_time = false;
+LIB_PRIVATE sem_t sem_launch;
 
 using namespace dmtcp;
 
@@ -49,12 +59,16 @@ void dmtcp_CoordinatorAPI_EventHook(DmtcpEvent_t event, DmtcpEventData_t *data)
       JASSERT(CoordinatorAPI::instance().isValid());
       break;
 
+    case DMTCP_EVENT_RESTART:
+      CoordinatorAPI::restart();
+      break;
+
     case DMTCP_EVENT_RESUME:
-      dmtcp::CoordinatorAPI::instance().sendCkptFilename();
+      CoordinatorAPI::instance().sendCkptFilename();
       break;
 
     case DMTCP_EVENT_EXIT:
-      JTRACE("exit() in progress, disconnecting from dmtcp coordinator");
+      JLOG(DMTCP)("exit() in progress, disconnecting from dmtcp coordinator");
       CoordinatorAPI::instance().closeConnection();
       break;
 
@@ -63,27 +77,11 @@ void dmtcp_CoordinatorAPI_EventHook(DmtcpEvent_t event, DmtcpEventData_t *data)
   }
 }
 
-static void getHostAndPort(CoordinatorAPI::CoordinatorMode mode,
-                           dmtcp::string *hostname,
-                           int *port)
+void CoordinatorAPI::restart()
 {
-  const char *addr = getenv (ENV_VAR_NAME_HOST);
-  if (addr == NULL) addr = DEFAULT_HOST;
-  *hostname = addr;
-
-  const char *portStr = getenv (ENV_VAR_NAME_PORT);
-
-  JASSERT(mode & CoordinatorAPI::COORD_NEW || mode & CoordinatorAPI::COORD_ANY);
-  if (portStr != NULL) {
-    *port = jalib::StringToInt(portStr);
-  } else if (mode & CoordinatorAPI::COORD_NEW) {
-    *port = 0;
-  } else {
-    *port = DEFAULT_PORT;
-  }
+  instance()._nsSock.close();
 }
 
-
 static uint32_t getCkptInterval()
 {
   uint32_t ret = DMTCPMESSAGE_SAME_CKPT_INTERVAL;
@@ -99,55 +97,21 @@ static uint32_t getCkptInterval()
   return ret;
 }
 
-static dmtcp::string getPrefixDir()
+static jalib::JSocket createNewSocketToCoordinator(CoordinatorMode mode)
 {
-  dmtcp::string prefixDir = "";
-  const char *prefixPathEnv = getenv(ENV_VAR_PREFIX_PATH);
-  if (prefixPathEnv != NULL) {
-    /* If --prefix was defined then this process is either running on the local
-     * node (the home of first process in the comptation) or a remote node.
-     *
-     * If the process is running on the local node, the prefix-path-env may be
-     * different from the prefix-dir of this binary, in which case, we want to
-     * send the prefix-path of this binary to the coordinator and the
-     * coordinator will save it as the local-prefix.
-     *
-     * However, if this is running on a remote node, the prefix-path-env would
-     * be the same as the prefix-path of this binary and we should send the
-     * prefix-path-env to the coordinator and the coordinator will note this as
-     * the remote-prefix.
-     */
-    const char *utilDirPath = getenv(ENV_VAR_UTILITY_DIR);
-    dmtcp::string utilDirPrefix = "";
-    if (utilDirPath != NULL) {
-      utilDirPrefix = jalib::Filesystem::DirName(utilDirPath);
-    }
-    if (utilDirPrefix == jalib::Filesystem::ResolveSymlink(prefixPathEnv)) {
-      prefixDir = prefixPathEnv;
-    } else {
-      prefixDir = utilDirPrefix;
-    }
-  }
-  return prefixDir;
+  string host = "";
+  int port = UNINITIALIZED_PORT;
+
+  Util::getCoordHostAndPort(COORD_ANY, host, &port);
+  return jalib::JClientSocket(host.c_str(), port);
 }
 
-static jalib::JSocket
-createNewSocketToCoordinator( CoordinatorAPI::CoordinatorMode mode)
-{
-  dmtcp::string addr;
-  int port;
-
-  getHostAndPort(mode, &addr, &port);
-  jalib::JSocket sock = jalib::JClientSocket(addr.c_str(), port);
-  return sock;
-}
-
-//dmtcp::CoordinatorAPI::CoordinatorAPI (int sockfd)
+//CoordinatorAPI::CoordinatorAPI (int sockfd)
   //: _coordinatorSocket(sockfd)
 //{ }
 
-static dmtcp::CoordinatorAPI *coordAPIInst = NULL;
-dmtcp::CoordinatorAPI& dmtcp::CoordinatorAPI::instance()
+static CoordinatorAPI *coordAPIInst = NULL;
+CoordinatorAPI& CoordinatorAPI::instance()
 {
   //static SysVIPC *inst = new SysVIPC(); return *inst;
   if (coordAPIInst == NULL) {
@@ -159,11 +123,11 @@ dmtcp::CoordinatorAPI& dmtcp::CoordinatorAPI::instance()
   return *coordAPIInst;
 }
 
-void dmtcp::CoordinatorAPI::init()
+void CoordinatorAPI::init()
 {
-  JTRACE("Informing coordinator of new process") (UniquePid::ThisProcess());
+  JLOG(DMTCP)("Informing coordinator of new process") (UniquePid::ThisProcess());
 
-  DmtcpMessage msg (DMT_UPDATE_PROCESS_INFO_AFTER_EXEC);
+  DmtcpMessage msg (DMT_UPDATE_PROCESS_INFO_AFTER_INIT_OR_EXEC);
   string progname = jalib::Filesystem::GetProgramName();
   msg.extraBytes = progname.length() + 1;
 
@@ -176,14 +140,14 @@ void dmtcp::CoordinatorAPI::init()
   // to call recvCoordinatorHandshake().
 }
 
-void dmtcp::CoordinatorAPI::resetOnFork(dmtcp::CoordinatorAPI& coordAPI)
+void CoordinatorAPI::resetOnFork(CoordinatorAPI& coordAPI)
 {
   JASSERT(coordAPI._coordinatorSocket.isValid());
   JASSERT(coordAPI._coordinatorSocket.sockfd() != PROTECTED_COORD_FD);
   instance() = coordAPI;
   instance()._coordinatorSocket.changeFd(PROTECTED_COORD_FD);
 
-  JTRACE("Informing coordinator of new process") (UniquePid::ThisProcess());
+  JLOG(DMTCP)("Informing coordinator of new process") (UniquePid::ThisProcess());
 
   DmtcpMessage msg (DMT_UPDATE_PROCESS_INFO_AFTER_FORK);
   if (dmtcp_virtual_to_real_pid) {
@@ -197,19 +161,32 @@ void dmtcp::CoordinatorAPI::resetOnFork(dmtcp::CoordinatorAPI& coordAPI)
   instance()._nsSock.close();
 }
 
-void dmtcp::CoordinatorAPI::setupVirtualCoordinator(CoordinatorInfo *coordInfo,
-                                                    struct in_addr  *localIP)
+/* Recompute the protected coordinator fd and reset the coordiantor socket.
+ *
+ * Used to handle the case where protectedFdBase() has changed, for example,
+ * on restart.
+ * */
+void CoordinatorAPI::resetCoordSocketFd()
 {
-  const char *portStr = getenv(ENV_VAR_NAME_PORT);
-  int port = (portStr == NULL) ? DEFAULT_PORT : jalib::StringToInt(portStr);
+  _coordinatorSocket = jalib::JSocket(PROTECTED_COORD_FD);
+}
+
+// FIXME:  Does "virtual coordinator" mean coordinator built into the
+//         the current process (no separate process?)
+void CoordinatorAPI::setupVirtualCoordinator(CoordinatorInfo *coordInfo,
+                                             struct in_addr  *localIP)
+{
+  string host = "";
+  int port = UNINITIALIZED_PORT;
+  Util::getCoordHostAndPort(COORD_NONE, host, &port);
   _coordinatorSocket = jalib::JServerSocket(jalib::JSockAddr::ANY, port);
   JASSERT(_coordinatorSocket.isValid()) (port) (JASSERT_ERRNO)
     .Text("Failed to create listen socket.");
   _coordinatorSocket.changeFd(PROTECTED_COORD_FD);
-  string coordPort= jalib::XToString(_coordinatorSocket.port());
-  setenv (ENV_VAR_NAME_PORT, coordPort.c_str(), 1);
+  Util::setCoordPort(_coordinatorSocket.port());
 
-  dmtcp::Util::setVirtualPidEnvVar(INITIAL_VIRTUAL_PID, getppid());
+  pid_t ppid = getppid();
+  Util::setVirtualPidEnvVar(INITIAL_VIRTUAL_PID, ppid, ppid);
 
   UniquePid coordId = UniquePid(INITIAL_VIRTUAL_PID,
                                 UniquePid::ThisProcess().hostid(),
@@ -227,13 +204,12 @@ void dmtcp::CoordinatorAPI::setupVirtualCoordinator(CoordinatorInfo *coordInfo,
   memset(localIP, 0, sizeof(*localIP));
 }
 
-void dmtcp::CoordinatorAPI::waitForCheckpointCommand()
+void CoordinatorAPI::waitForCheckpointCommand()
 {
   uint32_t ckptInterval = SharedData::getCkptInterval();
   struct timeval tmptime={0,0};
   long remaining = ckptInterval;
   do {
-    fd_set rfds;
     struct timeval *timeout = NULL;
     struct timeval start;
     if (ckptInterval > 0) {
@@ -241,18 +217,34 @@ void dmtcp::CoordinatorAPI::waitForCheckpointCommand()
       timeout->tv_sec = remaining;
       JASSERT(gettimeofday(&start, NULL) == 0) (JASSERT_ERRNO);
     }
-    FD_ZERO(&rfds);
-    FD_SET(PROTECTED_COORD_FD, &rfds );
-    int retval = select(PROTECTED_COORD_FD+1, &rfds, NULL, NULL, timeout);
+
+    // This call to poll() does nothing and returns.
+    // But we want to find address of poll() using dlsym/libc before
+    //   allowing the user thread to continue.
+    poll(NULL, 0, 0);
+    if (sem_launch_first_time) {
+      // Release user thread now that we've initialized the checkpoint thread.
+      // This code is reached if the --no-coordinator flag is used.
+      sem_post(&sem_launch);
+      sem_launch_first_time = false;
+    }
+
+    struct pollfd socketFd = {0};
+    socketFd.fd = PROTECTED_COORD_FD;
+    socketFd.events = POLLIN;
+    uint64_t millis = timeout ? ((timeout->tv_sec * (uint64_t)1000) +
+                                 (timeout->tv_usec / 1000))
+                              : -1;
+    int retval = poll(&socketFd, 1, millis);
     if (retval == 0) { // timeout expired, time for checkpoint
-      JTRACE("Timeout expired, checkpointing now.");
+      JLOG(DMTCP)("Timeout expired, checkpointing now.");
       return;
     } else if (retval > 0) {
-      JASSERT(FD_ISSET(PROTECTED_COORD_FD, &rfds));
-      JTRACE("Connect request on virtual coordinator socket.");
+      JASSERT(socketFd.revents & POLLIN);
+      JLOG(DMTCP)("Connect request on virtual coordinator socket.");
       break;
     }
-    JASSERT(errno == EINTR) (JASSERT_ERRNO);
+    JASSERT(errno == EINTR) (JASSERT_ERRNO); /* EINTR: a signal was caught */
     if (ckptInterval > 0) {
       struct timeval end;
       JASSERT(gettimeofday(&end, NULL) == 0) (JASSERT_ERRNO);
@@ -265,13 +257,13 @@ void dmtcp::CoordinatorAPI::waitForCheckpointCommand()
   } while (remaining > 0);
 
   jalib::JSocket cmdSock(-1);
-  dmtcp::DmtcpMessage msg;
-  dmtcp::DmtcpMessage reply(DMT_USER_CMD_RESULT);
+  DmtcpMessage msg;
+  DmtcpMessage reply(DMT_USER_CMD_RESULT);
   do {
     cmdSock.close();
     cmdSock = _coordinatorSocket.accept();
     msg.poison();
-    JTRACE("Reading from incoming connection...");
+    JLOG(DMTCP)("Reading from incoming connection...");
     cmdSock >> msg;
   } while (!cmdSock.isValid());
 
@@ -283,35 +275,35 @@ void dmtcp::CoordinatorAPI::waitForCheckpointCommand()
   bool exitWhenDone = false;
   switch (msg.coordCmd) {
 //    case 'b': case 'B':  // prefix blocking command, prior to checkpoint command
-//      JTRACE("blocking checkpoint beginning...");
+//      JLOG(DMTCP)("blocking checkpoint beginning...");
 //      blockUntilDone = true;
 //      break;
     case 's': case 'S':
-      JTRACE("Received status command");
+      JLOG(DMTCP)("Received status command");
       reply.numPeers = 1;
       reply.isRunning = 1;
       break;
     case 'c': case 'C':
-      JTRACE("checkpointing...");
+      JLOG(DMTCP)("checkpointing...");
       break;
     case 'k': case 'K':
     case 'q': case 'Q':
-      JTRACE("Received KILL command from user, exiting");
+      JLOG(DMTCP)("Received KILL command from user, exiting");
       exitWhenDone = true;
       break;
     default:
-      JTRACE("unhandled user command") (msg.coordCmd);
+      JLOG(DMTCP)("unhandled user command") (msg.coordCmd);
       reply.coordCmdStatus = CoordCmdStatus::ERROR_INVALID_COMMAND;
   }
   cmdSock << reply;
   cmdSock.close();
   if (exitWhenDone) {
-    _exit(0);
+    _real_exit(0);
   }
   return;
 }
 
-bool dmtcp::CoordinatorAPI::noCoordinator()
+bool CoordinatorAPI::noCoordinator()
 {
   static int virtualCoordinator = -1;
   if (virtualCoordinator == -1) {
@@ -328,15 +320,18 @@ bool dmtcp::CoordinatorAPI::noCoordinator()
   return virtualCoordinator;
 }
 
-void dmtcp::CoordinatorAPI::connectAndSendUserCommand(char c,
-                                                      int *coordCmdStatus,
-                                                      int *numPeers,
-                                                      int *isRunning)
+char* CoordinatorAPI::connectAndSendUserCommand(char c,
+                                                int *coordCmdStatus,
+                                                int *numPeers,
+                                                int *isRunning,
+                                                int *ckptInterval,
+                                                uint32_t logMask)
 {
+  char *replyData = NULL;
   _coordinatorSocket = createNewSocketToCoordinator(COORD_ANY);
   if (!_coordinatorSocket.isValid()) {
     *coordCmdStatus = CoordCmdStatus::ERROR_COORDINATOR_NOT_FOUND;
-    return;
+    return replyData;
   }
 
   //tell the coordinator to run given user command
@@ -345,6 +340,7 @@ void dmtcp::CoordinatorAPI::connectAndSendUserCommand(char c,
   //send
   msg.type = DMT_USER_CMD;
   msg.coordCmd = c;
+  msg.logMask = logMask;
 
   if (c == 'i') {
     const char* interval = getenv (ENV_VAR_CKPT_INTR);
@@ -358,12 +354,12 @@ void dmtcp::CoordinatorAPI::connectAndSendUserCommand(char c,
   //the coordinator will violently close our socket...
   if (c=='q' || c=='Q') {
     *coordCmdStatus = CoordCmdStatus::NOERROR;
-    return;
+    return replyData;
   }
 
   //receive REPLY
   reply.poison();
-  _coordinatorSocket >> reply;
+  recvMsgFromCoordinator(&reply, (void**)&replyData);
   reply.assertValid();
   JASSERT(reply.type == DMT_USER_CMD_RESULT);
 
@@ -376,29 +372,25 @@ void dmtcp::CoordinatorAPI::connectAndSendUserCommand(char c,
   if (isRunning != NULL) {
     *isRunning = reply.isRunning;
   }
+  if (ckptInterval != NULL) {
+    *ckptInterval = reply.theCheckpointInterval;
+  }
 
   _coordinatorSocket.close();
+
+  return replyData;
 }
 
-dmtcp::string dmtcp::CoordinatorAPI::getCoordCkptDir(void)
+string CoordinatorAPI::getCoordCkptDir(void)
 {
   // FIXME: Add a test for make-check.
-  char buf[PATH_MAX];
+  int coordCmdStatus = CoordCmdStatus::NOERROR;
   if (noCoordinator()) return "";
-  DmtcpMessage msg(DMT_GET_CKPT_DIR);
-  _coordinatorSocket << msg;
-
-  msg.poison();
-  _coordinatorSocket >> msg;
-  msg.assertValid();
-  JASSERT(msg.type == DMT_GET_CKPT_DIR_RESULT) (msg.type);
-
-  JASSERT(msg.extraBytes > 0);
-  _coordinatorSocket.readAll(buf, msg.extraBytes);
-  return buf;
+  string coordCkptDir = connectAndSendUserCommand('e', &coordCmdStatus);
+  return coordCkptDir;
 }
 
-void dmtcp::CoordinatorAPI::updateCoordCkptDir(const char *dir)
+void CoordinatorAPI::updateCoordCkptDir(const char *dir)
 {
   if (noCoordinator()) return;
   JASSERT(dir != NULL);
@@ -408,9 +400,36 @@ void dmtcp::CoordinatorAPI::updateCoordCkptDir(const char *dir)
   _coordinatorSocket.writeAll(dir, strlen(dir) + 1);
 }
 
-void dmtcp::CoordinatorAPI::sendMsgToCoordinator(const dmtcp::DmtcpMessage &msg,
-                                                 const void *extraData,
-                                                 size_t len)
+// Invoked by the application or a plugin via
+// dmtcp.h:dmtcp_set_global_ckpt_dir() and
+// via dmtcpplugin.cpp:dmtcp_set_global_ckpt_dir()
+bool CoordinatorAPI::updateGlobalCkptDir(const char *dir)
+{
+  if (noCoordinator()) return false;
+  JASSERT(dir != NULL);
+  jalib::JSocket sock = createNewSocketToCoordinator(COORD_ANY);
+  JASSERT(sock.isValid());
+  DmtcpMessage msg(DMT_UPDATE_GLOBAL_CKPT_DIR);
+  msg.extraBytes = strlen(dir) + 1;
+  sock << msg;
+  sock.writeAll(dir, msg.extraBytes);
+
+  msg.poison();
+  sock >> msg;
+  sock.close();
+
+  if (msg.type == DMT_UPDATE_GLOBAL_CKPT_DIR_SUCCEED) {
+    JTRACE("Updated global checkpoint dir") (dir);
+    return true;
+  } else {
+    JTRACE("Failed to update global checkpoint dir") (dir);
+    return false;
+  }
+}
+
+void CoordinatorAPI::sendMsgToCoordinator(const DmtcpMessage &msg,
+                                          const void *extraData,
+                                          size_t len)
 {
   if (noCoordinator()) return;
   _coordinatorSocket << msg;
@@ -421,16 +440,22 @@ void dmtcp::CoordinatorAPI::sendMsgToCoordinator(const dmtcp::DmtcpMessage &msg,
   }
 }
 
-void dmtcp::CoordinatorAPI::recvMsgFromCoordinator(dmtcp::DmtcpMessage *msg,
-                                                   void **extraData)
+void CoordinatorAPI::recvMsgFromCoordinator(DmtcpMessage *msg, void **extraData)
 {
-  JASSERT(!noCoordinator());
+  JASSERT(!noCoordinator()).Text("internal error");
+  if (sem_launch_first_time) {
+    // Release user thread now that we've initialized the checkpoint thread.
+    // This code is reached if the --no-coordinator flag is not used.
+    // FIXME:  Technically, some rare type of software could still execute
+    //   between here and when we readall() from coord, thus creating a race.
+    sem_post(&sem_launch);
+    sem_launch_first_time = false;
+  }
   msg->poison();
   _coordinatorSocket >> (*msg);
 
-  if (extraData != NULL) {
+  if (extraData != NULL && msg->extraBytes > 0) {
     msg->assertValid();
-    JASSERT(msg->extraBytes > 0);
     // Caller must free this buffer
     void *buf = JALLOC_HELPER_MALLOC(msg->extraBytes);
     _coordinatorSocket.readAll((char*)buf, msg->extraBytes);
@@ -439,17 +464,16 @@ void dmtcp::CoordinatorAPI::recvMsgFromCoordinator(dmtcp::DmtcpMessage *msg,
   }
 }
 
-void dmtcp::CoordinatorAPI::startNewCoordinator(CoordinatorAPI::CoordinatorMode
-                                                mode)
+void CoordinatorAPI::startNewCoordinator(CoordinatorMode mode)
 {
-  dmtcp::string addr;
-  int port;
-  getHostAndPort(mode, &addr, &port);
+  string host = "";
+  int port = UNINITIALIZED_PORT;
+  Util::getCoordHostAndPort(mode, host, &port);
 
-  JASSERT(addr == "localhost" ||
-          addr == "127.0.0.1" ||
-          addr == jalib::Filesystem::GetCurrentHostname())
-    (addr) (jalib::Filesystem::GetCurrentHostname())
+  JASSERT(strcmp(host.c_str(), "localhost") == 0 ||
+          strcmp(host.c_str(), "127.0.0.1") == 0 ||
+          jalib::Filesystem::GetCurrentHostname() == host)
+    (host) (jalib::Filesystem::GetCurrentHostname())
     .Text("Won't automatically start coordinator because DMTCP_HOST"
           " is set to a remote host.");
   // Create a socket and bind it to an unused port.
@@ -457,23 +481,40 @@ void dmtcp::CoordinatorAPI::startNewCoordinator(CoordinatorAPI::CoordinatorMode
   jalib::JServerSocket coordinatorListenerSocket(jalib::JSockAddr::ANY,
                                                  port, 128);
   JASSERT(coordinatorListenerSocket.isValid())
-    (coordinatorListenerSocket.port()) (JASSERT_ERRNO)
-    .Text("Failed to create listen socket."
-          "\nIf msg is \"Address already in use\", this may be an old coordinator."
-          "\nKill other coordinators and try again in a minute or so.");
+    (coordinatorListenerSocket.port()) (JASSERT_ERRNO) (host) (port)
+    .Text("Failed to create socket to coordinator port."
+          "\nIf the above message (sterror) is:"
+          "\n           \"Address already in use\" or \"Bad file descriptor\","
+          "\n  then this may be an old coordinator."
+          "\nEither try again a few seconds or a minute later,"
+          "\nOr kill other coordinator (using same host and port):"
+          "\n    dmtcp_command ---coord-host XXX --coord-port YYY --quit"
+          "\nOr specify --join-coordinator if joining existing computation.");
   // Now dup the sockfd to
   coordinatorListenerSocket.changeFd(PROTECTED_COORD_FD);
-  dmtcp::string coordPort= jalib::XToString(coordinatorListenerSocket.port());
-  setenv (ENV_VAR_NAME_PORT, coordPort.c_str(), 1);
+  Util::setCoordPort(coordinatorListenerSocket.port());
 
-  JTRACE("Starting a new coordinator automatically.") (coordPort);
+  JLOG(DMTCP)("Starting a new coordinator automatically.")
+        (coordinatorListenerSocket.port());
 
   if (fork() == 0) {
-    dmtcp::string coordinator = jalib::Filesystem::FindHelperUtility("dmtcp_coordinator");
+    /* NOTE:  This code assumes that dmtcp_launch (the current program)
+     *  and dmtcp_coordinator are in the same directory.  Namely,
+     *  GetProgramDir() gets the dir of the current program (dmtcp_launch).
+     *  Hence, if dmtcp_coordinator is in a different directory, then
+     *     jalib::Filesystem::GetProgramDir() + "/dmtcp_coordinator"
+     *  will not exist, and the child will fail.
+     */
+    // We can't use Util::getPath() here since the SharedData has not been
+    // initialized yet.
+    string coordinator =
+      jalib::Filesystem::GetProgramDir() + "/dmtcp_coordinator";
+
     char *modeStr = (char *)"--daemon";
     char * args[] = {
       (char*)coordinator.c_str(),
       (char*)"--quiet",
+      /* If we wish to also suppress coordinator warnings, call --quiet twice */
       (char*)"--exit-on-last",
       modeStr,
       NULL
@@ -481,19 +522,18 @@ void dmtcp::CoordinatorAPI::startNewCoordinator(CoordinatorAPI::CoordinatorMode
     execv(args[0], args);
     JASSERT(false)(coordinator)(JASSERT_ERRNO) .Text("exec(dmtcp_coordinator) failed");
   } else {
-    _real_close (PROTECTED_COORD_FD);
+    int status;
+    _real_close(PROTECTED_COORD_FD);
+    JASSERT(wait(&status) > 0) (JASSERT_ERRNO);
   }
-
-  int status;
-  JASSERT(wait(&status) > 0) (JASSERT_ERRNO);
 }
 
-void CoordinatorAPI::createNewConnToCoord(CoordinatorAPI::CoordinatorMode mode)
+void CoordinatorAPI::createNewConnToCoord(CoordinatorMode mode)
 {
   if (mode & COORD_JOIN) {
     _coordinatorSocket = createNewSocketToCoordinator(mode);
     JASSERT(_coordinatorSocket.isValid()) (JASSERT_ERRNO)
-      .Text("Coordinator not found, but --join was specified. Exiting.");
+     .Text("Coordinator not found, but --join-coordinator specified. Exiting.");
   } else if (mode & COORD_NEW) {
     startNewCoordinator(mode);
     _coordinatorSocket = createNewSocketToCoordinator(mode);
@@ -502,7 +542,7 @@ void CoordinatorAPI::createNewConnToCoord(CoordinatorAPI::CoordinatorMode mode)
   } else if (mode & COORD_ANY) {
     _coordinatorSocket = createNewSocketToCoordinator(mode);
     if (!_coordinatorSocket.isValid()) {
-      JTRACE("Coordinator not found, trying to start a new one.");
+      JLOG(DMTCP)("Coordinator not found, trying to start a new one.");
       startNewCoordinator(mode);
       _coordinatorSocket = createNewSocketToCoordinator(mode);
       JASSERT(_coordinatorSocket.isValid()) (JASSERT_ERRNO)
@@ -514,9 +554,9 @@ void CoordinatorAPI::createNewConnToCoord(CoordinatorAPI::CoordinatorMode mode)
   _coordinatorSocket.changeFd(PROTECTED_COORD_FD);
 }
 
-DmtcpMessage dmtcp::CoordinatorAPI::sendRecvHandshake(DmtcpMessage msg,
-                                                      string progname,
-                                                      UniquePid *compId)
+DmtcpMessage CoordinatorAPI::sendRecvHandshake(DmtcpMessage msg,
+                                               string progname,
+                                               UniquePid *compId)
 {
   if (dmtcp_virtual_to_real_pid) {
     msg.realPid = dmtcp_virtual_to_real_pid(getpid());
@@ -525,27 +565,19 @@ DmtcpMessage dmtcp::CoordinatorAPI::sendRecvHandshake(DmtcpMessage msg,
   }
 
   msg.theCheckpointInterval = getCkptInterval();
-  dmtcp::string hostname = jalib::Filesystem::GetCurrentHostname();
-  dmtcp::string prefixDir = getPrefixDir();
+  string hostname = jalib::Filesystem::GetCurrentHostname();
   msg.extraBytes = hostname.length() + 1 + progname.length() + 1;
-  if (!prefixDir.empty()) {
-    msg.extraBytes += prefixDir.length() + 1;
-  }
 
   _coordinatorSocket << msg;
   _coordinatorSocket.writeAll(hostname.c_str(), hostname.length() + 1);
   _coordinatorSocket.writeAll(progname.c_str(), progname.length() + 1);
-  if (!prefixDir.empty()) {
-    _coordinatorSocket.writeAll(prefixDir.c_str(), prefixDir.length() + 1);
-    msg.extraBytes += prefixDir.length() + 1;
-  }
 
   msg.poison();
   _coordinatorSocket >> msg;
   msg.assertValid();
   if (msg.type == DMT_KILL_PEER) {
-    JTRACE("Received KILL message from coordinator, exiting");
-    _exit (0);
+    JLOG(DMTCP)("Received KILL message from coordinator, exiting");
+    _real_exit (0);
   }
   if (msg.type == DMT_REJECT_NOT_RUNNING) {
     JASSERT(false)
@@ -557,22 +589,27 @@ DmtcpMessage dmtcp::CoordinatorAPI::sendRecvHandshake(DmtcpMessage msg,
     JASSERT(false) (*compId)
       .Text("Connection rejected by the coordinator.\n"
             " Reason: This process has a different computation group.");
-  } else if (msg.type == DMT_REJECT_WRONG_PREFIX) {
-    JASSERT(false) (prefixDir)
-      .Text("Connection rejected by the coordinator.\n"
-            "Reason: Prefix supplied with --prefix doesn't match the prefix\n"
-            "        of other processes in the computation.\n"
-            "        (DMTCP installed at different paths?)");
   }
-  JASSERT(msg.type == DMT_ACCEPT);
+  // Coordinator also prints this, but its stderr may go to /dev/null
+  if (msg.type == DMT_REJECT_NOT_RESTARTING) {
+    string coordinatorHost = ""; // C++ magic code; "" to be invisibly replaced
+    int coordinatorPort = 0;
+    Util::getCoordHostAndPort(COORD_ANY, coordinatorHost, &coordinatorPort);
+    JNOTE ("\n\n*** Computation not in RESTARTING or CHECKPOINTED state."
+        "\n***Can't join the existing coordinator, as it is serving a"
+        "\n***different computation.  Consider launching a new coordinator."
+        "\n***Consider, also, checking with:  dmtcp_command --status")
+        (coordinatorPort);
+  }
+  JASSERT(msg.type == DMT_ACCEPT)(msg.type);
   return msg;
 }
 
-void dmtcp::CoordinatorAPI::connectToCoordOnStartup(CoordinatorMode  mode,
-                                                    string           progname,
-                                                    DmtcpUniqueProcessId *compId,
-                                                    CoordinatorInfo *coordInfo,
-                                                    struct in_addr  *localIP)
+void CoordinatorAPI::connectToCoordOnStartup(CoordinatorMode mode,
+                                             string progname,
+                                             DmtcpUniqueProcessId *compId,
+                                             CoordinatorInfo *coordInfo,
+                                             struct in_addr  *localIP)
 {
   JASSERT(compId != NULL && localIP != NULL && coordInfo != NULL);
 
@@ -583,15 +620,17 @@ void dmtcp::CoordinatorAPI::connectToCoordOnStartup(CoordinatorMode  mode,
   }
 
   createNewConnToCoord(mode);
-  JTRACE("sending coordinator handshake")(UniquePid::ThisProcess());
+  JLOG(DMTCP)("sending coordinator handshake")(UniquePid::ThisProcess());
   DmtcpMessage hello_local(DMT_NEW_WORKER);
   hello_local.virtualPid = -1;
 
   DmtcpMessage hello_remote = sendRecvHandshake(hello_local, progname);
 
   JASSERT(hello_remote.virtualPid != -1);
-  JTRACE("Got virtual pid from coordinator") (hello_remote.virtualPid);
-  dmtcp::Util::setVirtualPidEnvVar(hello_remote.virtualPid, getppid());
+  JLOG(DMTCP)("Got virtual pid from coordinator") (hello_remote.virtualPid);
+
+  pid_t ppid = getppid();
+  Util::setVirtualPidEnvVar(hello_remote.virtualPid, ppid, ppid);
 
   JASSERT(compId != NULL && localIP != NULL && coordInfo != NULL);
   *compId = hello_remote.compGroup.upid();
@@ -605,9 +644,12 @@ void dmtcp::CoordinatorAPI::connectToCoordOnStartup(CoordinatorMode  mode,
   memcpy(localIP, &hello_remote.ipAddr, sizeof hello_remote.ipAddr);
 }
 
-void dmtcp::CoordinatorAPI::createNewConnectionBeforeFork(string& progname)
+void CoordinatorAPI::createNewConnectionBeforeFork(string& progname)
 {
-  JASSERT(!noCoordinator());
+  JASSERT(!noCoordinator())
+    .Text("Process attempted to call fork() while in --no-coordinator mode\n"
+          "  Because the coordinator is embedded in a single process,\n"
+          "    DMTCP will not work with multiple processes.");
   struct sockaddr_storage addr;
   uint32_t len;
   SharedData::getCoordAddr((struct sockaddr *)&addr, &len);
@@ -620,17 +662,21 @@ void dmtcp::CoordinatorAPI::createNewConnectionBeforeFork(string& progname)
   JASSERT(hello_remote.virtualPid != -1);
 
   if (dmtcp_virtual_to_real_pid) {
-    JTRACE("Got virtual pid from coordinator") (hello_remote.virtualPid);
-    dmtcp::Util::setVirtualPidEnvVar(hello_remote.virtualPid, getpid());
+    JLOG(DMTCP)("Got virtual pid from coordinator") (hello_remote.virtualPid);
+    pid_t pid = getpid();
+    pid_t realPid = dmtcp_virtual_to_real_pid(pid);
+    Util::setVirtualPidEnvVar(hello_remote.virtualPid, pid, realPid);
   }
 }
 
-void dmtcp::CoordinatorAPI::connectToCoordOnRestart(CoordinatorMode  mode,
-                                                    dmtcp::string progname,
-                                                    UniquePid compGroup,
-                                                    int np,
-                                                    CoordinatorInfo *coordInfo,
-                                                    struct in_addr  *localIP)
+void CoordinatorAPI::connectToCoordOnRestart(CoordinatorMode  mode,
+                                             string progname,
+                                             UniquePid compGroup,
+                                             int np,
+                                             CoordinatorInfo *coordInfo,
+                                             const char *host,
+                                             int port,
+                                             struct in_addr  *localIP)
 {
   if (mode & COORD_NONE) {
     setupVirtualCoordinator(coordInfo, localIP);
@@ -638,7 +684,7 @@ void dmtcp::CoordinatorAPI::connectToCoordOnRestart(CoordinatorMode  mode,
   }
 
   createNewConnToCoord(mode);
-  JTRACE("sending coordinator handshake")(UniquePid::ThisProcess());
+  JLOG(DMTCP)("sending coordinator handshake")(UniquePid::ThisProcess());
   DmtcpMessage hello_local(DMT_RESTART_WORKER);
   hello_local.virtualPid = -1;
   hello_local.numPeers = np;
@@ -660,42 +706,51 @@ void dmtcp::CoordinatorAPI::connectToCoordOnRestart(CoordinatorMode  mode,
     memcpy(localIP, &hello_remote.ipAddr, sizeof hello_remote.ipAddr);
   }
 
-  JTRACE("Coordinator handshake RECEIVED!!!!!");
+  JLOG(DMTCP)("Coordinator handshake RECEIVED!!!!!");
 }
 
-void dmtcp::CoordinatorAPI::sendCkptFilename()
+void CoordinatorAPI::sendCkptFilename()
 {
   if (noCoordinator()) return;
   // Tell coordinator to record our filename in the restart script
-  dmtcp::string ckptFilename = dmtcp::ProcessInfo::instance().getCkptFilename();
-  dmtcp::string hostname = jalib::Filesystem::GetCurrentHostname();
-  JTRACE("recording filenames") (ckptFilename) (hostname);
-  dmtcp::DmtcpMessage msg;
+  string ckptFilename = ProcessInfo::instance().getCkptFilename();
+  string hostname = jalib::Filesystem::GetCurrentHostname();
+  DmtcpMessage msg;
   if (dmtcp_unique_ckpt_enabled && dmtcp_unique_ckpt_enabled()) {
     msg.type = DMT_UNIQUE_CKPT_FILENAME;
   } else {
     msg.type = DMT_CKPT_FILENAME;
   }
-  msg.extraBytes = ckptFilename.length() + 1 + hostname.length() + 1;
+  // Tell coordinator type of remote shell command used ssh/rsh
+  const char* shellType = "";
+  const char *remoteShellType = getenv(ENV_VAR_REMOTE_SHELL_CMD);
+  if(remoteShellType != NULL) {
+    shellType = remoteShellType;
+  }
+
+  JLOG(DMTCP)("recording filenames") (ckptFilename) (hostname) (shellType);
+  msg.extraBytes = ckptFilename.length() + 1 + hostname.length() + 1+ strlen(shellType) + 1;
+ 
   _coordinatorSocket << msg;
   _coordinatorSocket.writeAll(ckptFilename.c_str(), ckptFilename.length() + 1);
+  _coordinatorSocket.writeAll(shellType, strlen(shellType) + 1);
   _coordinatorSocket.writeAll(hostname.c_str(), hostname.length() + 1);
 }
 
 
-int dmtcp::CoordinatorAPI::sendKeyValPairToCoordinator(const char *id,
-                                                       const void *key,
-                                                       uint32_t key_len,
-                                                       const void *val,
-                                                       uint32_t val_len,
-						       int sync)
+int CoordinatorAPI::sendKeyValPairToCoordinator(const char *id,
+                                                const void *key,
+                                                uint32_t key_len,
+                                                const void *val,
+                                                uint32_t val_len,
+                                                int sync)
 {
   DmtcpMessage msg (DMT_REGISTER_NAME_SERVICE_DATA);
   if (sync) {
     msg.type = DMT_REGISTER_NAME_SERVICE_DATA_SYNC;
   }
   JWARNING(strlen(id) < sizeof(msg.nsid));
-  strncpy(msg.nsid, id, 8);
+  strncpy(msg.nsid, id, sizeof msg.nsid);
   msg.keyLen = key_len;
   msg.valLen = val_len;
   msg.extraBytes = key_len + val_len;
@@ -728,15 +783,15 @@ int dmtcp::CoordinatorAPI::sendKeyValPairToCoordinator(const char *id,
 //   size of that buffer (the memory allocated by user).
 // On output, we copy data to val, and set *val_len to the actual buffer size
 //   (to the size of the data that we copied to the user buffer).
-int dmtcp::CoordinatorAPI::sendQueryToCoordinator(const char *id,
-                                                  const void *key,
-                                                  uint32_t key_len,
-                                                  void *val,
-                                                  uint32_t *val_len)
+int CoordinatorAPI::sendQueryToCoordinator(const char *id,
+                                           const void *key,
+                                           uint32_t key_len,
+                                           void *val,
+                                           uint32_t *val_len)
 {
   DmtcpMessage msg (DMT_NAME_SERVICE_QUERY);
   JWARNING(strlen(id) < sizeof(msg.nsid));
-  strncpy(msg.nsid, id, 8);
+  strncpy(msg.nsid, id, sizeof msg.nsid);
   msg.keyLen = key_len;
   msg.valLen = 0;
   msg.extraBytes = key_len;
@@ -775,3 +830,55 @@ int dmtcp::CoordinatorAPI::sendQueryToCoordinator(const char *id,
 
   return *val_len;
 }
+
+int CoordinatorAPI::getUniqueIdFromCoordinator(const char *id,
+                                               const void *key,
+                                               uint32_t key_len,
+                                               void *val,
+                                               uint32_t *val_len,
+                                               uint32_t offset /* = 1 */)
+{
+  DmtcpMessage msg(DMT_NAME_SERVICE_GET_UNIQUE_ID);
+
+  JWARNING(strlen(id) < sizeof(msg.nsid));
+  strncpy(msg.nsid, id, sizeof msg.nsid);
+  msg.keyLen = key_len;
+  msg.valLen = 0;
+  msg.extraBytes = key_len;
+  msg.uniqueIdOffset = offset;
+  jalib::JSocket sock = _coordinatorSocket;
+
+  if (key == NULL || key_len == 0 || val == NULL || val_len == 0) {
+    return 0;
+  }
+
+  msg.valLen = *val_len;
+
+  if (dmtcp_is_running_state()) {
+    if (!_nsSock.isValid()) {
+      _nsSock = createNewSocketToCoordinator(COORD_ANY);
+      JASSERT(_nsSock.isValid());
+      _nsSock.changeFd(PROTECTED_NS_FD);
+      DmtcpMessage m(DMT_NAME_SERVICE_WORKER);
+      _nsSock << m;
+    }
+    sock = _nsSock;
+    JASSERT(sock.isValid());
+  }
+
+  JASSERT(Util::writeAll(sock, &msg, sizeof(msg)) == sizeof(msg));
+  JASSERT(Util::writeAll(sock, key, key_len) == key_len);
+
+  msg.poison();
+
+  JASSERT(Util::readAll(sock, &msg, sizeof(msg)) == sizeof(msg));
+  msg.assertValid();
+  JASSERT(msg.type == DMT_NAME_SERVICE_GET_UNIQUE_ID_RESPONSE &&
+          msg.extraBytes == msg.valLen);
+
+  JASSERT(*val_len >= msg.valLen);
+  *val_len = msg.valLen;
+  JASSERT(Util::readAll(sock, val, *val_len) == *val_len);
+
+  return *val_len;
+}
