@@ -34,7 +34,9 @@
 
 using namespace dmtcp;
 
-static void processArgs(int *orig_argc, char ***orig_argv);
+static void processArgs(int *orig_argc, char ***orig_argv,
+                        string *tmpDir_p, string &host,
+                        const char **portStr);
 static int testMatlab(const char *filename);
 static int testJava(char **argv);
 static bool testSetuid(const char *filename);
@@ -49,30 +51,33 @@ static const char* theUsage =
   "Usage: dmtcp_launch [OPTIONS] <command> [args...]\n"
   "Start a process under DMTCP control.\n\n"
   "Connecting to the DMTCP Coordinator:\n"
-  "  -h, --host HOSTNAME (environment variable DMTCP_HOST)\n"
+  "  -h, --coord-host HOSTNAME (environment variable DMTCP_COORD_HOST)\n"
   "              Hostname where dmtcp_coordinator is run (default: localhost)\n"
-  "  -p, --port PORT_NUM (environment variable DMTCP_PORT)\n"
-  "              Port where dmtcp_coordinator is run (default: 7779)\n"
+  "  -p, --coord-port PORT_NUM (environment variable DMTCP_COORD_PORT)\n"
+  "              Port where dmtcp_coordinator is run (default: "
+                                                  STRINGIFY(DEFAULT_PORT) ")\n"
   "  --port-file FILENAME\n"
-  "              File to write listener port number.\n"
-  "              (Useful with '--port 0', which is used to assign a random port)\n"
-  "  -j, --join\n"
+  "              File to write listener port number.  (Useful with\n"
+  "              '--coord-port 0', which is used to assign a random port)\n"
+  "  -j, --join-coordinator\n"
   "              Join an existing coordinator, raise error if one doesn't\n"
   "              already exist\n"
   "  --new-coordinator\n"
-  "              Create a new coordinator at the given port. Fail if one \n"
+  "              Create a new coordinator at the given port. Fail if one\n"
   "              already exists on the given port. The port can be specified\n"
-  "              with --port, or with environment variable DMTCP_PORT.  If no\n"
-  "              port is specified, start coordinator at a random port (same\n"
-  "              as specifying port '0').\n"
+  "              with --coord-port, or with environment variable DMTCP_COORD_PORT.\n"
+  "              If no port is specified, start coordinator at a random port\n"
+  "              (same as specifying port '0').\n"
   "  --no-coordinator\n"
-  "              Execute the process in stand-alone coordinator-less mode.\n"
+  "              Execute the process in standalone coordinator-less mode.\n"
   "              Use dmtcp_command or --interval to request checkpoints.\n"
-  "  -i, -interval SECONDS (environment variable DMTCP_CHECKPOINT_INTERVAL)\n"
+  "              Note that this is incompatible with calls to fork(), since\n"
+  "              an embedded coordinator runs in the original process only.\n"
+  "  -i, --interval SECONDS (environment variable DMTCP_CHECKPOINT_INTERVAL)\n"
   "              Time in seconds between automatic checkpoints.\n"
   "              0 implies never (manual ckpt only); if not set and no env var,\n"
   "              use default value set in dmtcp_coordinator or dmtcp_command.\n"
-  "              Not allowed if --join is specified\n"
+  "              Not allowed if --join-coordinator is specified\n"
   "\n"
   "Checkpoint image generation:\n"
   "  --gzip, --no-gzip, (environment variable DMTCP_GZIP=[01])\n"
@@ -89,10 +94,13 @@ static const char* theUsage =
   "  --checkpoint-open-files\n"
   "              Checkpoint open files and restore old working dir.\n"
   "              (default: do neither)\n"
+  "  --allow-file-overwrite\n"
+  "              If used with --checkpoint-open-files, allows a saved file\n"
+  "              to overwrite its existing copy at original location\n"
+  "              (default: file overwrites are not allowed)\n"
   "  --ckpt-signal signum\n"
-  "  --mtcp-checkpoint-signal signum\n"
   "              Signal number used internally by DMTCP for checkpointing\n"
-  "              (default: 12). --mtcp-checkpoint-signal is deprecated.\n"
+  "              (default: SIGUSR2/12).\n"
   "\n"
   "Enable/disable plugins:\n"
   "  --with-plugin (environment variable DMTCP_PLUGIN)\n"
@@ -101,13 +109,16 @@ static const char* theUsage =
   "  --batch-queue, --rm\n"
   "              Enable support for resource managers (Torque PBS and SLURM).\n"
   "              (default: disabled)\n"
-  "  --ptrace    \n"
+  "  --ptrace\n"
   "              Enable support for PTRACE system call for gdb/strace etc.\n"
   "              (default: disabled)\n"
   "  --modify-env\n"
-  "              Update environment variables based on the environment on the \n"
+  "              Update environment variables based on the environment on the\n"
   "              restart host (e.g., DISPLAY=$DISPLAY).\n"
-  "              This can be set in a file dmtcp_env.txt. \n"
+  "              This can be set in a file dmtcp_env.txt.\n"
+  "              (default: disabled)\n"
+  "  --pathvirt\n"
+  "              Update file pathnames based on DMTCP_PATH_PREFIX\n"
   "              (default: disabled)\n"
   "  --ib, --infiniband\n"
   "              Enable InfiniBand plugin. (default: disabled)\n"
@@ -119,13 +130,14 @@ static const char* theUsage =
   "              Disable all plugins.\n"
   "\n"
   "Other options:\n"
-  "  --prefix PATH\n"
-  "              Prefix where DMTCP is installed on remote nodes.\n"
   "  --tmpdir PATH (environment variable DMTCP_TMPDIR)\n"
-  "              Directory to store temporary files \n"
-  "              (default: $TMDPIR/dmtcp-$USER@$HOST or /tmp/dmtcp-$USER@$HOST)\n"
+  "              Directory to store temporary files (default: $TMDPIR or /tmp)\n"
+  "              (Behavior is undefined if two launched processes specify\n"
+  "               different tmpdirs.)\n"
   "  -q, --quiet (or set environment variable DMTCP_QUIET = 0, 1, or 2)\n"
   "              Skip NOTE messages; if given twice, also skip WARNINGs\n"
+  "  --coord-logfile PATH (environment variable DMTCP_COORD_LOG_FILENAME\n"
+  "              Coordinator will dump its logs to the given file\n"
   "  --help\n"
   "              Print this message and exit.\n"
   "  --version\n"
@@ -144,12 +156,12 @@ static const char* theUsage =
 // ;
 
 static bool disableAllPlugins=false;
-static bool isSSHSlave=false;
 static bool checkpointOpenFiles=false;
 
 static bool enablePtracePlugin=false;
 static bool enableModifyEnvPlugin=false;
 static bool enableRMPlugin=false;
+static bool explicitSrun = false;
 
 static bool enableIB2TcpPlugin=false;
 static bool enableIBPlugin=false;
@@ -157,15 +169,22 @@ static bool enableIBPlugin=false;
 static bool enableAllocPlugin=true;
 static bool enableDlPlugin=true;
 static bool enableIPCPlugin=true;
-static bool enableLibDMTCP=true;
-static bool enablePIDPlugin=true;
+static bool enableSvipcPlugin=true;
+static bool enablePathVirtPlugin=false;
+static bool enableTimerPlugin=true;
+
 #ifdef UNIQUE_CHECKPOINT_FILENAMES
 static bool enableUniqueCkptPlugin=true;
 #else
 static bool enableUniqueCkptPlugin=false;
 #endif
+// This is the base library.
+static bool enableLibDMTCP=true;
 
-static dmtcp::string thePortFile;
+// PID plugin must come last.
+static bool enablePIDPlugin=true;
+
+static string thePortFile;
 
 struct PluginInfo {
   bool *enabled;
@@ -182,20 +201,30 @@ static struct PluginInfo pluginInfo[] = {               // Default value
   {&enableAllocPlugin,      "libdmtcp_alloc.so"},       // Enabled
   {&enableDlPlugin,         "libdmtcp_dl.so"},          // Enabled
   {&enableIPCPlugin,        "libdmtcp_ipc.so"},         // Enabled
+  {&enableSvipcPlugin,      "libdmtcp_svipc.so"},       // Enabled
+  {&enablePathVirtPlugin,   "libdmtcp_pathvirt.so"},    // Enabled
+  {&enableTimerPlugin,      "libdmtcp_timer.so"},       // Enabled
   {&enableLibDMTCP,         "libdmtcp.so"},             // Enabled
+  // PID plugin must come last.
   {&enablePIDPlugin,        "libdmtcp_pid.so"}          // Enabled
 };
 
 const size_t numLibs = sizeof(pluginInfo) / sizeof (struct PluginInfo);
 
-static CoordinatorAPI::CoordinatorMode allowedModes = CoordinatorAPI::COORD_ANY;
+static CoordinatorMode allowedModes = COORD_ANY;
+static uint32_t mask = jassert_internal::UNKNOWN;
 
 //shift args
 #define shift argc--,argv++
-static void processArgs(int *orig_argc, char ***orig_argv)
+static void processArgs(int *orig_argc, char ***orig_argv,
+                        string *tmpDir_p,
+                        string &host, const char **portStr)
 {
-  char argc = *orig_argc;
+  int argc = *orig_argc;
   char **argv = *orig_argv;
+  char *tmpdir_arg = NULL;
+  host = ""; // uninitialized
+  *portStr = NULL; // uninitialized
 
   if (argc == 1) {
     printf("%s", DMTCP_VERSION_AND_COPYRIGHT_INFO);
@@ -206,20 +235,18 @@ static void processArgs(int *orig_argc, char ***orig_argv)
   //process args
   shift;
   while (true) {
-    dmtcp::string s = argc>0 ? argv[0] : "--help";
-    if ((s=="--help") && argc==1) {
+    string s = argc>0 ? argv[0] : "--help";
+    if ((s=="--help") && argc<=1) {
       printf("%s", theUsage);
       exit(DMTCP_FAIL_RC);
     } else if ((s=="--version") && argc==1) {
       printf("%s", DMTCP_VERSION_AND_COPYRIGHT_INFO);
       exit(DMTCP_FAIL_RC);
-    } else if (s=="--ssh-slave") {
-      isSSHSlave = true;
-      shift;
-    } else if (s == "-j" || s == "--join") {
-      allowedModes = dmtcp::CoordinatorAPI::COORD_JOIN;
+    } else if (s == "-j" || s == "--join-coordinator" || s == "--join") {
+      allowedModes = COORD_JOIN;
       shift;
     } else if (s == "--gzip") {
+      /* NOTE: If gzip flags not set, default is --gzip */
       setenv(ENV_VAR_COMPRESSION, "1", 1);
       shift;
     } else if (s == "--no-gzip") {
@@ -236,52 +263,61 @@ static void processArgs(int *orig_argc, char ***orig_argv)
     }
 #endif
     else if (s == "--new-coordinator") {
-      allowedModes = dmtcp::CoordinatorAPI::COORD_NEW;
+      allowedModes = COORD_NEW;
       shift;
     } else if (s == "--no-coordinator") {
-      allowedModes = dmtcp::CoordinatorAPI::COORD_NONE;
+      allowedModes = COORD_NONE;
       shift;
-    } else if (s == "-i" || s == "--interval" ||
-             (s.c_str()[0] == '-' && s.c_str()[1] == 'i' &&
-              isdigit(s.c_str()[2]) ) ) {
-      if (isdigit(s.c_str()[2])) { // if -i5, for example
-        setenv(ENV_VAR_CKPT_INTR, s.c_str()+2, 1);
-        shift;
-      } else { // else -i 5
-        setenv(ENV_VAR_CKPT_INTR, argv[1], 1);
-        shift; shift;
-      }
-    } else if (argc>1 && (s == "-h" || s == "--host")) {
-      setenv(ENV_VAR_NAME_HOST, argv[1], 1);
+    } else if (s == "-i" || s == "--interval") {
+      setenv(ENV_VAR_CKPT_INTR, argv[1], 1);
       shift; shift;
-    } else if (argc>1 && (s == "-p" || s == "--port")) {
-      setenv(ENV_VAR_NAME_PORT, argv[1], 1);
+    } else if (s == "--coord-logfile") {
+      setenv(ENV_VAR_COORD_LOGFILE, argv[1], 1);
       shift; shift;
+    } else if (s == "--debug-logs") {
+      mask = Util::processDebugLogsArg(argv[1]);
+      shift; shift;
+    } else if (argv[0][0] == '-' && argv[0][1] == 'i' &&
+               isdigit(argv[0][2])) { // else if -i5, for example
+      setenv(ENV_VAR_CKPT_INTR, argv[0]+2, 1);
+      shift;
+    } else if (argc>1 && (s == "-h" || s == "--coord-host" || s == "--host")) {
+      host = argv[1];
+      shift; shift;
+    } else if (argc>1 && (s == "-p" || s == "--coord-port" || s == "--port")) {
+      *portStr = argv[1];
+      shift; shift;
+    } else if (argv[0][0] == '-' && argv[0][1] == 'p' &&
+               isdigit(argv[0][2])) { // else if -p0, for example
+      *portStr = argv[0]+2; // Must use argv[0] here, and not s
+      shift;
     } else if (argc>1 && s == "--port-file"){
       thePortFile = argv[1];
       shift; shift;
-    } else if (argc>1 && (s == "--prefix")) {
-      setenv(ENV_VAR_PREFIX_PATH, argv[1], 1);
-      shift; shift;
     } else if (argc>1 && (s == "-c" || s == "--ckptdir")) {
       setenv(ENV_VAR_CHECKPOINT_DIR, argv[1], 1);
       shift; shift;
     } else if (argc>1 && (s == "-t" || s == "--tmpdir")) {
-      setenv(ENV_VAR_TMPDIR, argv[1], 1);
+      tmpdir_arg = argv[1];
       shift; shift;
-    } else if (argc>1 && (s == "--mtcp-checkpoint-signal" ||
-                          s == "--ckpt-signal")) {
+    } else if (argc>1 && s == "--ckpt-signal") {
       setenv(ENV_VAR_SIGCKPT, argv[1], 1);
       shift; shift;
     } else if (s == "--checkpoint-open-files" || s == "--ckpt-open-files") {
       checkpointOpenFiles = true;
       shift;
+    } else if (s == "--allow-file-overwrite") {
+      setenv(ENV_VAR_ALLOW_OVERWRITE_WITH_CKPTED_FILES, "1", 0);
+      shift;
     } else if (s == "--ptrace") {
       enablePtracePlugin = true;
       shift;
     } else if (s == "--modify-env") {
       enableModifyEnvPlugin = true;
       shift;
+    } else if (s == "--pathvirt") {
+      enablePathVirtPlugin = true;
+      shift;
     } else if (s == "--ib" || s == "--infiniband") {
       enableIBPlugin = true;
       shift;
@@ -300,9 +336,15 @@ static void processArgs(int *orig_argc, char ***orig_argv)
     } else if (s == "--rm" || s == "--batch-queue") {
       enableRMPlugin = true;
       shift;
+    } else if (s == "--explicit-srun" ) {
+      explicitSrun = true;
+      shift;
     } else if (s == "--with-plugin") {
       setenv(ENV_VAR_PLUGIN, argv[1], 1);
       shift; shift;
+    } else if (s == "--with-plugin-32") {
+      setenv(ENV_VAR_PLUGIN_32, argv[1], 1);
+      shift; shift;
     } else if (s == "-q" || s == "--quiet") {
       *getenv(ENV_VAR_QUIET) = *getenv(ENV_VAR_QUIET) + 1;
       // Just in case a non-standard version of setenv is being used:
@@ -319,37 +361,54 @@ static void processArgs(int *orig_argc, char ***orig_argv)
       break;
     }
   }
+ 
+#ifdef FAST_RST_VIA_MMAP
+  // In case of fast restart, we shall not use gzip. 
+  setenv(ENV_VAR_COMPRESSION, "0", 1);
+#endif
+
+#if __aarch64__
+  /* FIXME:  Currently, there is a bug exposed by SIGRETURN for aarch64,
+   *      when we create a SIGCHLD handler for the gzip process.
+   *      So, we're temporarily disabling GZIP for aarch64.
+   * NOTE:  This occurs _only_ on second CKPT of 'make check' tests.
+   */
+  if (getenv(ENV_VAR_COMPRESSION) == NULL /* NULL default => --gzip */ ||
+      strcmp(getenv(ENV_VAR_COMPRESSION), "1") == 0) {
+    setenv(ENV_VAR_COMPRESSION, "0", 1);
+    if (getenv(ENV_VAR_QUIET) != NULL &&
+        strcmp(getenv(ENV_VAR_QUIET), "0") == 0) {
+      JASSERT_STDERR <<
+        "\n*** Turning off gzip compression.  The armv8 CPU support is"
+        " still experimental.\n*** Gzip not yet supported.\n\n";
+    }
+  }
+#endif
+  *tmpDir_p = Util::calcTmpDir(tmpdir_arg);
+  *orig_argc = argc;
   *orig_argv = argv;
 }
 
 int main ( int argc, char** argv )
 {
-  for (size_t fd = PROTECTED_FD_START; fd < PROTECTED_FD_END; fd++) {
+  Util::setProtectedFdBase();
+  for (int fd = PROTECTED_FD_START; fd < PROTECTED_FD_END; fd++) {
     close(fd);
   }
 
   if (! getenv(ENV_VAR_QUIET))
     setenv(ENV_VAR_QUIET, "0", 0);
 
-  processArgs(&argc, &argv);
+  string tmpDir = "tmpDir is not set";
+  string host;
+  const char *portStr;
+  // This will change argv to refer to the target application.
+  processArgs(&argc, &argv, &tmpDir, host, &portStr);
 
   initializeJalib();
-  // If --ssh-slave and --prefix both are present, verify that the prefix-dir
-  // of this binary (dmtcp_launch) is same as the one provided with
-  // --prefix
-  if (isSSHSlave && getenv(ENV_VAR_PREFIX_PATH) != NULL) {
-    char buf[PATH_MAX];
-    string prefixPath = getenv(ENV_VAR_PREFIX_PATH);
-    prefixPath += "/bin/dmtcp_launch";
-    JASSERT(realpath(prefixPath.c_str(), buf) != NULL) (prefixPath);
-    prefixPath = buf;
-    string programPath = jalib::Filesystem::GetProgramPath();
-    JASSERT(prefixPath == programPath) (prefixPath) (programPath);
-  }
 
-  dmtcp::Util::setTmpDir(getenv(ENV_VAR_TMPDIR));
-  dmtcp::UniquePid::ThisProcess(true);
-  dmtcp::Util::initializeLogFile();
+  UniquePid::ThisProcess(true);
+  Util::initializeLogFile(tmpDir);
 
 #ifdef FORKED_CHECKPOINTING
   /* When this is robust, add --forked-checkpointing option on command-line,
@@ -411,7 +470,7 @@ int main ( int argc, char** argv )
   if (testSetuid(argv[0])) {
     char **newArgv;
     // THIS NEXT LINE IS DANGEROUS.  MOST setuid PROGRAMS CAN'T RUN UNPRIVILEGED
-    dmtcp::Util::patchArgvIfSetuid(argv[0], argv, &newArgv);
+    Util::patchArgvIfSetuid(argv[0], argv, &newArgv);
     argv = newArgv;
   };
 
@@ -424,7 +483,7 @@ int main ( int argc, char** argv )
     const char* ckptDir = get_current_dir_name();
     if(ckptDir != NULL ){
       //copy to private buffer
-      static dmtcp::string _buf = ckptDir;
+      static string _buf = ckptDir;
       ckptDir = _buf.c_str();
     }else{
       ckptDir=".";
@@ -439,7 +498,7 @@ int main ( int argc, char** argv )
     unsetenv( ENV_VAR_CKPT_OPEN_FILES);
 
   bool isElf, is32bitElf;
-  if  (dmtcp::Util::elfType(argv[0], &isElf, &is32bitElf) == -1) {
+  if  (Util::elfType(argv[0], &isElf, &is32bitElf) == -1) {
     // Couldn't read argv_buf
     // FIXME:  This could have been a symbolic link.  Don't issue an error,
     //         unless we're sure that the executable is not readable.
@@ -458,29 +517,74 @@ int main ( int argc, char** argv )
     unsetenv("DISPLAY");
   }
 
+  // Unset SESSION_MANAGER environment variable. SESSION_MANAGER is used by the
+  // X Window system to save and restore the current state of applications.
+  // For example, on a session crash, the X session manager attempts to reopen
+  // the applications that were open before the crash.  An application has to
+  // create a (socket) connection to the SESSION_MANAGER to communicate the
+  // current state.
+  // See http://en.wikipedia.org/wiki/X_session_manager for more information.
+  //
+  // For our purposes, we don't care about session management and thus we
+  // should disable the SESSION_MANAGER environment variable to prevent the
+  // application from communication with the X session manager.
+  if (getenv("SESSION_MANAGER") != NULL) {
+    JTRACE("Unsetting SESSION_MANAGER environment variable.");
+    unsetenv("SESSION_MANAGER");
+  }
+
+  if( explicitSrun ){
+      setenv(ENV_VAR_EXPLICIT_SRUN, "1", 1);
+  }
+
 // FIXME:  Unify this code with code prior to execvp in execwrappers.cpp
 //   Can use argument to dmtcpPrepareForExec() or getenv("DMTCP_...")
 //   from DmtcpWorker constructor, to distinguish the two cases.
-  dmtcp::Util::adjustRlimitStack();
-
-  // Set DLSYM_OFFSET env var(s).
-  dmtcp::Util::prepareDlsymWrapper();
+  Util::adjustRlimitStack();
 
   DmtcpUniqueProcessId compId;
   CoordinatorInfo coordInfo;
   struct in_addr localIPAddr;
+  int port = (portStr ? jalib::StringToInt(portStr) : UNINITIALIZED_PORT);
+  // Initialize host and port now.  Will be used in low-level functions.
+  Util::getCoordHostAndPort(allowedModes, host, &port);
   CoordinatorAPI::instance().connectToCoordOnStartup(allowedModes, argv[0],
                                                      &compId, &coordInfo,
                                                      &localIPAddr);
-  Util::writeCoordPortToFile(getenv(ENV_VAR_NAME_PORT), thePortFile.c_str());
+  // If port was 0, we'll get new random port when coordinator starts up.
+  Util::getCoordHostAndPort(allowedModes, host, &port);
+  Util::writeCoordPortToFile(port, thePortFile.c_str());
+
+  string installDir =
+    jalib::Filesystem::DirName(jalib::Filesystem::GetProgramDir());
+
+#if defined(__i386__) || defined(__arm__)
+  if (Util::strEndsWith(installDir, "/lib/dmtcp/32")) {
+    // If dmtcp_launch was compiled for 32 bits in a 64-bit O/S, then note:
+    // DMTCP_ROOT/bin/dmtcp_launch is a symbolic link to:
+    //    DMTCP_ROOT/bin/dmtcp_launch/lib/dmtcp/32/bin
+    // GetProgramDir() followed the link.  So, we need to remove the suffix.
+    char *str = const_cast<char*>(installDir.c_str());
+    str[strlen(str) - strlen("/lib/dmtcp/32")] = '\0';
+    installDir = str;
+  }
+#endif
+
   /* We need to initialize SharedData here to make sure that it is
    * initialized with the correct coordinator timestamp.  The coordinator
    * timestamp is updated only during postCkpt callback. However, the
    * SharedData area may be initialized earlier (for example, while
    * recreating threads), causing it to use *older* timestamp.
    */
-  SharedData::initialize(Util::getTmpDir().c_str(), &compId, &coordInfo,
+  SharedData::initialize(tmpDir.c_str(),
+                         installDir.c_str(),
+                         &compId,
+                         &coordInfo,
                          &localIPAddr);
+  SharedData::setLogMask(mask);
+
+  // Set DLSYM_OFFSET env var(s).
+  Util::prepareDlsymWrapper();
 
   setLDPreloadLibs(is32bitElf);
 
@@ -508,12 +612,12 @@ static int testMatlab(const char *filename)
   static const char* theMatlabWarning =
     "\n**** WARNING:  Earlier Matlab releases (e.g. release 7.4) use an\n"
     "****  older glibc.  Later releases (e.g. release 7.9) have no problem.\n"
-    "****  \n"
+    "****\n"
     "****  If you are using an _earlier_ Matlab, please re-compile DMTCP\n"
     "****  with gcc-4.1 and g++-4.1\n"
     "**** env CC=gcc-4.1 CXX=g++-4.1 ./configure\n"
     "**** [ Also modify mtcp/Makefile to:  CC=gcc-4.1 ]\n"
-    "**** [ Next, you may need an alternative Java JVM (see QUICK-START) ]\n"
+    "**** [ Next, you may need an alternative Java JVM (see QUICK-START.md) ]\n"
     "**** [ Finally, run as:   dmtcp_launch matlab -nodisplay ]\n"
     "**** [   (DMTCP does not yet checkpoint X-Windows applications.) ]\n"
     "**** [ You may see \"Not checkpointing libc-2.7.so\".  This is normal. ]\n"
@@ -521,7 +625,9 @@ static int testMatlab(const char *filename)
 	    " executing.)\n\n" ;
 
   // FIXME:  should expand filename and "matlab" before checking
-  if ( strcmp(filename, "matlab") == 0 && getenv(ENV_VAR_QUIET) == NULL) {
+  if ( strcmp(filename, "matlab") == 0 &&
+       ( getenv(ENV_VAR_QUIET) == NULL || strcmp(getenv(ENV_VAR_QUIET), "0") ) )
+  {
     JASSERT_STDERR << theMatlabWarning;
     return -1;
   }
@@ -567,7 +673,7 @@ static int testJava(char **argv)
 
 static bool testSetuid(const char *filename)
 {
-  if (dmtcp::Util::isSetuid(filename) &&
+  if (Util::isSetuid(filename) &&
       strcmp(filename, "screen") != 0 && strstr(filename, "/screen") == NULL) {
 
     static const char* theSetuidWarning =
@@ -586,7 +692,7 @@ static bool testSetuid(const char *filename)
 
 void testStaticallyLinked(const char *pathname)
 {
-  if (dmtcp::Util::isStaticallyLinked(pathname)) {
+  if (Util::isStaticallyLinked(pathname)) {
     JASSERT_STDERR <<
       "*** WARNING:  " ELF_INTERPRETER " --verify " << pathname << " returns\n"
       << "***  nonzero status.\n"
@@ -606,9 +712,9 @@ void testStaticallyLinked(const char *pathname)
 // Test for 'screen' program, argvPtr is an in- and out- parameter
 static bool testScreen(char **argv, char ***newArgv)
 {
-  if (dmtcp::Util::isScreen(argv[0])) {
-    dmtcp::Util::setScreenDir();
-    dmtcp::Util::patchArgvIfSetuid(argv[0], argv, newArgv);
+  if (Util::isScreen(argv[0])) {
+    Util::setScreenDir();
+    Util::patchArgvIfSetuid(argv[0], argv, newArgv);
     return true;
   }
   return false;
@@ -618,19 +724,26 @@ static void setLDPreloadLibs(bool is32bitElf)
 {
   // preloadLibs are to set LD_PRELOAD:
   //   LD_PRELOAD=PLUGIN_LIBS:UTILITY_DIR/libdmtcp.so:R_LIBSR_UTILITY_DIR/
-  dmtcp::string preloadLibs = "";
+  string preloadLibs = "";
+  string preloadLibs32 = "";
   // FIXME:  If the colon-separated elements of ENV_VAR_PLUGIN are not
   //     absolute pathnames, then they must be expanded to absolute pathnames.
   //     Warn user if an absolute pathname is not valid.
   if ( getenv(ENV_VAR_PLUGIN) != NULL ) {
     preloadLibs += getenv(ENV_VAR_PLUGIN);
     preloadLibs += ":";
+    preloadLibs32 = preloadLibs;
+  }
+  if ( getenv(ENV_VAR_PLUGIN_32) != NULL ) {
+    preloadLibs32 = getenv(ENV_VAR_PLUGIN_32);
+    preloadLibs32 += ":";
   }
-  dmtcp::string preloadLibs32 = preloadLibs;
 
-  // FindHelperUtiltiy requires ENV_VAR_UTILITY_DIR to be set
-  dmtcp::string searchDir = jalib::Filesystem::GetProgramDir();
-  setenv ( ENV_VAR_UTILITY_DIR, searchDir.c_str(), 0 );
+  // In case user has only given --with-plugin-32, then set
+  // preloadLibs same as preloadLibs32 and vice-versa
+  if(preloadLibs == "") {
+    preloadLibs = preloadLibs32;
+  }
 
   //set up Alloc plugin
   if (getenv(ENV_VAR_ALLOC_PLUGIN) != NULL){
@@ -659,46 +772,47 @@ static void setLDPreloadLibs(bool is32bitElf)
   }
 
   if (disableAllPlugins) {
-    preloadLibs = jalib::Filesystem::FindHelperUtility("libdmtcp.so");
-#if defined(__x86_64__)
-    preloadLibs32 = jalib::Filesystem::FindHelperUtility("libdmtcp.so", true);
+    preloadLibs = Util::getPath("libdmtcp.so");
+#if defined(__x86_64__) || defined(__aarch64__)
+    preloadLibs32 = Util::getPath("libdmtcp.so", true);
 #endif
   } else {
     for (size_t i = 0; i < numLibs; i++) {
       struct PluginInfo *p= &pluginInfo[i];
       if (*p->enabled) {
-        preloadLibs += jalib::Filesystem::FindHelperUtility(p->lib) + ":";
-#if defined(__x86_64__)
-        preloadLibs32 +=
-          jalib::Filesystem::FindHelperUtility(p->lib, true) + ":";
+        preloadLibs += Util::getPath(p->lib) + ":";
+#if defined(__x86_64__) || defined(__aarch64__)
+        preloadLibs32 += Util::getPath(p->lib, true) + ":";
 #endif
       }
     }
   }
 
   setenv(ENV_VAR_HIJACK_LIBS, preloadLibs.c_str(), 1);
-#if defined(__x86_64__)
+#if defined(__x86_64__) || defined(__aarch64__)
   setenv(ENV_VAR_HIJACK_LIBS_M32, preloadLibs32.c_str(), 1);
 #endif
 
   // If dmtcp_launch was called with user LD_PRELOAD, and if
-  //   if dmtcp_launch survived the experience, then pass it back to user.
+  //   dmtcp_launch survived the experience, then pass it back to user.
   if (getenv("LD_PRELOAD")) {
     setenv(ENV_VAR_ORIG_LD_PRELOAD, getenv("LD_PRELOAD"), 1);
     preloadLibs = preloadLibs + ":" + getenv("LD_PRELOAD");
-#if defined(__x86_64__)
+#if defined(__x86_64__) || defined(__aarch64__)
     preloadLibs32 = preloadLibs32 + ":" + getenv("LD_PRELOAD");
 #endif
   }
 
   setenv("LD_PRELOAD", preloadLibs.c_str(), 1);
-#if defined(__x86_64__)
+#if defined(__x86_64__) || defined(__aarch64__)
   if (is32bitElf) {
-    string libdmtcp = jalib::Filesystem::FindHelperUtility("libdmtcp.so", true);
-    JWARNING(libdmtcp != "libdmtcp.so")
+    string libdmtcp = Util::getPath("libdmtcp.so", true);
+    JWARNING(libdmtcp != "libdmtcp.so") (libdmtcp)
       .Text("You appear to be checkpointing a 32-bit target under 64-bit Linux.\n"
             "DMTCP was unable to find the 32-bit installation.\n"
-            "Try configure --enable-m32 ; make clean ; make ; make install");
+            "See DMTCP FAQ or try:\n"
+            "  ./configure --enable-m32 && make clean && make -j && make install\n"
+            "  ./configure && make clean && make -j && make install\n");
     setenv("LD_PRELOAD", preloadLibs32.c_str(), 1);
   }
 #endif
