@@ -20,12 +20,21 @@
  ****************************************************************************/
 
 #include <sys/syscall.h>
+#ifdef __aarch64__
+# define __ARCH_WANT_SYSCALL_DEPRECATED
+// SYS_fork is a deprecated kernel call in aarch64; in favor of SYS_clone?
+# include <asm-generic/unistd.h>
+// SYS_fork undefined in aarch64, but add extra insurance
+# undef SYS_fork
+# define SYS_fork __NR_fork
+#endif
 #include "constants.h"
 #include "uniquepid.h"
 #include "dmtcpworker.h"
 #include "processinfo.h"
 #include "syscallwrappers.h"
 #include "syslogwrappers.h"
+#include "threadlist.h"
 #include "util.h"
 #include "coordinatorapi.h"
 #include "mtcpinterface.h"
@@ -39,22 +48,22 @@
 
 using namespace dmtcp;
 
-#ifdef DEBUG
-  const static bool dbg = true;
-#else
-  const static bool dbg = false;
-#endif
+#ifdef LOGGING
+const static bool dbg = true;
+#else // ifdef LOGGING
+const static bool dbg = false;
+#endif // ifdef LOGGING
 
 static bool pthread_atfork_enabled = false;
 static uint64_t child_time;
-static dmtcp::CoordinatorAPI coordinatorAPI;
+static CoordinatorAPI coordinatorAPI;
 
 // Allow plugins to call fork/exec/system to perform specific tasks during
 // preCKpt/postCkpt/PostRestart etc. event.
 static bool isPerformingCkptRestart()
 {
-  if (dmtcp::WorkerState::currentState() != dmtcp::WorkerState::UNKNOWN &&
-      dmtcp::WorkerState::currentState() != dmtcp::WorkerState::RUNNING) {
+  if (WorkerState::currentState() != WorkerState::UNKNOWN &&
+      WorkerState::currentState() != WorkerState::RUNNING) {
     return true;
   }
   return false;
@@ -62,7 +71,7 @@ static bool isPerformingCkptRestart()
 
 static bool isBlacklistedProgram(const char *path)
 {
-  dmtcp::string programName = jalib::Filesystem::BaseName(path);
+  string programName = jalib::Filesystem::BaseName(path);
 
   JASSERT(programName != "dmtcp_coordinator" &&
           programName != "dmtcp_launch"  &&
@@ -89,7 +98,7 @@ static bool isBlacklistedProgram(const char *path)
   }
 
   if (programName == "dmtcp_nocheckpoint" || programName == "dmtcp_command" ||
-      programName == "ssh") {
+      programName == "ssh" || programName == "rsh" ) {
     return true;
   }
   return false;
@@ -134,28 +143,31 @@ LIB_PRIVATE void pthread_atfork_child()
   }
   pthread_atfork_enabled = false;
 
-  uint64_t host = dmtcp::UniquePid::ThisProcess().hostid();
-  dmtcp::UniquePid parent = dmtcp::UniquePid::ThisProcess();
-  dmtcp::UniquePid child = dmtcp::UniquePid(host, getpid(), child_time);
-  dmtcp::string child_name = jalib::Filesystem::GetProgramName() + "_(forked)";
-  JALIB_RESET_ON_FORK();
+  uint64_t host = UniquePid::ThisProcess().hostid();
+  UniquePid parent = UniquePid::ThisProcess();
+  UniquePid child = UniquePid(host, getpid(), child_time);
+  string child_name = jalib::Filesystem::GetProgramName() + "_(forked)";
   _dmtcp_remutex_on_fork();
-  dmtcp::SyslogCheckpointer::resetOnFork();
-  dmtcp::ThreadSync::resetLocks();
+  ThreadSync::resetLocks();
 
-  dmtcp::UniquePid::resetOnFork(child);
-  dmtcp::Util::initializeLogFile(child_name);
+  UniquePid::resetOnFork(child);
+  Util::initializeLogFile(child_name);
 
-  dmtcp::ProcessInfo::instance().resetOnFork();
+  ProcessInfo::instance().resetOnFork();
 
-  JTRACE("fork()ed [CHILD]") (child) (parent);
-  dmtcp::CoordinatorAPI::resetOnFork(coordinatorAPI);
-  dmtcp::DmtcpWorker::resetOnFork();
+  JLOG(DMTCP)("fork()ed [CHILD]") (child) (parent);
+  CoordinatorAPI::resetOnFork(coordinatorAPI);
+  DmtcpWorker::resetOnFork();
 }
 
 extern "C" pid_t fork()
 {
-  if (isPerformingCkptRestart()) {
+  if (isPerformingCkptRestart() ||
+      /*
+       * We don't want any user program between post checkpointing and before
+       * resuming to run under dmtcp
+       */
+      (dmtcp_delay_resume_blocked != NULL && dmtcp_delay_resume_blocked())) {
     return _real_syscall(SYS_fork);
   }
 
@@ -163,14 +175,14 @@ extern "C" pid_t fork()
    * processing this system call.
    */
   WRAPPER_EXECUTION_GET_EXCL_LOCK();
-  dmtcp::DmtcpWorker::eventHook(DMTCP_EVENT_ATFORK_PREPARE, NULL);
+  DmtcpWorker::eventHook(DMTCP_EVENT_ATFORK_PREPARE, NULL);
 
   /* Little bit cheating here: child_time should be same for both parent and
    * child, thus we compute it before forking the child. */
   child_time = time(NULL);
-  uint64_t host = dmtcp::UniquePid::ThisProcess().hostid();
-  dmtcp::UniquePid parent = dmtcp::UniquePid::ThisProcess();
-  dmtcp::string child_name = jalib::Filesystem::GetProgramName() + "_(forked)";
+  uint64_t host = UniquePid::ThisProcess().hostid();
+  UniquePid parent = UniquePid::ThisProcess();
+  string child_name = jalib::Filesystem::GetProgramName() + "_(forked)";
 
   coordinatorAPI.createNewConnectionBeforeFork(child_name);
 
@@ -180,6 +192,16 @@ extern "C" pid_t fork()
 
   if (childPid == -1) {
   } else if (childPid == 0) { /* child process */
+    // ThreadList::resetOnFork calls pthread_create which in turn calls
+    // malloc/calloc, etc. This can result in a deadlock if the parent process
+    // was holding malloc-lock while forking and might reset the lock only
+    // during atfork_child handler. Because our own atfork_child handler is
+    // called at the very beginning, the parent process won't have a chance to
+    // reset the lock. Calling ThreadList::resetOnFork here ensures that any
+    // such locks would have been reset by the caller and hence it's safe to
+    // call pthread_creat at this point.
+    ThreadList::resetOnFork();
+
     /* NOTE: Any work that needs to be done for the newly created child
      * should be put into pthread_atfork_child() function. That function is
      * hooked to the libc:fork() and will be called right after the new
@@ -189,21 +211,21 @@ extern "C" pid_t fork()
      * within the DmtcpWorker constructor to make sure that this is the first
      * registered handle.
      */
-    dmtcp::UniquePid child = dmtcp::UniquePid(host, getpid(), child_time);
-    JTRACE("fork() done [CHILD]") (child) (parent);
+    UniquePid child = UniquePid(host, getpid(), child_time);
+    JLOG(DMTCP)("fork() done [CHILD]") (child) (parent);
 
-    dmtcp::initializeMtcpEngine();
+    initializeMtcpEngine();
   } else if (childPid > 0) { /* Parent Process */
-    dmtcp::UniquePid child = dmtcp::UniquePid(host, childPid, child_time);
-    dmtcp::ProcessInfo::instance().insertChild(childPid, child);
-    JTRACE("fork()ed [PARENT] done") (child);;
+    UniquePid child = UniquePid(host, childPid, child_time);
+    ProcessInfo::instance().insertChild(childPid, child);
+    JLOG(DMTCP)("fork()ed [PARENT] done") (child);;
   }
 
   pthread_atfork_enabled = false;
 
   if (childPid != 0) {
     coordinatorAPI.closeConnection();
-    dmtcp::DmtcpWorker::eventHook(DMTCP_EVENT_ATFORK_PARENT, NULL);
+    DmtcpWorker::eventHook(DMTCP_EVENT_ATFORK_PARENT, NULL);
     WRAPPER_EXECUTION_RELEASE_EXCL_LOCK();
   }
   return childPid;
@@ -211,7 +233,7 @@ extern "C" pid_t fork()
 
 extern "C" pid_t vfork()
 {
-  JTRACE("vfork wrapper calling fork");
+  JLOG(DMTCP)("vfork wrapper calling fork");
   // This might not preserve the full semantics of vfork.
   // Used for checkpointing gdb.
   return fork();
@@ -231,13 +253,14 @@ static void execShortLivedProcessAndExit(const char *path, char *const argv[])
   if (argv[0] == NULL) {
     output = _real_popen(path, "r");
   } else {
-    dmtcp::string command = path;
+    string command = path;
     for (int i = 1; argv[i] != NULL; i++)
       command = command + " " + argv[i];
     output = _real_popen(command.c_str(), "r");
   }
   int numRead = fread(buf, 1, bufSize - 1, output);
   numRead++, numRead--; // suppress unused-var warning
+  buf[bufSize - 1] = '\0'; // NULL terminate in case the last char is not null
 
   pclose(output); // /lib/libXXX process is now done; can checkpoint now
   // FIXME:  code currently allows wrapper to proceed without lock if
@@ -257,34 +280,34 @@ static void execShortLivedProcessAndExit(const char *path, char *const argv[])
 static void dmtcpPrepareForExec(const char *path, char *const argv[],
                                 char **filename, char ***newArgv)
 {
-  JTRACE("Preparing for Exec") (path);
+  JLOG(DMTCP)("Preparing for Exec") (path);
 
   const char * libPrefix = "/lib/lib";
   const char * lib64Prefix = "/lib64/lib";
-  if (path != NULL && dmtcp::Util::strStartsWith(path, libPrefix))
+  if (path != NULL && Util::strStartsWith(path, libPrefix))
     execShortLivedProcessAndExit(path, argv);
-  if (path != NULL && dmtcp::Util::strStartsWith(path, lib64Prefix))
+  if (path != NULL && Util::strStartsWith(path, lib64Prefix))
     execShortLivedProcessAndExit(path, argv);
   // Needed for /usr/libexec/utempter/utempter and other short-lived
   //  setuid/setgid processes.
   // FIXME:  USE THIS FOR ALL setuid/setgid PROCESSES EXCEPT ONES THAT
   //         WE DIRECTLY HANDLE, LIKE 'screen'.  (Need to name special routine,
   //         execScreenProcess() ??)
-  if (path != NULL && dmtcp::Util::strEndsWith(path, "/utempter")) {
-    JTRACE("Trying to exec: utempter")(path)(argv[0])(argv[1]);
+  if (path != NULL && Util::strEndsWith(path, "/utempter")) {
+    JLOG(DMTCP)("Trying to exec: utempter")(path)(argv[0])(argv[1]);
     int oldIdx = -1;
     char *oldStr = NULL;
-    dmtcp::string realPtsNameStr;
+    string realPtsNameStr;
     // utempter takes a pts slave name as an argument. Since we virtualize
     // ptys, the slave name points to a virtual slave name, thus we need to
     // replace it with the real one.
     for (size_t i = 0; argv[i] != NULL; i++) {
-      if (dmtcp::Util::strStartsWith(argv[i], VIRT_PTS_PREFIX_STR)) {
+      if (Util::strStartsWith(argv[i], VIRT_PTS_PREFIX_STR)) {
         // FIXME: Potential memory leak if exec() fails.
         char *realPtsNameStr = (char*)JALLOC_HELPER_MALLOC(PTS_PATH_MAX);
         oldStr = argv[i];
         oldIdx = i;
-        dmtcp::SharedData::getRealPtyName(argv[i], realPtsNameStr,
+        SharedData::getRealPtyName(argv[i], realPtsNameStr,
                                           PTS_PATH_MAX);
         // Override const restriction
         *(const char**)&argv[i] = realPtsNameStr;
@@ -298,13 +321,13 @@ static void dmtcpPrepareForExec(const char *path, char *const argv[],
   }
 
   // FIXME:  SEE COMMENTS IN dmtcp_launch.cpp, rev. 1087; AND CHANGE THIS.
-  if (dmtcp::Util::isSetuid(path)) {
-    if (dmtcp::Util::isScreen(path)) {
-      dmtcp::Util::setScreenDir();
+  if (path != NULL && Util::isSetuid(path)) {
+    if (Util::isScreen(path)) {
+      Util::setScreenDir();
     }
     // THIS NEXT LINE IS DANGEROUS.  MOST setuid PROGRAMS CAN'T RUN UNPRIVILEGED
-    dmtcp::Util::patchArgvIfSetuid(path, argv, newArgv);
-    // BUG:  dmtcp::Util::patchArgvIfSetuid() DOES NOT SET newArgv WHEN COPYING
+    Util::patchArgvIfSetuid(path, argv, newArgv);
+    // BUG:  Util::patchArgvIfSetuid() DOES NOT SET newArgv WHEN COPYING
     //   BINARY IN CODE RE-FACTORING FROM REVISION 911.
     *filename = (*newArgv)[0];
   } else {
@@ -322,37 +345,33 @@ static void dmtcpPrepareForExec(const char *path, char *const argv[],
   JASSERT(unlink(buf) == 0) (JASSERT_ERRNO);
   Util::changeFd(fd, PROTECTED_LIFEBOAT_FD);
   jalib::JBinarySerializeWriterRaw wr ("", PROTECTED_LIFEBOAT_FD);
-  dmtcp::UniquePid::serialize (wr);
+  UniquePid::serialize (wr);
   DmtcpEventData_t edata;
   edata.serializerInfo.fd = PROTECTED_LIFEBOAT_FD;
-  dmtcp::DmtcpWorker::eventHook(DMTCP_EVENT_PRE_EXEC, &edata);
+  DmtcpWorker::eventHook(DMTCP_EVENT_PRE_EXEC, &edata);
 
-  JTRACE("Will exec filename instead of path") (path) (*filename);
+  JLOG(DMTCP)("Will exec filename instead of path") (path) (*filename);
 
-  dmtcp::Util::adjustRlimitStack();
-
-  char str[21] = {0};
-  sprintf(str, "%d", SharedData::getDlsymOffset());
-  setenv(ENV_VAR_DLSYM_OFFSET, str, 1);
-  sprintf(str, "%d", SharedData::getDlsymOffset_m32());
-  setenv(ENV_VAR_DLSYM_OFFSET_M32, str, 1);
+  Util::adjustRlimitStack();
+  Util::prepareDlsymWrapper();
+  Util::setProtectedFdBase();
 
   // Remove FD_CLOEXEC flag from protected file descriptors.
-  for (size_t i  = PROTECTED_FD_START; i < PROTECTED_FD_END; i++) {
+  for (int i  = PROTECTED_FD_START; i < PROTECTED_FD_END; i++) {
     int flags = fcntl(i, F_GETFD, NULL);
     if (flags != -1) {
       fcntl(i, F_SETFD, flags & ~FD_CLOEXEC);
     }
   }
-  JTRACE("Prepared for Exec") (getenv("LD_PRELOAD"));
+  JLOG(DMTCP)("Prepared for Exec") (getenv("LD_PRELOAD"));
 }
 
 static void dmtcpProcessFailedExec(const char *path, char *newArgv[])
 {
   int saved_errno = errno;
 
-  if (dmtcp::Util::isSetuid(path)) {
-    dmtcp::Util::freePatchedArgv(newArgv);
+  if (Util::isSetuid(path)) {
+    Util::freePatchedArgv(newArgv);
   }
 
   restoreUserLDPRELOAD();
@@ -360,31 +379,69 @@ static void dmtcpProcessFailedExec(const char *path, char *newArgv[])
   unsetenv(ENV_VAR_DLSYM_OFFSET);
   unsetenv(ENV_VAR_DLSYM_OFFSET_M32);
 
-  JTRACE("Processed failed Exec Attempt") (path) (getenv("LD_PRELOAD"));
+  JLOG(DMTCP)("Processed failed Exec Attempt") (path) (getenv("LD_PRELOAD"));
   errno = saved_errno;
   JASSERT(_real_close(PROTECTED_LIFEBOAT_FD) == 0) (JASSERT_ERRNO);
 }
 
-static dmtcp::string getUpdatedLdPreload(const char* filename,
-                                         const char* currLdPreload = NULL)
+static string getUpdatedLdPreload(const char* filename,
+                                  const char* currLdPreload)
 {
-  dmtcp::string preload = getenv(ENV_VAR_HIJACK_LIBS);
-  bool isElf, is32bitElf;
-  if  (dmtcp::Util::elfType(filename, &isElf, &is32bitElf) != -1) {
-    if (isElf && is32bitElf && getenv(ENV_VAR_HIJACK_LIBS_M32) != NULL) {
-      preload = getenv(ENV_VAR_HIJACK_LIBS_M32);
+  string preload = getenv(ENV_VAR_HIJACK_LIBS);
+
+  bool isElf = false;
+  bool is32bitElf = false;
+  if (getenv(ENV_VAR_HIJACK_LIBS_M32) != NULL &&
+      Util::elfType(filename, &isElf, &is32bitElf) != -1 &&
+      isElf &&
+      is32bitElf) {
+    preload = getenv(ENV_VAR_HIJACK_LIBS_M32);
+  }
+
+  vector<string> pluginLibraries = Util::tokenizeString(preload, ":");
+  for (size_t i = 0; i < pluginLibraries.size(); i++) {
+    // If the plugin doesn't exist, try to search it in the current install
+    // directory.
+    if (!jalib::Filesystem::FileExists(pluginLibraries[i])) {
+      pluginLibraries[i] =
+        Util::getPath(jalib::Filesystem::BaseName(pluginLibraries[i]),
+                                                  is32bitElf);
     }
   }
 
   const char *preloadEnv = getenv("LD_PRELOAD");
   if (currLdPreload != NULL && strlen(currLdPreload) > 0) {
+    pluginLibraries.push_back(currLdPreload);
     setenv(ENV_VAR_ORIG_LD_PRELOAD, currLdPreload, 1);
-    preload = preload + ":" + currLdPreload;
   } else if (preloadEnv != NULL && strlen(preloadEnv) > 0) {
-    setenv(ENV_VAR_ORIG_LD_PRELOAD, getenv("LD_PRELOAD"), 1);
-    preload = preload + ":" + getenv("LD_PRELOAD");
+    pluginLibraries.push_back(preloadEnv);
+    setenv(ENV_VAR_ORIG_LD_PRELOAD, preloadEnv, 1);
   }
-  return preload;
+
+  string newPreload = Util::joinStrings(pluginLibraries, ":");
+  return newPreload;
+}
+
+static vector<string> copyEnv(char *const envp[])
+{
+  vector<string> result;
+  if (envp) {
+    for (size_t i = 0; envp[i] != NULL; i++) {
+      result.push_back(envp[i]);
+    }
+  }
+  return result;
+}
+
+static vector<const char*> stringVectorToPointerArray(const vector<string>& s)
+{
+  vector<const char*> result;
+  // Now get the pointers.
+  for (size_t i = 0; i < s.size(); i++) {
+    result.push_back(s[i].c_str());
+  }
+  result.push_back(NULL);
+  return result;
 }
 
 static const char* ourImportantEnvs[] =
@@ -393,7 +450,7 @@ static const char* ourImportantEnvs[] =
 };
 #define ourImportantEnvsCnt ((sizeof(ourImportantEnvs))/(sizeof(const char*)))
 
-static bool isImportantEnv (dmtcp::string str)
+static bool isImportantEnv (string str)
 {
   str = str.substr(0, str.find("="));
 
@@ -404,124 +461,98 @@ static bool isImportantEnv (dmtcp::string str)
   return false;
 }
 
-static dmtcp::vector<dmtcp::string> copyUserEnv (char *const envp[])
+static vector<string> patchUserEnv (vector<string> env, const char* filename)
 {
-  dmtcp::vector<dmtcp::string> strStorage;
+  vector<string> result;
+  string userPreloadStr;
 
-  dmtcp::ostringstream out;
+  ostringstream out;
   out << "non-DMTCP env vars:\n";
-  for (; *envp != NULL; ++envp) {
-    if (isImportantEnv (*envp)) {
+
+  for (size_t i = 0; i < env.size(); i++) {
+    if (isImportantEnv (env[i])) {
       if (dbg) {
-        out << "     skipping: " << *envp << '\n';
+        out << "     skipping: " << env[i] << '\n';
       }
       continue;
     }
-    dmtcp::string e(*envp);
-    strStorage.push_back (e);
-    if (dbg) {
-      out << "     addenv[user]:" << strStorage.back() << '\n';
-    }
-  }
-  JTRACE("Creating a copy of (non-DMTCP) user env vars...") (out.str());
-
-  return strStorage;
-}
-
-static
-dmtcp::vector<const char*> patchUserEnv (dmtcp::vector<dmtcp::string> &envp,
-                                         const char* filename)
-{
-  dmtcp::vector<const char*> envVect;
-  const char *userPreloadStr = NULL;
-  envVect.clear();
-  JASSERT(envVect.size() == 0);
-
-  dmtcp::ostringstream out;
-  out << "non-DMTCP env vars:\n";
-
-  for (size_t i = 0; i < envp.size(); i++) {
-    if (isImportantEnv (envp[i].c_str())) {
-      if (dbg) {
-        out << "     skipping: " << envp[i] << '\n';
-      }
-      continue;
-    }
-    if (dmtcp::Util::strStartsWith(envp[i], "LD_PRELOAD=")) {
-      userPreloadStr = envp[i].c_str() + strlen("LD_PRELOAD=");
+    if (Util::strStartsWith(env[i], "LD_PRELOAD=")) {
+      userPreloadStr = env[i].substr(strlen("LD_PRELOAD="));
       continue;
     }
 
-    envVect.push_back (envp[i].c_str());
+    result.push_back(env[i]);
     if (dbg) {
-      out << "     addenv[user]:" << envVect.back() << '\n';
+      out << "     addenv[user]:" << result.back() << '\n';
     }
   }
-  JTRACE("Creating a copy of (non-DMTCP) user env vars...") (out.str());
+  JLOG(DMTCP)("Creating a copy of (non-DMTCP) user env vars...") (out.str());
 
   //pack up our ENV into the new ENV
   out.str("DMTCP env vars:\n");
   for (size_t i=0; i<ourImportantEnvsCnt; ++i) {
-    const char* v = getenv (ourImportantEnvs[i]);
-    dmtcp::string e = ourImportantEnvs[i];
-    if (strcmp(ourImportantEnvs[i], ENV_VAR_ORIG_LD_PRELOAD) == 0 &&
-        userPreloadStr != NULL && strlen(userPreloadStr) > 0) {
-      envp.push_back(e + "=" + userPreloadStr);
+    const char* v = getenv(ourImportantEnvs[i]);
+    const string e = ourImportantEnvs[i];
+    if (e == ENV_VAR_ORIG_LD_PRELOAD && !userPreloadStr.empty()) {
+      result.push_back(e + "=" + userPreloadStr);
     } else if (v != NULL) {
-      envp.push_back (e + '=' + v);
-      const char *ptr = envp.back().c_str();
-      JASSERT(ptr != NULL);
-      envVect.push_back(ptr);
+      result.push_back (e + '=' + v);
       if (dbg) {
-        out << "     addenv[dmtcp]:" << envVect.back() << '\n';
+        out << "     addenv[dmtcp]:" << result.back() << '\n';
       }
     }
   }
 
-  dmtcp::string ldPreloadStr = "LD_PRELOAD=";
-  ldPreloadStr += getUpdatedLdPreload(filename, userPreloadStr);
+  string ldPreloadStr = "LD_PRELOAD=";
+  ldPreloadStr += getUpdatedLdPreload(filename, userPreloadStr.c_str());
 
-  envp.push_back(ldPreloadStr);
-  envVect.push_back(envp.back().c_str());
-  if (dbg) {
-    out << "     addenv[dmtcp]:" << envVect.back() << '\n';
+  result.push_back(ldPreloadStr);
+
+  /*
+   * The path virtualization plugin tries to set the new prefix env var on
+   * restart. However, for some applications (e.g. bash), setenv() won't affect
+   * the envp area, possibly because the application has its private copy of
+   * the env vars. We need to patch the right env vars before calling execve().
+   * */
+  if (getenv("DMTCP_NEW_PATH_PREFIX")) {
+    string pathVirt = "DMTCP_NEW_PATH_PREFIX=";
+    pathVirt += getenv("DMTCP_NEW_PATH_PREFIX");
+    result.push_back(pathVirt);
   }
 
-  JTRACE("patching user envp...")  (out.str());
+  if (dbg) {
+    out << "     addenv[dmtcp]:" << result.back() << '\n';
+  }
 
-  envVect.push_back (NULL);
+  JLOG(DMTCP)("Patched user envp...")  (out.str());
 
-  JTRACE("Done patching environ");
-  return envVect;
+  return result;
 }
 
 extern "C" int execve (const char *filename, char *const argv[],
                         char *const envp[])
 {
 
-  if( strstr(filename,"srun") != NULL ){
-    JTRACE("execvp() wrapper (srun) - will call ecexvp directly") (filename);
-  }
-
-  if (isPerformingCkptRestart() || isBlacklistedProgram(filename) || (strstr(filename,"srun") != NULL) ) {
+  if (isPerformingCkptRestart() || isBlacklistedProgram(filename) ) {
     return _real_execve(filename, argv, envp);
   }
-  JTRACE("execve() wrapper") (filename);
+  JLOG(DMTCP)("execve() wrapper") (filename);
 
   /* Acquire the wrapperExeution lock to prevent checkpoint to happen while
    * processing this system call.
    */
   WRAPPER_EXECUTION_GET_EXCL_LOCK();
 
-  dmtcp::vector<dmtcp::string> origUserEnv = copyUserEnv(envp);
+  const vector<string> env = copyEnv(envp);
 
   char *newFilename;
   char **newArgv;
   dmtcpPrepareForExec(filename, argv, &newFilename, &newArgv);
 
-  dmtcp::vector<const char*> envVect = patchUserEnv(origUserEnv, filename);
+  const vector<string> envStrings = patchUserEnv(env, filename);
+  const vector<const char*> newEnv = stringVectorToPointerArray(envStrings);
 
-  int retVal = _real_execve (newFilename, newArgv, (char* const*)&envVect[0]);
+  int retVal = _real_execve (newFilename, newArgv, (char* const*)&newEnv[0]);
 
   dmtcpProcessFailedExec(filename, newArgv);
 
@@ -532,20 +563,22 @@ extern "C" int execve (const char *filename, char *const argv[],
 
 extern "C" int execv (const char *path, char *const argv[])
 {
-  JTRACE("execv() wrapper, calling execve with environ") (path);
-  return execve(path, argv, environ);
+  JLOG(DMTCP)("execv() wrapper, calling execve with environ") (path);
+
+  // Make a copy of the environ coz it might change after a setenv().
+  const vector<string> envStrings = copyEnv(environ);
+  // Now get the pointers.
+  const vector<const char*> env = stringVectorToPointerArray(envStrings);
+
+  return execve(path, argv, (char* const*) &env[0]);
 }
 
 extern "C" int execvp (const char *filename, char *const argv[])
 {
-  if( strstr(filename,"srun") != NULL ){
-    JTRACE("execvp() wrapper (srun) - will call ecexvp directly") (filename);
-  }
-
-  if (isPerformingCkptRestart() || isBlacklistedProgram(filename) || (strstr(filename,"srun") != NULL) ) {
+  if (isPerformingCkptRestart() || isBlacklistedProgram(filename) ) {
     return _real_execvp(filename, argv);
   }
-  JTRACE("execvp() wrapper") (filename);
+  JLOG(DMTCP)("execvp() wrapper") (filename);
   /* Acquire the wrapperExeution lock to prevent checkpoint to happen while
    * processing this system call.
    */
@@ -554,7 +587,7 @@ extern "C" int execvp (const char *filename, char *const argv[])
   char *newFilename;
   char **newArgv;
   dmtcpPrepareForExec(filename, argv, &newFilename, &newArgv);
-  setenv("LD_PRELOAD", getUpdatedLdPreload(filename).c_str(), 1);
+  setenv("LD_PRELOAD", getUpdatedLdPreload(filename, NULL).c_str(), 1);
 
   int retVal = _real_execvp (newFilename, newArgv);
 
@@ -572,21 +605,23 @@ extern "C" int execvpe (const char *filename, char *const argv[],
   if (isPerformingCkptRestart() || isBlacklistedProgram(filename)) {
     return _real_execvpe(filename, argv, envp);
   }
-  JTRACE("execvpe() wrapper") (filename);
+  JLOG(DMTCP)("execvpe() wrapper") (filename);
   /* Acquire the wrapperExeution lock to prevent checkpoint to happen while
    * processing this system call.
    */
   WRAPPER_EXECUTION_GET_EXCL_LOCK();
 
-  dmtcp::vector<dmtcp::string> origUserEnv = copyUserEnv(envp);
+  // Make a copy of the environ coz it might change after a setenv().
+  const vector<string> env = copyEnv(envp);
 
   char *newFilename;
   char **newArgv;
   dmtcpPrepareForExec(filename, argv, &newFilename, &newArgv);
 
-  dmtcp::vector<const char*> envVect = patchUserEnv(origUserEnv, filename);
+  const vector<string> newEnvStrings = patchUserEnv(env, filename);
+  const vector<const char*> newEnv = stringVectorToPointerArray(newEnvStrings);
 
-  int retVal = _real_execvpe(newFilename, newArgv, (char* const*)&envVect[0]);
+  int retVal = _real_execvpe(newFilename, newArgv, (char* const*)&newEnv[0]);
 
   dmtcpProcessFailedExec(filename, newArgv);
 
@@ -600,14 +635,14 @@ extern "C" int fexecve (int fd, char *const argv[], char *const envp[])
   char buf[sizeof "/proc/self/fd/" + sizeof (int) * 3];
   snprintf (buf, sizeof (buf), "/proc/self/fd/%d", fd);
 
-  JTRACE("fexecve() wrapper calling execve()") (fd) (buf);
+  JLOG(DMTCP)("fexecve() wrapper calling execve()") (fd) (buf);
   return execve(buf, argv, envp);
 }
 
 
 extern "C" int execl (const char *path, const char *arg, ...)
 {
-  JTRACE("execl() wrapper") (path);
+  JLOG(DMTCP)("execl() wrapper") (path);
 
   size_t argv_max = INITIAL_ARGV_MAX;
   const char *initial_argv[INITIAL_ARGV_MAX];
@@ -629,6 +664,7 @@ extern "C" int execl (const char *path, const char *arg, ...)
       {
         if (argv != initial_argv)
           free (argv);
+        va_end (args);
         return -1;
       }
       if (argv == initial_argv)
@@ -652,7 +688,7 @@ extern "C" int execl (const char *path, const char *arg, ...)
 
 extern "C" int execlp (const char *file, const char *arg, ...)
 {
-  JTRACE("execlp() wrapper") (file);
+  JLOG(DMTCP)("execlp() wrapper") (file);
 
   size_t argv_max = INITIAL_ARGV_MAX;
   const char *initial_argv[INITIAL_ARGV_MAX];
@@ -674,6 +710,7 @@ extern "C" int execlp (const char *file, const char *arg, ...)
       {
         if (argv != initial_argv)
           free (argv);
+        va_end (args);
         return -1;
       }
       if (argv == initial_argv)
@@ -697,7 +734,7 @@ extern "C" int execlp (const char *file, const char *arg, ...)
 
 extern "C" int execle(const char *path, const char *arg, ...)
 {
-  JTRACE("execle() wrapper") (path);
+  JLOG(DMTCP)("execle() wrapper") (path);
 
   size_t argv_max = INITIAL_ARGV_MAX;
   const char *initial_argv[INITIAL_ARGV_MAX];
@@ -718,6 +755,7 @@ extern "C" int execle(const char *path, const char *arg, ...)
       {
         if (argv != initial_argv)
           free (argv);
+        va_end (args);
         return -1;
       }
       if (argv == initial_argv)
@@ -745,7 +783,7 @@ extern int do_system (const char *line);
 
 extern "C" int system (const char *line)
 {
-  JTRACE("before system(), checkpointing may not work")
+  JLOG(DMTCP)("before system(), checkpointing may not work")
     (line) (getenv (ENV_VAR_HIJACK_LIBS)) (getenv ("LD_PRELOAD"));
 
   if (line == NULL)
@@ -755,7 +793,7 @@ extern "C" int system (const char *line)
 
   int result = do_system (line);
 
-  JTRACE("after system()");
+  JLOG(DMTCP)("after system()");
 
   return result;
 }
