@@ -29,10 +29,13 @@
 #include "mtcp_sys.h"
 #include "restore_libc.h"
 #include "tlsutil.h"
+#include "config.h" // define WSL if present
+#include <pthread.h> // for pthread_self(), needed for WSL
+                     //    or for alternate way to verify thread descriptor
 
 int mtcp_sys_errno;
 
-#ifdef __x86_64__
+#if defined(__x86_64__) || defined(__aarch64__)
 # define ELF_AUXV_T Elf64_auxv_t
 # define UINT_T uint64_t
 #else
@@ -43,6 +46,27 @@ int mtcp_sys_errno;
 
 extern MYINFO_GS_T myinfo_gs;
 
+static int glibcMajorVersion()
+{
+  static int major = 0;
+  if (major == 0) {
+    major = (int) strtol(gnu_get_libc_version(), NULL, 10);
+    ASSERT (major == 2);
+  }
+  return major;
+}
+
+static int glibcMinorVersion()
+{
+  static long minor = 0;
+  if (minor == 0) {
+    char *ptr;
+    int major = (int) strtol(gnu_get_libc_version(), &ptr, 10);
+    ASSERT (major == 2);
+    minor = (int) strtol(ptr+1, NULL, 10);
+  }
+  return minor;
+}
 
 /* Offset computed (&x.pid - &x) for
  *   struct pthread x;
@@ -89,25 +113,58 @@ extern MYINFO_GS_T myinfo_gs;
 //     of cached tid to clone, and MTCP grabs it; But MTCP is still missing
 //     the address where pthread cached the tid of motherofall.  So, it can't
 //     update.
+/*
+ * For those who want to dig deeper into glibc and its thread descriptor,
+ *   see below.  Note that pthread_self returns a pthread_t, which is a
+ *   pointer to the 'struct pthread' below.
+ * FROM: glibc-2.23/sysdeps/x86_64/nptl/tls.h
+ * // Return the thread descriptor for the current thread.
+ * # define THREAD_SELF \
+ *   ({ struct pthread *__self;                                                  \
+ *      asm ("mov %%fs:%c1,%0" : "=r" (__self)                                   \
+ *           : "i" (offsetof (struct pthread, header.self)));                    \
+ *      __self;})
+ * // struct pthread is defined in glibc-2.23/nptl/descr.h
+ * // type = struct pthread {
+ * //     union {
+ * //         tcbhead_t header;
+ * //         void *__padding[24];
+ * //     };
+ * //     list_t list; // 2*sizeof(void *)
+ * //     pid_t tid;
+ * //     pid_t pid;
+ * // NOTE: sizeof(tcbhead_t) + 2*sizeof(void *) == 720
+ * //       where:  sizeof(tcbhead_t) == 704
+ */
+
+/* NOTE:  For future reference, the STATIC_TLS_TID_OFFSET() for a glibc version
+ *  can be easily discvoered as long as a debug version of glibc is present:
+ *  (gdb) p (char *)&(((struct pthread *)pthread_self())->tid) - \
+ *                                                      (char *)pthread_self()
+ *  $14 = 720  # So, 720 is the correct offset in this example.
+ */
 static int STATIC_TLS_TID_OFFSET()
 {
   static int offset = -1;
   if (offset != -1)
     return offset;
 
-  char *ptr;
-  long major = strtol(gnu_get_libc_version(), &ptr, 10);
-  long minor = strtol(ptr+1, NULL, 10);
-  ASSERT (major == 2);
-
-  if (minor >= 11) {
 #ifdef __x86_64__
-    offset = 26*sizeof(void *) + 512;
-#else
-    offset = 26*sizeof(void *);
+  // NEEDED FOR WSL
+  if (glibcMinorVersion() >= 23) {
+    offset = 720;
+    return offset;
+  }
 #endif
-  } else if (minor == 10) {
-    offset = 26*sizeof(void *);
+
+  if (glibcMinorVersion() >= 11) {
+#ifdef __x86_64__
+    offset = 26 * sizeof(void *) + 512;
+#else /* ifdef __x86_64__ */
+    offset = 26 * sizeof(void *);
+#endif /* ifdef __x86_64__ */
+  } else if (glibcMinorVersion() == 10) {
+    offset = 26 * sizeof(void *);
   } else {
     offset = 18*sizeof(void *);
   }
@@ -147,7 +204,7 @@ static int STATIC_TLS_TID_OFFSET()
 /* Can remove the unused attribute when this __GLIBC_PREREQ is the only one. */
 static char *memsubarray (char *array, char *subarray, size_t len)
 					 __attribute__ ((unused));
-static int get_tls_segreg(void);
+static uint64_t get_tls_segreg(void);
 static void *get_tls_base_addr(void);
 extern void **motherofall_saved_sp;
 extern ThreadTLSInfo *motherofall_tlsInfo;
@@ -156,11 +213,11 @@ extern ThreadTLSInfo *motherofall_tlsInfo;
  *
  *****************************************************************************/
 int TLSInfo_GetTidOffset(void)
-
 {
   static int tid_offset = -1;
   if (tid_offset == -1) {
-    struct {pid_t tid; pid_t pid;} tid_pid;
+    struct { pid_t tid; pid_t pid; } tid_pid;
+
     /* struct pthread has adjacent fields, tid and pid, in that order.
      * Try to find at what offset that bit patttern occurs in struct pthread.
      */
@@ -176,7 +233,7 @@ int TLSInfo_GetTidOffset(void)
      * Then fill in the entry_number field of an empty 'struct user_desc', and
      * get_thread_area(struct user_desc *uinfo) will fill in the rest.
      * Then use the filled in base_address field to get the 'struct pthread'.
-     * The function tcp_get_tls_base_addr() returns this 'struct pthread' addr.
+     * The function get_tls_base_addr() returns this 'struct pthread' addr.
      */
     void * pthread_desc = get_tls_base_addr();
     /* A false hit for tid_offset probably can't happen since a new
@@ -184,6 +241,14 @@ int TLSInfo_GetTidOffset(void)
      * pthread_desc below is defined as 'struct pthread' in glibc:nptl/descr.h
      */
     tmp = memsubarray((char *)pthread_desc, (char *)&tid_pid, sizeof(tid_pid));
+
+    if (tmp == NULL && glibcMajorVersion() == 2 && glibcMinorVersion() >= 24) {
+      // starting with glibc-2.25 (including 2.24.90 on Fedora), the pid field
+      // is deprecated and set to zero.
+      tid_pid.pid = 0;
+      tmp = memsubarray((char *)pthread_desc, (char *)&tid_pid, sizeof(tid_pid));
+    }
+
     if (tmp == NULL) {
       PRINTF("WARNING: Couldn't find offsets of tid/pid in thread_area.\n"
              "  Now relying on the value determined using the\n"
@@ -200,7 +265,7 @@ int TLSInfo_GetTidOffset(void)
              "  Continuing anyway.  If this fails, please try again.",
              tid_offset);
     }
-    DPRINTF("tid_offset: %d\n", tid_offset);
+    DPRINTF("DEBUGGING INFO: tid_offset: %d\n", tid_offset);
     if (tid_offset % sizeof(int) != 0) {
       PRINTF("WARNING: tid_offset is not divisible by sizeof(int).\n"
              "  Now relying on the value determined using the\n"
@@ -226,7 +291,7 @@ int TLSInfo_GetPidOffset(void)
   if (pid_offset == -1) {
     int tid_offset = TLSInfo_GetTidOffset();
     pid_offset = tid_offset + (char *)&(tid_pid.pid) - (char *)&tid_pid;
-    DPRINTF("pid_offset: %d\n", pid_offset);
+    DPRINTF("DEBUGGING_INFO: pid_offset: %d\n", pid_offset);
   }
   return pid_offset;
 }
@@ -250,7 +315,7 @@ static char *memsubarray (char *array, char *subarray, size_t len)
    return NULL;
 }
 
-static int get_tls_segreg(void)
+static uint64_t get_tls_segreg(void)
 {
   segreg_t tlssegreg;
 #ifdef __i386__
@@ -261,20 +326,50 @@ static int get_tls_segreg(void)
 #elif __arm__
   asm volatile ("mrc     p15, 0, %0, c13, c0, 3  @ load_tp_hard\n\t"
                 : "=r" (tlssegreg));
+#elif defined(__aarch64__)
+  asm volatile ("mrs  %0, tpidr_el0" : "=r" (tlssegreg) );
 #endif
-  return (int)tlssegreg;
+  return (uint64_t)tlssegreg;
 }
 
+/*
+static int
+tls_get_thread_area(void *uinfo, MYINFO_GS_T dummy)
+{
+  asm volatile ("mrs   %0, tpidr_el0" : "=r" (myinfo_gs) );
+  myinfo_gs = myinfo_gs - 1776;
+  *(unsigned long int *)&(((struct user_desc *)uinfo)->base_addr) = myinfo_gs;
+   return myinfo_gs;
+}
+
+static int
+tls_set_thread_area(void *uinfo, MYINFO_GS_T dummy)
+{
+  myinfo_gs = *(unsigned long int *)&(((struct user_desc *)uinfo)->base_addr);
+  myinfo_gs = myinfo_gs + 1776;
+  asm volatile ("msr     tpidr_el0, %[gs]" : : [gs] "r" (myinfo_gs) );
+}
+*/
+
 static void* get_tls_base_addr()
 {
-  struct user_desc gdtentrytls;
+  struct user_desc gdtentrytls[2];
 
-  gdtentrytls.entry_number = get_tls_segreg() / 8;
-  if (tls_get_thread_area(&gdtentrytls, myinfo_gs) == -1) {
+  gdtentrytls[0].entry_number = get_tls_segreg() / 8;
+  if (tls_get_thread_area(&gdtentrytls[0], &gdtentrytls[1]) == -1) {
     PRINTF("Error getting GDT TLS entry: %d\n", errno);
     _exit(0);
   }
-  return (void *)(*(unsigned long *)&(gdtentrytls.base_addr));
+#ifdef WSL
+  // This could be used for regular non-WSL also, although it depends on
+  //   the GNU implementation.  pthread_self() returns thread area.
+  // NOTE:  The syscall arch_prctl(ARCH_GET_FS,...) discovers the base
+  //   address of the thread descriptor for non-WSL, but it returns
+  //   different values when invoked at different call sites in WSL.
+  //   Probably, this is because Windows is also using the $fs register.
+  return (void *)pthread_self();
+#endif
+  return (void *)(*(unsigned long *)&(gdtentrytls[0].base_addr));
 }
 
 // Returns value for AT_SYSINFO in kernel's auxv
@@ -371,6 +466,8 @@ int TLSInfo_HaveThreadSysinfoOffset()
 #elif defined(__arm__)
   asm volatile ("mrc     p15, 0, %0, c13, c0, 3  @ load_tp_hard\n\t"
                 : "=r" (sysinfo) );
+#elif defined(__aarch64__)
+  asm volatile ("mrs     %0, tpidr_el0" : "=r" (sysinfo) );
 #else
 # error "current architecture not supported"
 #endif
@@ -391,6 +488,8 @@ void *TLSInfo_GetThreadSysinfo()
 #elif defined(__arm__)
   asm volatile ("mrc     p15, 0, %0, c13, c0, 3  @ load_tp_hard\n\t"
                 : "=r" (sysinfo) );
+#elif defined(__aarch64__)
+  asm volatile ("mrs     %0, tpidr_el0" : "=r" (sysinfo) );
 #else
 # error "current architecture not supported"
 #endif
@@ -403,6 +502,8 @@ void TLSInfo_SetThreadSysinfo(void *sysinfo) {
                 : : "r" (sysinfo) );
 #elif defined(__arm__)
   mtcp_sys_kernel_set_tls(sysinfo);
+#elif defined(__aarch64__)
+  asm volatile ("msr     tpidr_el0, %[gs]" : : [gs] "r" (sysinfo) );
 #else
 # error "current architecture not supported"
 #endif
@@ -419,18 +520,29 @@ void TLSInfo_VerifyPidTid(pid_t pid, pid_t tid)
   tls_pid = *(pid_t *) (addr + TLSInfo_GetPidOffset());
   tls_tid = *(pid_t *) (addr + TLSInfo_GetTidOffset());
 
-  if ((tls_pid != pid) || (tls_tid != tid)) {
-    PRINTF("ERROR: getpid(%d), tls pid(%d), and tls tid(%d) must all match\n",
-           (int)mtcp_sys_getpid(), tls_pid, tls_tid);
+  if (tls_tid != tid) {
+    PRINTF("ERROR: tls tid(%d) doesn't match the thread tid (%d)\n",
+           tls_tid, tid);
+    _exit(0);
+  }
+
+  // For glibc > 2.24, pid field is unused. Here we do the <24 check to ensure
+  // that distros with glibc 2.24-NNN are covered as well.
+  if (glibcMajorVersion() == 2 && glibcMinorVersion() < 24 && tls_pid != pid) {
+    PRINTF("ERROR: tls pid (%d) doesn't match getpid (%d)\n",
+           tls_pid, (int)mtcp_sys_getpid());
     _exit(0);
   }
 }
 
 void TLSInfo_UpdatePid()
 {
-  pid_t  *tls_pid = (pid_t *) ((char*)get_tls_base_addr() +
-                               TLSInfo_GetPidOffset());
-  *tls_pid = mtcp_sys_getpid();
+  // For glibc > 2.24, pid field is unused.
+  if (glibcMajorVersion() == 2 && glibcMinorVersion() <= 24) {
+    pid_t *tls_pid =
+      (pid_t *)((char *)get_tls_base_addr() + TLSInfo_GetPidOffset());
+    *tls_pid = mtcp_sys_getpid();
+  }
 }
 
 /*****************************************************************************
@@ -449,7 +561,7 @@ void TLSInfo_SaveTLSState (ThreadTLSInfo *tlsInfo)
 #elif __x86_64__
   //asm volatile ("movl %%fs,%0" : "=m" (tlsInfo->fs));
   //asm volatile ("movl %%gs,%0" : "=m" (tlsInfo->gs));
-#elif __arm__
+#elif __arm__ || __aarch64__
   // Follow x86_64 for arm.
 #endif
 
@@ -461,7 +573,7 @@ void TLSInfo_SaveTLSState (ThreadTLSInfo *tlsInfo)
  */
   i = tlsInfo->TLSSEGREG / 8;
   tlsInfo->gdtentrytls[0].entry_number = i;
-  if (tls_get_thread_area (&(tlsInfo->gdtentrytls[0]), myinfo_gs) == -1) {
+  if (tls_get_thread_area (&(tlsInfo->gdtentrytls[0]), &(tlsInfo->gdtentrytls[1])) == -1) {
     PRINTF("Error saving GDT TLS entry: %d\n", errno);
     _exit(0);
   }
@@ -487,16 +599,21 @@ void TLSInfo_RestoreTLSState(ThreadTLSInfo *tlsInfo)
   /* Patch 'struct user_desc' (gdtentrytls) of glibc to contain the
    * the new pid and tid.
    */
-  *(pid_t *)(*(unsigned long *)&(tlsInfo->gdtentrytls[0].base_addr)
-             + TLSInfo_GetPidOffset()) = mtcp_sys_getpid();
+
+  if (glibcMajorVersion() == 2 && glibcMinorVersion() <= 24) {
+    *(pid_t *)(*(unsigned long *)&(tlsInfo->gdtentrytls[0].base_addr)
+               + TLSInfo_GetPidOffset()) = mtcp_sys_getpid();
+  }
+
   if (mtcp_sys_kernel_gettid() == mtcp_sys_getpid()) {
     *(pid_t *)(*(unsigned long *)&(tlsInfo->gdtentrytls[0].base_addr)
                + TLSInfo_GetTidOffset()) = mtcp_sys_getpid();
   }
 
-  /* Now pass this to the kernel, so it can adjust the segment descriptor.
+  /* Now pass this to the kernel, so it can adjust the segment descriptors:
+   *   tls_set_thread_areaa() uses arg1 for fs and arg2 for gs.
    * This will make different kernel calls according to the CPU architecture. */
-  if (tls_set_thread_area (&(tlsInfo->gdtentrytls[0]), myinfo_gs) != 0) {
+  if (tls_set_thread_area (&(tlsInfo->gdtentrytls[0]), &(tlsInfo->gdtentrytls[1])) != 0) {
     PRINTF("Error restoring GDT TLS entry: %d\n", errno);
     mtcp_abort();
   }
@@ -518,6 +635,8 @@ void TLSInfo_RestoreTLSState(ThreadTLSInfo *tlsInfo)
  */
 #elif __arm__
 /* ARM treats this same as x86_64 above. */
+#elif defined(__aarch64__)
+# warning "TODO: Implementation for ARM64?"
 #endif
 }
 
@@ -553,6 +672,8 @@ void TLSInfo_PostRestart()
            sizeof("shell ../../util/gdb-add-libdmtcp-symbol-file.py PID PC\n"));
 #if defined(__i386__) || defined(__x86_64__)
       asm volatile ("int3"); // Do breakpoint; send SIGTRAP, caught by gdb
+#elif defined(__aarch64__)
+      asm("brk 0");
 #else
       DPRINTF("IN GDB: interrupt (^C); add-symbol-file ...; (gdb) print x=0\n");
       { int x = 1; while (x); } // Stop execution for user to type command.
