@@ -19,6 +19,42 @@
  *  <http://www.gnu.org/licenses/>.                                         *
  ****************************************************************************/
 
+/*
+ * Ckpt policy for handling files and shared memory segments.
+ *
+ * TODO(kapil): Fill the holes in this policy.
+ *
+ * Regular File:
+ * - Ckpt file descriptor
+ * - Leader election
+ * - Ckpt file based on heuristics
+ * Unlinked File:
+ * - Ckpt file descriptor
+ * - Leader election
+ * - Ckpt file
+ *
+ * Shared-memory area with regular file:
+ * - TODO(kapil): Any file descriptor with the file? If yes, use that to
+ *   ckpt-file.
+ * - Open a file descriptor
+ * - Ckpt file based on heuristics
+ * - recreate file on restart
+ * - close fd on restart.
+ *
+ * Shared-memory area with unlinked file:
+ * + Ckpt:
+ *   - TODO(kapil): Any file descriptor pointing to the file? If yes, delegate
+ *     ckpt to the file descriptor.
+ *   - everyone saves the contents of the shared-area.
+ * + Restart
+ *   - File already exists: verify that the file is at least as large as
+ *     (area.offset+area.size).
+ *   - File doesn't exist: try to recreate the file; write content
+ * - on restart, everyone tries to recreate the file and write their data
+ *   (offset, length) to the file.
+ * - everyone tries to unlink the file in a subsequent barrier.
+ */
+
 // THESE INCLUDES ARE IN RANDOM ORDER.  LET'S CLEAN IT UP AFTER RELEASE. - Gene
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -30,6 +66,7 @@
 #include <mqueue.h>
 #include <stdint.h>
 #include <signal.h>
+#include "procselfmaps.h"
 #include "util.h"
 #include "shareddata.h"
 #include "jfilesystem.h"
@@ -43,11 +80,14 @@ using namespace dmtcp;
 
 void dmtcp_FileConnList_EventHook(DmtcpEvent_t event, DmtcpEventData_t *data)
 {
-  dmtcp::FileConnList::instance().eventHook(event, data);
+  FileConnList::instance().eventHook(event, data);
 }
 
-static dmtcp::vector<ProcMapsArea> shmAreas;
-static dmtcp::vector<dmtcp::FileConnection*> shmAreaConn;
+static vector<ProcMapsArea> shmAreas;
+static vector<ProcMapsArea> unlinkedShmAreas;
+static vector<ProcMapsArea> missingUnlinkedShmFiles;
+static vector<FileConnection*> shmAreaConn;
+static uint32_t virtPtyId = 0;
 
 void dmtcp_FileConn_ProcessFdEvent(int event, int arg1, int arg2)
 {
@@ -61,7 +101,7 @@ void dmtcp_FileConn_ProcessFdEvent(int event, int arg1, int arg2)
 }
 
 static FileConnList *fileConnList = NULL;
-dmtcp::FileConnList& dmtcp::FileConnList::instance()
+FileConnList& FileConnList::instance()
 {
   if (fileConnList == NULL) {
     fileConnList = new FileConnList();
@@ -69,7 +109,7 @@ dmtcp::FileConnList& dmtcp::FileConnList::instance()
   return *fileConnList;
 }
 
-void dmtcp::FileConnList::preLockSaveOptions()
+void FileConnList::preLockSaveOptions()
 {
   // Now create a list of all shared-memory areas.
   prepareShmList();
@@ -77,9 +117,10 @@ void dmtcp::FileConnList::preLockSaveOptions()
   ConnectionList::preLockSaveOptions();
 }
 
-void dmtcp::FileConnList::drain()
+void FileConnList::drain()
 {
   ConnectionList::drain();
+  virtPtyId = SharedData::getVirtualPtyId();
 
   vector<SharedData::InodeConnIdMap> inodeConnIdMaps;
   for (iterator i = begin(); i != end(); ++i) {
@@ -98,10 +139,42 @@ void dmtcp::FileConnList::drain()
   if (inodeConnIdMaps.size() > 0) {
     SharedData::insertInodeConnIdMaps(inodeConnIdMaps);
   }
+  dmtcp_update_max_required_fd(getMaxFd());
 }
 
-void dmtcp::FileConnList::postRestart()
+/*
+ * This function is called after preCkpt() for all FileConnection
+ * objects and is useful for writing out information about the
+ * open files saved by DMTCP.
+ */
+void FileConnList::preCkpt()
 {
+  ConnectionList::preCkpt();
+
+  string fdInfoFile = dmtcp_get_ckpt_files_subdir();
+  fdInfoFile += "/fd-info.txt";
+  int tmpfd = _real_open(fdInfoFile.c_str(),
+                         O_CREAT | O_WRONLY | O_TRUNC,
+			 0644);
+
+  for (iterator i = begin(); i != end(); ++i) {
+    Connection* con =  i->second;
+    if (con->hasLock() && con->conType() == Connection::FILE) {
+      FileConnection *fileCon = (FileConnection*) con;
+      if (fileCon->checkpointed() == true) {
+        string buf = jalib::Filesystem::BaseName(fileCon->savedFilePath()) +
+                      ":" + fileCon->filePath() + "\n";
+        JASSERT(Util::writeAll(tmpfd, buf.c_str(),
+	                       buf.length()) == (ssize_t)buf.length());
+      }
+    }
+  }
+  _real_close(tmpfd);
+}
+
+void FileConnList::postRestart()
+{
+  SharedData::setVirtualPtyId(virtPtyId);
   /* It is possible to have two different connection-ids for a pre-existing
    * CTTY in two or more different process trees. In this case, only one of the
    * several process trees would be able to acquire a lock on the underlying
@@ -122,10 +195,24 @@ void dmtcp::FileConnList::postRestart()
     }
   }
 
+  /* Try to map the file as is, if it already exists on the disk.
+   */
+  for (size_t i = 0; i < unlinkedShmAreas.size(); i++) {
+    if (jalib::Filesystem::FileExists(unlinkedShmAreas[i].name)) {
+      // TODO(kapil): Verify the file contents.
+      JWARNING(false) (unlinkedShmAreas[i].name)
+        .Text("File was unlinked at ckpt but is currently present on disk; "
+              "remove it and try again.");
+      restoreShmArea(unlinkedShmAreas[i]);
+    } else {
+      missingUnlinkedShmFiles.push_back(unlinkedShmAreas[i]);
+    }
+  }
+
   ConnectionList::postRestart();
 }
 
-void dmtcp::FileConnList::refill(bool isRestart)
+void FileConnList::refill(bool isRestart)
 {
   // Check comments in PtyConnection::preRefill()/refill()
   for (iterator i = begin(); i != end(); ++i) {
@@ -136,24 +223,43 @@ void dmtcp::FileConnList::refill(bool isRestart)
     }
   }
 
+  if (isRestart) {
+    // The backing file will be created as a result of restoreShmArea. We need
+    // to unlink all such files in the resume() call below.
+    for (size_t i = 0; i < missingUnlinkedShmFiles.size(); i++) {
+      recreateShmFileAndMap(missingUnlinkedShmFiles[i]);
+    }
+  }
+
   ConnectionList::refill(isRestart);
 }
 
-void dmtcp::FileConnList::resume(bool isRestart)
+void FileConnList::resume(bool isRestart)
 {
   ConnectionList::resume(isRestart);
   remapShmMaps();
+
+  if (isRestart) {
+    // Now unlink the files that we created as a side-effect of restoreShmArea.
+    for (size_t i = 0; i < missingUnlinkedShmFiles.size(); i++) {
+      JWARNING(unlink(missingUnlinkedShmFiles[i].name) != -1)
+        (missingUnlinkedShmFiles[i].name) (JASSERT_ERRNO)
+        .Text("The file was unlinked at the time of checkpoint. "
+              "Unlinking it after restart failed");
+    }
+  }
 }
 
-void dmtcp::FileConnList::prepareShmList()
+void FileConnList::prepareShmList()
 {
+  ProcSelfMaps procSelfMaps;
   ProcMapsArea area;
-  int mapsfd = _real_open("/proc/self/maps", O_RDONLY, 0);
-  JASSERT(mapsfd != -1) (JASSERT_ERRNO);
 
   shmAreas.clear();
+  unlinkedShmAreas.clear();
+  missingUnlinkedShmFiles.clear();
   shmAreaConn.clear();
-  while (Util::readProcMapsLine(mapsfd, &area)) {
+  while (procSelfMaps.getNextArea(&area)) {
     if ((area.flags & MAP_SHARED) && area.prot != 0) {
       if (strstr(area.name, "ptraceSharedInfo") != NULL ||
           strstr(area.name, "dmtcpPidMap") != NULL ||
@@ -164,9 +270,23 @@ void dmtcp::FileConnList::prepareShmList()
           strstr(area.name, "synchronization-read-log") != NULL) {
         continue;
       }
+
+      if (Util::isNscdArea(area) ||
+          Util::isIBShmArea(area) ||
+          Util::isSysVShmArea(area)) {
+        continue;
+      }
+
+      /* Invalidate shared memory pages so that the next read to it (when we are
+       * writing them to ckpt file) will cause them to be reloaded from the
+       * disk.
+       */
+      JWARNING(msync(area.addr, area.size, MS_INVALIDATE) == 0)
+        (area.addr) (area.size) (area.name) (area.offset) (JASSERT_ERRNO);
+
       if (jalib::Filesystem::FileExists(area.name)) {
         if (_real_access(area.name, W_OK) == 0) {
-          JTRACE("Will checkpoint shared memory area") (area.name);
+          JLOG(FILEP)("Will checkpoint shared memory area") (area.name);
           int flags = Util::memProtToOpenFlags(area.prot);
           int fd = _real_open(area.name, flags, 0);
           JASSERT(fd != -1) (JASSERT_ERRNO) (area.name);
@@ -190,34 +310,83 @@ void dmtcp::FileConnList::prepareShmList()
                              MAP_PRIVATE | MAP_ANONYMOUS | MAP_FIXED,
                              -1, 0) != MAP_FAILED) (JASSERT_ERRNO);
         } else {
-          JTRACE("Will not checkpoint shared memory area") (area.name);
+          JLOG(FILEP)("Will not checkpoint shared memory area") (area.name);
         }
       } else {
         // TODO: Shared memory areas with unlinked backing files.
-#if 0
         JASSERT(Util::strEndsWith(area.name, DELETED_FILE_SUFFIX)) (area.name);
         if (Util::strStartsWith(area.name, DEV_ZERO_DELETED_STR) ||
             Util::strStartsWith(area.name, DEV_NULL_DELETED_STR)) {
           JWARNING(false) (area.name)
-            .Text("Ckpt/Restart of Anon Shared memory not supported.");
+            .Text("Ckpt/Restart of anonymous shared memory not supported.");
         } else {
-          JTRACE("Will recreate shm file on restart.") (area.name);
-          //shmAreas[area] = NULL;
+          JLOG(FILEP)("Will recreate shm file on restart.") (area.name);
+
+          // Remove the DELETED suffix.
+          area.name[strlen(area.name) - strlen(DELETED_FILE_SUFFIX)] = '\0';
+          unlinkedShmAreas.push_back(area);
         }
-#endif
       }
     }
   }
-  _real_close(mapsfd);
 }
 
-void dmtcp::FileConnList::remapShmMaps()
+void FileConnList::recreateShmFileAndMap(const ProcMapsArea& area)
+{
+  // TODO(kapil): Handle /dev/zero, /dev/random, etc.
+  // Recreate file in dmtcp-tmpdir;
+  string filename = Util::removeSuffix(area.name, DELETED_FILE_SUFFIX);
+  JASSERT(Util::createDirectoryTree(area.name)) (area.name)
+    .Text("Unable to create directory in File Path");
+
+  /* Now try to create the file with O_EXCL. If we fail with EEXIST, there
+   * are two possible scenarios:
+   * - The file was created by a different restarting process with data from
+   *   checkpointed copy. It is possible that the data is "in flight", so we
+   *   should wait until the next barrier to compare the data from our copy.
+   * - The file existed before restart. After the next barrier, abort if the
+   *   contents differ from our checkpointed copy.
+   */
+  int fd = _real_open(area.name, O_CREAT | O_EXCL | O_RDWR,
+                      S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
+  JASSERT(fd != -1 || errno == EEXIST) (area.name);
+
+  if (fd == -1) {
+    fd = _real_open(area.name, O_RDWR);
+    JASSERT(fd != -1) (JASSERT_ERRNO);
+  }
+
+  // Get to the correct offset.
+  JASSERT(lseek(fd, area.offset, SEEK_SET) == area.offset) (JASSERT_ERRNO);
+  // Now populate file contents from memory.
+  JASSERT(Util::writeAll(fd, area.addr, area.size) == (ssize_t)area.size)
+    (JASSERT_ERRNO);
+  restoreShmArea(area, fd);
+}
+
+void FileConnList::restoreShmArea(const ProcMapsArea& area, int fd)
+{
+  if (fd == -1) {
+    fd = _real_open(area.name, Util::memProtToOpenFlags(area.prot));
+  }
+
+  JASSERT(fd != -1) (area.name) (JASSERT_ERRNO);
+
+  JLOG(FILEP)("Restoring shared memory area") (area.name) ((void*)area.addr);
+  void *addr = _real_mmap(area.addr, area.size, area.prot,
+                          MAP_FIXED | area.flags, fd, area.offset);
+  JASSERT(addr != MAP_FAILED) (area.flags) (area.prot) (JASSERT_ERRNO)
+    .Text("mmap failed");
+  _real_close(fd);
+}
+
+void FileConnList::remapShmMaps()
 {
   for (size_t i = 0; i < shmAreas.size(); i++) {
     ProcMapsArea *area = &shmAreas[i];
     FileConnection *fileCon = shmAreaConn[i];
     int fd = fileCon->getFds()[0];
-    JTRACE("Restoring shared memory area") (area->name) ((void*)area->addr);
+    JLOG(FILEP)("Restoring shared memory area") (area->name) ((void*)area->addr);
     void *addr = _real_mmap(area->addr, area->size, area->prot,
                             MAP_FIXED | area->flags,
                             fd, area->offset);
@@ -230,12 +399,12 @@ void dmtcp::FileConnList::remapShmMaps()
 }
 
 //examine /proc/self/fd for unknown connections
-void dmtcp::FileConnList::scanForPreExisting()
+void FileConnList::scanForPreExisting()
 {
   // FIXME: Detect stdin/out/err fds to detect duplicates.
-  dmtcp::vector<int> fds = jalib::Filesystem::ListOpenFds();
-  dmtcp::string ctty = jalib::Filesystem::GetControllingTerm();
-  dmtcp::string parentCtty = jalib::Filesystem::GetControllingTerm(getppid());
+  vector<int> fds = jalib::Filesystem::ListOpenFds();
+  string ctty = jalib::Filesystem::GetControllingTerm();
+  string parentCtty = jalib::Filesystem::GetControllingTerm(getppid());
   for (size_t i = 0; i < fds.size(); ++i) {
     int fd = fds[i];
     if (!Util::isValidFd(fd)) continue;
@@ -245,9 +414,9 @@ void dmtcp::FileConnList::scanForPreExisting()
     bool isRegularFile = (S_ISREG(statbuf.st_mode) || S_ISCHR(statbuf.st_mode) ||
                           S_ISDIR(statbuf.st_mode) || S_ISBLK(statbuf.st_mode));
 
-    dmtcp::string device = jalib::Filesystem::GetDeviceName(fd);
+    string device = jalib::Filesystem::GetDeviceName(fd);
 
-    JTRACE("scanning pre-existing device") (fd) (device);
+    JLOG(FILEP)("scanning pre-existing device") (fd) (device);
     if (device == ctty || device == parentCtty) {
       // Search if this is duplicate connection
       iterator conit;
@@ -281,6 +450,19 @@ void dmtcp::FileConnList::scanForPreExisting()
       processFileConnection(fd, device.c_str(), -1, -1);
     } else if( fd <= 2 ){
       add(fd, new StdioConnection(fd));
+    } else if (getenv("PBS_JOBID") &&
+               (Util::strStartsWith(device, "/proc") &&
+                Util::strEndsWith(device, "environ"))) {
+      /*
+       * This is a workaround for an issue seen with PBS at ANU-NCI.
+       *
+       * Application processes would inherit a "/proc/<pid>/environ"
+       * file-descriptor when launched under PBS. As far as I can tell,
+       * the file-descriptor is inherited from the PBS launcher process
+       * on the compute node. The workaround is to recognize this fd as
+       * a pre-existing device and ignore it for checkpoint-restart.
+       */
+      continue;
     } else if (Util::strStartsWith(device, "/")) {
       if (isRegularFile) {
         Connection *c = findDuplication(fd, device.c_str());
@@ -294,7 +476,7 @@ void dmtcp::FileConnList::scanForPreExisting()
   }
 }
 
-Connection *dmtcp::FileConnList::findDuplication(int fd, const char *path)
+Connection *FileConnList::findDuplication(int fd, const char *path)
 {
   string npath(path);
   for (iterator i = begin(); i != end(); ++i) {
@@ -305,21 +487,21 @@ Connection *dmtcp::FileConnList::findDuplication(int fd, const char *path)
 
     FileConnection *fcon = (FileConnection*)con;
     // check for duplication
-    if( fcon->filePath() == npath && fcon->checkDup(fd) ){
+    if( fcon->checkDup(fd, path) ){
       return con;
     }
   }
   return NULL;
 }
 
-void dmtcp::FileConnList::processFileConnection(int fd, const char *path,
+void FileConnList::processFileConnection(int fd, const char *path,
                                                 int flags, mode_t mode)
 {
   Connection *c = NULL;
   struct stat statbuf;
   JASSERT(fstat(fd, &statbuf) == 0);
 
-  dmtcp::string device;
+  string device;
   if (path == NULL) {
     device = jalib::Filesystem::GetDeviceName(fd);
   } else {
@@ -335,19 +517,36 @@ void dmtcp::FileConnList::processFileConnection(int fd, const char *path,
     c = new PtyConnection(fd, path, flags, mode, PtyConnection::PTY_DEV_TTY);
   } else if (strcmp(path, "/dev/pty") == 0) {
     JASSERT(false) .Text("Not Implemented");
-  } else if (dmtcp::Util::strStartsWith(path, "/dev/pty")) {
+  } else if (Util::strStartsWith(path, "/dev/pty")) {
     // BSD Master
     c = new PtyConnection(fd, path, flags, mode, PtyConnection::PTY_BSD_MASTER);
-  } else if (dmtcp::Util::strStartsWith(path, "/dev/tty")) {
+  } else if (Util::strStartsWith(path, "/dev/tty")) {
     // BSD Slave
     c = new PtyConnection(fd, path, flags, mode, PtyConnection::PTY_BSD_SLAVE);
   } else if (strcmp(path, "/dev/ptmx") == 0 ||
              strcmp(path, "/dev/pts/ptmx") == 0) {
     // POSIX Master PTY
     c = new PtyConnection(fd, path, flags, mode, PtyConnection::PTY_MASTER);
-  } else if (dmtcp::Util::strStartsWith(path, "/dev/pts/")) {
+  } else if (Util::strStartsWith(path, "/dev/pts/")) {
     // POSIX Slave PTY
-    c = new PtyConnection(fd, path, flags, mode, PtyConnection::PTY_SLAVE);
+    PtyConnection *con = new PtyConnection(fd,
+                                           path,
+                                           flags,
+                                           mode,
+                                           PtyConnection::PTY_SLAVE);
+    /*
+     * '/dev/pts/_' can also point to controlling terminal,
+     *  especially when program tries to open controlling terminal
+     *  with open system call.
+     */
+    string ctty = jalib::Filesystem::GetControllingTerm();
+    string parentCtty = jalib::Filesystem::GetControllingTerm(getppid());
+    if (device == parentCtty) {
+        con -> markPreExistingPCTTY();
+    } else if (device == ctty) {
+        con -> markPreExistingCTTY();
+    }
+    c = (Connection *)con;
   } else if (S_ISREG(statbuf.st_mode) || S_ISCHR(statbuf.st_mode) ||
              S_ISDIR(statbuf.st_mode) || S_ISBLK(statbuf.st_mode)) {
     int type = FileConnection::FILE_REGULAR;
@@ -367,7 +566,7 @@ void dmtcp::FileConnList::processFileConnection(int fd, const char *path,
 }
 
 
-Connection *dmtcp::FileConnList::createDummyConnection(int type)
+Connection *FileConnList::createDummyConnection(int type)
 {
   switch (type) {
     case Connection::FILE:
