@@ -22,8 +22,15 @@
 #ifndef SYSCALLWRAPPERS_H
 #define SYSCALLWRAPPERS_H
 
+// FIXME:  Why are we adding all these includes here, if we're declaring
+//         only our own _real_XXX() functions?  Some *wrappers.cpp files
+//         use these includes.  But, then we should split up these includes
+//         among the individual *wrappers.cpp files that actually need them,
+//         and not declare every possible include in one giant .h file.
+#include <features.h>
 #include <pthread.h>
 #include <stdio.h>
+#include <sys/resource.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <signal.h>
@@ -33,7 +40,15 @@
 #include <stdarg.h>
 #include <sys/shm.h>
 #include <sys/sem.h>
-#include <sys/msg.h>
+// FIXME:  The SLES 10 (glibc-2.4) declaration for msgctl differs from
+//         our wrapper's declaration, which uses the POSIX declaration.
+//         If for syscallsreal.c, use the local (e.g., SLES 10) decl.
+//         Otherwise, skip declaration, and use decl. in pidwrappers.h
+//         or in syscallwrappers.h.  The correct solution should be to fix
+//         the way that *wrappers.h get their decl. of msgctl().
+#ifdef FOR_SYSCALLSREAL_C
+# include <sys/msg.h>
+#endif
 #ifdef __cplusplus
 # include <sys/stat.h>
 #else
@@ -58,9 +73,8 @@
 #include "constants.h"
 #include "dmtcp.h"
 #include "mtcp/ldt.h"
-#ifdef HAVE_CONFIG_H
-# include "config.h"
-#endif
+#include "config.h"
+
 #ifdef HAVE_SYS_EPOLL_H
 # include <sys/epoll.h>
 #else
@@ -69,9 +83,9 @@
 #  define _SYS_EPOLL_H    1
    struct epoll_event {int dummy;};
    /* Valid opcodes ("op" parameter) to issue to epoll_ctl().  */
-#  define EPOLL_CTL_ADD 1 /* Add a file decriptor to the interface.  */
-#  define EPOLL_CTL_DEL 2 /* Remove a file decriptor from the interface.  */
-#  define EPOLL_CTL_MOD 3 /* Change file decriptor epoll_event structure.  */
+#  define EPOLL_CTL_ADD 1 /* Add a file descriptor to the interface.  */
+#  define EPOLL_CTL_DEL 2 /* Remove a file descriptor from the interface.  */
+#  define EPOLL_CTL_MOD 3 /* Change file descriptor epoll_event structure.  */
 # endif
 #endif
 
@@ -82,14 +96,13 @@ extern "C"
 {
 #endif
 
-#ifdef __arm__
+#if defined(__arm__) || defined(__aarch64__)
 # define DISABLE_PTHREAD_GETSPECIFIC_TRICK
 #endif
 
-LIB_PRIVATE pid_t gettid();
-LIB_PRIVATE int tkill(int tid, int sig);
-LIB_PRIVATE int tgkill(int tgid, int tid, int sig);
-
+LIB_PRIVATE pid_t dmtcp_gettid();
+LIB_PRIVATE int dmtcp_tkill(int tid, int sig);
+LIB_PRIVATE int dmtcp_tgkill(int tgid, int tid, int sig);
 
 extern int dmtcp_wrappers_initializing;
 
@@ -189,6 +202,7 @@ LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
   MACRO(close)                              \
   MACRO(fclose)                             \
   MACRO(closedir)                           \
+  MACRO(setrlimit)                          \
   MACRO(dup)                                \
   MACRO(dup2)                               \
   MACRO(dup3)                               \
@@ -234,18 +248,6 @@ LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
   MACRO(select)                             \
   MACRO(poll)                               \
                                             \
-  MACRO(epoll_create)                       \
-  MACRO(epoll_create1)                      \
-  MACRO(epoll_ctl)                          \
-  MACRO(epoll_wait)                         \
-  MACRO(epoll_pwait)                        \
-  MACRO(eventfd)                            \
-  MACRO(signalfd)                           \
-  MACRO(inotify_init)                       \
-  MACRO(inotify_init1)                      \
-  MACRO(inotify_add_watch)                  \
-  MACRO(inotify_rm_watch)                   \
-                                            \
   MACRO(pthread_create)                     \
   MACRO(pthread_exit)                       \
   MACRO(pthread_tryjoin_np)                 \
@@ -277,7 +279,9 @@ LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
 # define GEN_ENUM(x) ENUM(x),
   typedef enum {
     FOREACH_DMTCP_WRAPPER(GEN_ENUM)
+#ifdef ENABLE_PTHREAD_COND_WRAPPERS
     FOREACH_LIBPTHREAD_WRAPPERS(GEN_ENUM)
+#endif // #ifdef ENABLE_PTHREAD_COND_WRAPPERS
     numLibcWrappers
   } LibcWrapperOffset;
 
@@ -296,8 +300,7 @@ LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
   LIB_PRIVATE void dmtcpResetPidPpid();
 
   int _dmtcp_unsetenv(const char *name);
-  void initialize_libc_wrappers();
-  void initialize_libpthread_wrappers();
+  void dmtcp_prepare_wrappers();
 
   int _real_socket (int domain, int type, int protocol);
   int _real_connect (int sockfd,  const  struct sockaddr *serv_addr,
@@ -339,6 +342,7 @@ LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
   int _real_close (int fd);
   int _real_fclose (FILE *fp);
   int _real_closedir (DIR *dir);
+  int _real_setrlimit(int resource, const struct rlimit *rlim);
   void _real_exit (int status);
   int _real_dup (int oldfd);
   int _real_dup2 (int oldfd, int newfd);
@@ -365,7 +369,9 @@ LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
                       struct sigaction *oldact);
   int _real_rt_sigaction(int signum, const struct sigaction *act,
                          struct sigaction *oldact);
+#if !__GLIBC_PREREQ(2,21)
   int _real_sigvec(int sig, const struct sigvec *vec, struct sigvec *ovec);
+#endif
 
   //set the mask
   int _real_sigblock(int mask);
@@ -390,11 +396,7 @@ LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
   int _real_sigtimedwait(const sigset_t *set, siginfo_t *info,
                          const struct timespec *timeout);
 
-  pid_t _real_gettid(void);
-  int   _real_tkill(int tid, int sig);
-  int   _real_tgkill(int tgid, int tid, int sig);
-
-  SYSCALL_ARG_RET_TYPE _real_syscall(SYSCALL_ARG_RET_TYPE sys_num, ...);
+  long _real_syscall(long sys_num, ...);
 
   int _real_pthread_create(pthread_t *thread, const pthread_attr_t *attr,
       void *(*start_routine)(void*), void *arg);
@@ -446,6 +448,8 @@ LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
   int _real_pthread_rwlock_tryrdlock(pthread_rwlock_t *rwlock);
   int _real_pthread_rwlock_wrlock(pthread_rwlock_t *rwlock);
   int _real_pthread_rwlock_trywrlock(pthread_rwlock_t *rwlock);
+
+#ifdef ENABLE_PTHREAD_COND_WRAPPERS
   int _real_pthread_cond_broadcast(pthread_cond_t *cond);
   int _real_pthread_cond_destroy(pthread_cond_t *cond);
   int _real_pthread_cond_init(pthread_cond_t *cond,
@@ -454,22 +458,9 @@ LIB_PRIVATE extern __thread int thread_performing_dlopen_dlsym;
   int _real_pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex,
                                    const struct timespec *abstime);
   int _real_pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex);
+#endif // #ifdef ENABLE_PTHREAD_COND_WRAPPERS
 
-  int _real_poll(struct pollfd *fds, nfds_t nfds, POLL_TIMEOUT_TYPE timeout);
-
-  int _real_epoll_create(int size);
-  int _real_epoll_create1(int flags);
-  int _real_epoll_ctl(int epfd, int op, int fd, struct epoll_event *event);
-  int _real_epoll_wait(int epfd, struct epoll_event *events,
-                       int maxevents, int timeout);
-  int _real_epoll_pwait(int epfd, struct epoll_event *events,
-                        int maxevents, int timeout, const sigset_t *sigmask);
-  int _real_eventfd(EVENTFD_VAL_TYPE initval, int flags);
-  int _real_signalfd (int fd, const sigset_t *mask, int flags);
-  int _real_inotify_init(void);
-  int _real_inotify_init1(int flags);
-  int _real_inotify_add_watch(int fd, const char *pathname, uint32_t mask);
-  int _real_inotify_rm_watch(int fd, int wd);
+  int _real_poll(struct pollfd *fds, nfds_t nfds, int timeout);
 
   int   _real_waitid(idtype_t idtype, id_t id, siginfo_t *infop, int options);
   pid_t _real_wait4(pid_t pid, __WAIT_STATUS status, int options,
