@@ -18,7 +18,6 @@
  *  License along with DMTCP:dmtcp/src.  If not, see                        *
  *  <http://www.gnu.org/licenses/>.                                         *
  ****************************************************************************/
-
 #include <errno.h>
 #include <sched.h>
 #include <signal.h>
@@ -31,11 +30,22 @@
 #include <sys/stat.h>
 #include <sys/fcntl.h>
 #include "dmtcp.h"
+#include "constants.h"
 #include "processinfo.h"
 #include "procmapsarea.h"
+#include "procselfmaps.h"
 #include "jassert.h"
 #include "util.h"
 
+#define DEV_ZERO_DELETED_STR "/dev/zero (deleted)"
+#define DEV_NULL_DELETED_STR "/dev/null (deleted)"
+
+/* Shared memory regions for Direct Rendering Infrastructure */
+#define DEV_DRI_SHMEM "/dev/dri/card"
+
+#define DELETED_FILE_SUFFIX " (deleted)"
+
+
 #define _real_open NEXT_FNC(open)
 #define _real_close NEXT_FNC(close)
 
@@ -43,22 +53,38 @@ using namespace dmtcp;
 
 EXTERNC int dmtcp_infiniband_enabled(void) __attribute__((weak));
 
-static const int END_OF_NSCD_AREAS = -1;
+static bool skipWritingTextSegments = false;
+
+// FIXME:  Why do we create two global variable here?  They should at least
+//         be static (file-private), and preferably local to a function.
+ProcSelfMaps *procSelfMaps = NULL;
+vector<ProcMapsArea> *nscdAreas = NULL;
+// FIXME:  If we allocate in the middle of reading
+//         /proc/self/maps, we modify the mapping.  But whenever we
+//         add to nscdAreas, we risk allocating memory.  So, we're depending
+//         on this memory being smaller than any pre-allocated memory,
+//         so that the memory allocator does not call mmap in the middle
+//         of reading /proc/self/maps.  A better design would be to create
+//         an nscdArea method of the ProcSelfMaps class, and that
+//         class can then be careful about allocating memory.
+
 
 /* Internal routines */
 //static void sync_shared_mem(void);
 static void writememoryarea (int fd, Area *area,
-                             int stack_was_seen, int vsyscall_exists);
-static void preprocess_special_segments(int *vsyscall_exists);
+                             int stack_was_seen);
 
-static void remap_nscd_areas(Area remap_nscd_areas_array[],
-                             int  num_remap_nscd_areas);
+static void remap_nscd_areas(const vector<ProcMapsArea> & areas);
 
 /*****************************************************************************
  *
  *  This routine is called from time-to-time to write a new checkpoint file.
  *  It assumes all the threads are suspended.
  *
+ *  NOTE: Any memory allocated in this function should be released explicitly
+ *  during the next ckpt cycle. Otherwise, on restart, we never come back to
+ *  this function which can cause memory leaks.
+ *
  *****************************************************************************/
 
 void mtcp_writememoryareas(int fd)
@@ -67,17 +93,17 @@ void mtcp_writememoryareas(int fd)
   //DeviceInfo dev_info;
   int stack_was_seen = 0;
 
-  JTRACE("Performing checkpoint.");
+  if (getenv(ENV_VAR_SKIP_WRITING_TEXT_SEGMENTS) != NULL) {
+    skipWritingTextSegments = true;
+  }
+
+  JLOG(DMTCP)("Performing checkpoint.");
 
   // Here we want to sync the shared memory pages with the backup files
   // FIXME: Why do we need this?
-  //JTRACE("syncing shared memory with backup files");
+  //JLOG(DMTCP)("syncing shared memory with backup files");
   //sync_shared_mem();
 
-  int vsyscall_exists = 0;
-  // Preprocess special segments like vsyscall, stack, heap etc.
-  preprocess_special_segments(&vsyscall_exists);
-
   /**************************************************************************/
   /* We can't do any more mallocing at this point because malloc stuff is   */
   /* outside the limits of the libmtcp.so image, so it won't get            */
@@ -85,20 +111,49 @@ void mtcp_writememoryareas(int fd)
   /* inconsistent state.  See note in restoreverything routine.             */
   /**************************************************************************/
 
-  int num_remap_nscd_areas = 0;
-  Area remap_nscd_areas_array[10];
-  remap_nscd_areas_array[9].flags = END_OF_NSCD_AREAS;
+  {
+    if (nscdAreas == NULL) {
+      nscdAreas = new vector<ProcMapsArea>();
+    }
+    nscdAreas->clear();
+    // This block is to ensure that the object is deleted as soon as we leave
+    // this block.
+    ProcSelfMaps procSelfMaps;
+    // Preprocess memory regions as needed.
+    while (procSelfMaps.getNextArea(&area)) {
+      if (Util::isNscdArea(area)) {
+        /* Special Case Handling: nscd is enabled*/
+        JLOG(DMTCP)("NSCD daemon shared memory area present.\n"
+               "  DMTCP will now try to remap this area in read/write mode as\n"
+               "  private (zero pages), so that glibc will automatically\n"
+               "  stop using NSCD or ask NSCD daemon for new shared area\n")
+          (area.name);
+
+        nscdAreas->push_back(area);
+      }
+    }
+  }
+
+  if (procSelfMaps != NULL) {
+    // We need to explicitly delete this object here because on restart, we
+    // never get back to this function and the object is never released.
+    delete procSelfMaps;
+  };
 
   /* Finally comes the memory contents */
-  int mapsfd = _real_open("/proc/self/maps", O_RDONLY);
-  while (Util::readProcMapsLine(mapsfd, &area)) {
-    VA area_begin = area.addr;
-    /* VA area_end   = area_begin + area.size; */
+  procSelfMaps = new ProcSelfMaps();
+  while (procSelfMaps->getNextArea(&area)) {
+    // TODO(kapil): Verify that we are not doing any operation that might
+    // result in a change of memory layout. For example, a call to JALLOC_NEW
+    // will invoke mmap if the JAlloc arena is full. Similarly, for STL objects
+    // such as vector and string.
 
     if ((uint64_t)area.addr == ProcessInfo::instance().restoreBufAddr()) {
       JASSERT(area.size == ProcessInfo::instance().restoreBufLen())
         ((void*) area.addr) (area.size) (ProcessInfo::instance().restoreBufLen());
       continue;
+    } else if (SharedData::isSharedDataRegion(area.addr)) {
+      continue;
     }
 
     /* Original comment:  Skip anything in kernel address space ---
@@ -106,20 +161,20 @@ void mtcp_writememoryareas(int fd)
      * Added: That's the vdso section for earlier Linux 2.6 kernels.  For later
      *  2.6 kernels, vdso occurs at an earlier address.  If it's unreadable,
      *  then we simply won't copy it.  But let's try to read all areas, anyway.
-     * **COMMENTED OUT:** if (area_begin >= HIGHEST_VA) continue;
+     * **COMMENTED OUT:** if (area.addr >= HIGHEST_VA) continue;
      */
     /* If it's readable, but it's VDSO, it will be dangerous to restore it.
      * In 32-bit mode later Red Hat RHEL Linux 2.6.9 releases use 0xffffe000,
      * the last page of virtual memory.  Note 0xffffe000 >= HIGHEST_VA
      * implies we're in 32-bit mode.
      */
-    if (area_begin >= HIGHEST_VA && area_begin == (VA)0xffffe000)
+    if (area.addr >= HIGHEST_VA && area.addr == (VA)0xffffe000)
       continue;
 #ifdef __x86_64__
     /* And in 64-bit mode later Red Hat RHEL Linux 2.6.9 releases
      * use 0xffffffffff600000 for VDSO.
      */
-    if (area_begin >= HIGHEST_VA && area_begin == (VA)0xffffffffff600000)
+    if (area.addr >= HIGHEST_VA && area.addr == (VA)0xffffffffff600000)
       continue;
 #endif
 
@@ -148,16 +203,21 @@ void mtcp_writememoryareas(int fd)
      *
      * We suspect that libpthread is using mmap() instead of mprotect to change
      * the permission from "---p" to "rw-p".
+     *
+     * Also, on SUSE 12, if this region was part of heap, the protected region
+     * may have the label "[heap]".  So, we also save the memory region if it
+     * has label "[heap]", "[stack]", or  "[stack:XXX]".
      */
 
     if (!((area.prot & PROT_READ) || (area.prot & PROT_WRITE)) &&
-        area.name[0] != '\0') {
+        (area.name[0] != '\0') && strcmp(area.name, "[heap]") &&
+        strcmp(area.name, "[stack]") && (!Util::strStartsWith(area.name, "[stack:XXX]"))){
       continue;
     }
 
     if (Util::strStartsWith(area.name, DEV_ZERO_DELETED_STR) ||
         Util::strStartsWith(area.name, DEV_NULL_DELETED_STR)) {
-      /* If the process has an area labelled as "/dev/zero (deleted)", we mark
+      /* If the process has an area labeled as "/dev/zero (deleted)", we mark
        *   the area as Anonymous and save the contents to the ckpt image file.
        * If this area has a MAP_SHARED attribute, it should be replaced with
        *   MAP_PRIVATE and we won't do any harm because, the /dev/zero file is
@@ -168,66 +228,29 @@ void mtcp_writememoryareas(int fd)
        *
        * The above explanation also applies to "/dev/null (deleted)"
        */
-      JTRACE("saving area as Anonymous") (area.name);
+      JLOG(DMTCP)("saving area as Anonymous") (area.name);
       area.flags = MAP_PRIVATE | MAP_ANONYMOUS;
       area.name[0] = '\0';
-    } else if (Util::strStartsWith(area.name, SYS_V_SHMEM_FILE)) {
-      JTRACE("saving area as Anonymous") (area.name);
+    } else if (Util::isSysVShmArea(area)) {
+      JLOG(DMTCP)("saving area as Anonymous") (area.name);
       area.flags = MAP_PRIVATE | MAP_ANONYMOUS;
       area.name[0] = '\0';
-    } else if (Util::strStartsWith(area.name, NSCD_MMAP_STR1) ||
-               Util::strStartsWith(area.name, NSCD_MMAP_STR2) ||
-               Util::strStartsWith(area.name, NSCD_MMAP_STR3)) {
+    } else if (Util::isNscdArea(area)) {
       /* Special Case Handling: nscd is enabled*/
-      JTRACE("NSCD daemon shared memory area present.\n"
-              "  MTCP will now try to remap this area in read/write mode as\n"
-              "  private (zero pages), so that glibc will automatically\n"
-              "  stop using NSCD or ask NSCD daemon for new shared area\n")
-        (area.name);
-      area.prot = PROT_READ | PROT_WRITE | MTCP_PROT_ZERO_PAGE;
+      area.prot = PROT_READ | PROT_WRITE;
+      area.properties |= DMTCP_ZERO_PAGE;
       area.flags = MAP_PRIVATE | MAP_ANONYMOUS;
-
-      /* We're still using proc-maps in mtcp_readmapsline();
-       * So, remap NSCD later.
-       */
-      remap_nscd_areas_array[num_remap_nscd_areas++] = area;
       Util::writeAll(fd, &area, sizeof(area));
       continue;
-    }
-    else if (Util::strStartsWith(area.name, INFINIBAND_SHMEM_FILE)) {
+    } else if (Util::isIBShmArea(area)) {
       // TODO: Don't checkpoint infiniband shared area for now.
       continue;
-    }
-    else if (Util::strEndsWith(area.name, DELETED_FILE_SUFFIX)) {
+    } else if (Util::strEndsWith(area.name, DELETED_FILE_SUFFIX)) {
       /* Deleted File */
     } else if (area.name[0] == '/' && strstr(&area.name[1], "/") != NULL) {
       /* If an absolute pathname
        * Posix and SysV shared memory segments can be mapped as /XYZ
        */
-#if 0
-      struct stat statbuf;
-      unsigned int long devnum;
-      if (stat(area.name, &statbuf) < 0) {
-        JWARNING(false) (area.name) (JASSERT_ERRNO)
-          .Text("Error statting file.");
-      } else {
-        devnum = makedev(dev_info.devmajor, dev_info.devminor);
-        JWARNING (devnum == statbuf.st_dev && dev_info.inodenum == statbuf.st_ino)
-          (area.name) (statbuf.st_dev) (statbuf.st_ino)
-          (devnum) (dev_info.inodenum);
-      }
-#endif
-    }
-
-    area.filesize = 0;
-    if (area.name[0] != '\0') {
-      int ffd = _real_open(area.name, O_RDONLY, 0);
-      if (ffd != -1) {
-        area.filesize = lseek(ffd, 0, SEEK_END);
-        if (area.filesize == -1)
-          area.filesize = 0;
-      }
-      _real_close(ffd);
     }
 
     /* Force the anonymous flag if it's a private writeable section, as the
@@ -242,16 +265,6 @@ void mtcp_writememoryareas(int fd)
       area.flags |= MAP_ANONYMOUS;
     }
 
-    if (area.flags & MAP_SHARED) {
-      /* invalidate shared memory pages so that the next read to it (when we are
-       * writing them to ckpt file) will cause them to be reloaded from the
-       * disk.
-       */
-      JASSERT(msync(area.addr, area.size, MS_INVALIDATE) == 0)
-        (area.addr) (area.size) (area.name) (area.offset) (JASSERT_ERRNO);
-    }
-
-
     /* Only write this image if it is not CS_RESTOREIMAGE.
      * Skip any mapping for this image - it got saved as CS_RESTOREIMAGE
      * at the beginning.
@@ -259,14 +272,17 @@ void mtcp_writememoryareas(int fd)
 
     if (strstr (area.name, "[stack]"))
       stack_was_seen = 1;
+
     // the whole thing comes after the restore image
-    writememoryarea(fd, &area, stack_was_seen, vsyscall_exists);
+    writememoryarea(fd, &area, stack_was_seen);
   }
 
-  /* It's now safe to do this, since we're done using mtcp_readmapsline() */
-  remap_nscd_areas(remap_nscd_areas_array, num_remap_nscd_areas);
+  // Release the memory.
+  delete procSelfMaps;
+  procSelfMaps = NULL;
 
-  close (mapsfd);
+  /* It's now safe to do this, since we're done using writememoryarea() */
+  remap_nscd_areas(*nscdAreas);
 
   area.addr = NULL; // End of data
   area.size = -1; // End of data
@@ -276,26 +292,14 @@ void mtcp_writememoryareas(int fd)
   JASSERT(_real_close (fd) == 0);
 }
 
-/* FIXME:
- * We should read /proc/self/maps into temporary array and mtcp_readmapsline
- * should then read from it.  This is cleaner than this hack here.
- * Then this body can go back to replacing:
- *    remap_nscd_areas_array[num_remap_nscd_areas++] = area;
- * - Gene
- */
-static void remap_nscd_areas(Area remap_nscd_areas_array[],
-			     int  num_remap_nscd_areas)
+static void remap_nscd_areas(const vector<ProcMapsArea>& areas)
 {
-  Area *area;
-  for (area = remap_nscd_areas_array; num_remap_nscd_areas-- > 0; area++) {
-    JASSERT(area->flags != END_OF_NSCD_AREAS)
-      .Text("Too many NSCD areas to remap.");
-    JASSERT(munmap(area->addr, area->size) == 0) (JASSERT_ERRNO)
+  for (size_t i = 0; i < areas.size(); i++) {
+    JASSERT(munmap(areas[i].addr, areas[i].size) == 0) (JASSERT_ERRNO)
       .Text("error unmapping NSCD shared area");
-    JASSERT(mmap(area->addr, area->size, area->prot, area->flags, 0, 0)
-            != MAP_FAILED)
+    JASSERT(mmap(areas[i].addr, areas[i].size, areas[i].prot,
+            MAP_ANONYMOUS | MAP_PRIVATE | MAP_FIXED, 0, 0) != MAP_FAILED)
       (JASSERT_ERRNO) .Text("error remapping NSCD shared area.");
-    memset(area->addr, 0, area->size);
   }
 }
 
@@ -341,9 +345,9 @@ static void mtcp_get_next_page_range(Area *area, size_t *size, int *is_zero)
 static void mtcp_write_non_rwx_and_anonymous_pages(int fd, Area *orig_area)
 {
   Area area = *orig_area;
-  /* Now give read permission to the anonymous pages that do not have read
-   * permission. We should remove the permission as soon as we are done
-   * writing the area to the checkpoint image
+  /* Now give read permission to the anonymous/[heap]/[stack]/[stack:XXX] pages
+   * that do not have read permission. We should remove the permission
+   * as soon as we are done writing the area to the checkpoint image
    *
    * NOTE: Changing the permission here can results in two adjacent memory
    * areas to become one (merged), if they have similar permissions. This can
@@ -352,7 +356,10 @@ static void mtcp_write_non_rwx_and_anonymous_pages(int fd, Area *orig_area)
    * code and that should reset the /proc/self/maps files to its original
    * condition.
    */
-  JASSERT(orig_area->name[0] == '\0');
+
+  JASSERT(orig_area->name[0] == '\0' || (strcmp(orig_area->name, "[heap]") == 0) ||
+         (strcmp(orig_area->name, "[stack]") == 0) ||
+         (Util::strStartsWith(area.name, "[stack:XXX]")));
 
   if ((orig_area->prot & PROT_READ) == 0) {
     JASSERT(mprotect(orig_area->addr, orig_area->size,
@@ -372,7 +379,7 @@ static void mtcp_write_non_rwx_and_anonymous_pages(int fd, Area *orig_area)
       mtcp_get_next_page_range(&a, &size, &is_zero);
     }
 
-    a.prot |= is_zero ? MTCP_PROT_ZERO_PAGE : 0;
+    a.properties = is_zero ? DMTCP_ZERO_PAGE : 0;
     a.size = size;
 
     Util::writeAll(fd, &a, sizeof(a));
@@ -397,34 +404,17 @@ static void mtcp_write_non_rwx_and_anonymous_pages(int fd, Area *orig_area)
   }
 }
 
-static void writememoryarea (int fd, Area *area, int stack_was_seen,
-			     int vsyscall_exists)
+static void writememoryarea (int fd, Area *area, int stack_was_seen)
 {
-  static void * orig_stack = NULL;
   void *addr = area->addr;
 
-  /* Write corresponding descriptor to the file */
-
-  if (orig_stack == NULL && 0 == strcmp(area -> name, "[stack]"))
-    orig_stack = area -> addr + area -> size;
-
-  if (0 == strcmp(area -> name, "[vdso]") && !stack_was_seen)
-    JTRACE("skipping over [vdso] section") (addr) (area->size);
-  else if (0 == strcmp(area -> name, "[vsyscall]") && !stack_was_seen)
-    JTRACE("skipping over [vsyscall] section") (addr) (area->size);
-  else if (0 == strcmp(area -> name, "[vectors]") && !stack_was_seen)
-    JTRACE("skipping over [vectors] section") (addr) (area->size);
-  else if (0 == strcmp(area -> name, "[stack]") &&
-	   orig_stack != area -> addr + area -> size)
-    /* Kernel won't let us munmap this.  But we don't need to restore it. */
-    JTRACE("skipping over [stack] segment (not the orig stack)")
-      (addr) (area->size);
-  else if (!(area -> flags & MAP_ANONYMOUS))
-    JTRACE("save region") (addr) (area->size) (area->name) (area->offset);
-  else if (area -> name[0] == '\0')
-    JTRACE("save anonymous") (addr) (area->size);
-  else
-    JTRACE("save anonymous") (addr) (area->size) (area->name) (area->offset);
+  if (!(area -> flags & MAP_ANONYMOUS)) {
+    JLOG(DMTCP)("save region") (addr) (area->size) (area->name) (area->offset);
+  } else if (area -> name[0] == '\0') {
+    JLOG(DMTCP)("save anonymous") (addr) (area->size);
+  } else {
+    JLOG(DMTCP)("save anonymous") (addr) (area->size) (area->name) (area->offset);
+  }
 
   if ((area -> name[0]) == '\0') {
     char *brk = (char*)sbrk(0);
@@ -432,7 +422,45 @@ static void writememoryarea (int fd, Area *area, int stack_was_seen,
       strcpy(area -> name, "[heap]");
   }
 
-  if (area->prot == 0 ||
+  if (area->size == 0) {
+    /* Kernel won't let us munmap this.  But we don't need to restore it. */
+    JLOG(DMTCP)("skipping over [stack] segment (not the orig stack)")
+      (addr) (area->size);
+  } else if (0 == strcmp(area -> name, "[vsyscall]") ||
+             0 == strcmp(area -> name, "[vectors]") ||
+             0 == strcmp(area -> name, "[vvar]"))
+             // NOTE: We can't trust kernel's "[vdso]" label here.  See below.
+  {
+    JLOG(DMTCP)("skipping over memory special section")
+      (area->name) (addr) (area->size);
+  } else if ( area->__addr == ProcessInfo::instance().vdsoStart() ) {
+    //vDSO issue:
+    //    As always, we never want to save the vdso section.  We will use
+    //  the vdso section code provided by the kernel on restart.  Further,
+    //  the user code on restart has already been initialized and so it
+    //  will continue to use the original vdso section determined during
+    //  program launch.  Luckily, during the DMTCP_INIT event, DMTCP recorded
+    //  this vdso address when it called ProcessInfo::instance().init().
+    //    Now, here's the bad news.  During the first restart, the kernel
+    //  may choose to locate the vdso at a new address.  So, in
+    //  src/mtcp/mtcp_restart, DMTCP will mremap the kernel's vdso back
+    //  to the original address known during program launch.  This is as
+    //  it should be.  But when DMTCP does an mremap of vdso, the kernel
+    //  fails to update its own "[vds0]" label.  So, during the second
+    //  checkpoint (after the first restart), we can't trust the "[vdso]"
+    //  label to tell us where the vdso section really is.  And it's even
+    //  worse.  During mtcp_restart, we may have done the mremap, and there
+    //  may even now be some user data that was restored to the address
+    //  where the kernel thinks the "[vdso]" label belongs.  So, we would
+    //  be saving the original vdso section (which is wrong), and we would
+    //  be failing to save the user's memory that was restored into the
+    //  location labelled by the kernel's "[vdso]" label.  This last
+    //  case is even worse, since we have now failed to restore some user data. 
+    //    This was observed to happen in RHEL 6.6.  The solution is to
+    //  trust DMTCP for the vdso location (as in the if condition above),
+    //  and not to trust the kernel's "[vdso]" label.
+    JLOG(DMTCP)("skipping vDSO section") ;
+  } else if (area->prot == 0 ||
       (area->name[0] == '\0' &&
        ((area->flags & MAP_ANONYMOUS) != 0) &&
        ((area->flags & MAP_PRIVATE) != 0))) {
@@ -441,127 +469,21 @@ static void writememoryarea (int fd, Area *area, int stack_was_seen,
      * mappings only
      */
     mtcp_write_non_rwx_and_anonymous_pages(fd, area);
-  } else if (0 != strcmp(area -> name, "[vsyscall]")
-             && 0 != strcmp(area -> name, "[vectors]")
-             && ((0 != strcmp(area -> name, "[vdso]")
-                  || vsyscall_exists /* which implies vdso can be overwritten */
-                  || !stack_was_seen))) /* If vdso appeared before stack, it can be
-                                         replaced */
-  {
+  } else {
     /* Anonymous sections need to have their data copied to the file,
      *   as there is no file that contains their data
      * We also save shared files to checkpoint file to handle shared memory
      *   implemented with backing files
      */
     JASSERT((area->flags & MAP_ANONYMOUS) || (area->flags & MAP_SHARED));
-    Util::writeAll(fd, area, sizeof(*area));
-    Util::writeAll(fd, area->addr, area->size);
-  }
-}
 
-static void preprocess_special_segments(int *vsyscall_exists)
-{
-  Area area;
-  int mapsfd = _real_open("/proc/self/maps", O_RDONLY);
-  JASSERT(mapsfd != -1) (JASSERT_ERRNO) .Text("Error opening /proc/self/maps");
-
-  while (Util::readProcMapsLine(mapsfd, &area)) {
-    if (0 == strcmp(area.name, "[vsyscall]")) {
-      /* Determine if [vsyscall] exists.  If [vdso] and [vsyscall] exist,
-       * [vdso] will be saved and restored.
-       * NOTE:  [vdso] is relocated if /proc/sys/kernel/randomize_va_space == 2.
-       * We must restore old [vdso] and also keep [vdso] in that case.
-       * On Linux 2.6.25:
-       *   32-bit Linux has:  [heap], /lib/ld-2.7.so, [vdso], libs, [stack].
-       *   64-bit Linux has:  [stack], [vdso], [vsyscall].
-       * and at least for gcl, [stack], libmtcp.so, [vsyscall] seen.
-       * If 32-bit process in 64-bit Linux:
-       *     [stack] (0xffffd000), [vdso] (0xffffe0000)
-       * On 32-bit Linux, mtcp_restart has [vdso], /lib/ld-2.7.so, [stack]
-       * Need to restore old [vdso] into mtcp_restart, to restart.
-       * With randomize_va_space turned off, libraries start at high address
-       *     0xb8000000 and are loaded progressively at lower addresses.
-       * mtcp_restart loads vdso (which looks like a shared library) first.
-       * But libpthread/libdl/libc libraries are loaded above vdso in user
-       * image.
-       * So, we must use the opposite of the user's setting (no randomization if
-       *     user turned it on, and vice versa).  We must also keep the
-       *     new vdso segment, provided by mtcp_restart.
-       */
-      *vsyscall_exists = 1;
-    } else if (/*!mtcp_saved_heap_start && */strcmp(area.name, "[heap]") == 0) {
-      // Record start of heap which will later be used in mtcp_restore_finish()
-      //mtcp_saved_heap_start = area.addr;
-    } else if (strcmp(area.name, "[stack]") == 0) {
-      /*
-       * When using Matlab with dmtcp_launch, sometimes the bottom most
-       * page of stack (the page with highest address) which contains the
-       * environment strings and the argv[] was not shown in /proc/self/maps.
-       * This is arguably a bug in the Linux kernel as of version 2.6.32, etc.
-       * This happens on some odd combination of environment passed on to
-       * Matlab process. As a result, the page was not checkpointed and hence
-       * the process segfaulted on restart. The fix is to try to mprotect this
-       * page with RWX permission to make the page visible again. This call
-       * will fail if no stack page was invisible to begin with.
-       */
-      // FIXME : If the area following the stack is not empty, don't
-      //         exercise this path.
-      int ret = mprotect(area.addr + area.size, 0x1000,
-                         PROT_READ | PROT_WRITE | PROT_EXEC);
-      if (ret == 0) {
-        JNOTE("bottom-most page of stack (page with highest address) was \n"
-              "  invisible in /proc/self/maps. It is made visible again now.");
-      }
+    if (skipWritingTextSegments && (area->prot & PROT_EXEC)) {
+      area->properties |= DMTCP_SKIP_WRITING_TEXT_SEGMENTS;
+      Util::writeAll(fd, area, sizeof(*area));
+      JLOG(DMTCP)("Skipping over text segments") (area->name) ((void*)area->addr);
+    } else {
+      Util::writeAll(fd, area, sizeof(*area));
+      Util::writeAll(fd, area->addr, area->size);
     }
   }
-  close(mapsfd);
 }
-
-#if 0
-
-/*****************************************************************************
- *
- *  Sync shared memory pages with backup files on disk
- *
- *****************************************************************************/
-static void sync_shared_mem(void)
-{
-  int mapsfd;
-  Area area;
-
-  mapsfd = _real_open("/proc/self/maps", O_RDONLY);
-  JASSERT(mapsfd != -1) (JASSERT_ERRNO) .Text("Error opening /proc/self/maps");
-
-  while (mtcp_readmapsline (mapsfd, &area, NULL)) {
-    /* Skip anything that has no read or execute permission.  This occurs on one
-     * page in a Linux 2.6.9 installation.  No idea why.  This code would also
-     * take care of kernel sections since we don't have read/execute permission
-     * there.
-     */
-
-    if (!((area.prot & PROT_READ) || (area.prot & PROT_WRITE))) continue;
-
-    if (!(area.flags & MAP_SHARED)) continue;
-
-    if (Util::strEndsWith(area.name, DELETED_FILE_SUFFIX)) continue;
-
-    /* Don't sync the DRI shared memory region for OpenGL */
-    if (Util::strStartsWith(area.name, DEV_DRI_SHMEM)) continue;
-
-#ifdef IBV
-    // TODO: Don't checkpoint infiniband shared area for now.
-   if (Util::strstartswith(area.name, INFINIBAND_SHMEM_FILE)) {
-     continue;
-   }
-#endif
-
-    JTRACE("syncing shared memory region")
-      (area.size) (area.addr) (area.name) (area.offset);
-
-    JASSERT(msync(area.addr, area.size, MS_SYNC) == 0)
-      (area.addr) (area.size) (area.name) (area.offset) (JASSERT_ERRNO);
-  }
-
-  _real_close(mapsfd);
-}
-#endif
