@@ -14,6 +14,20 @@ import pwd
 import stat
 import re
 
+
+# FIX for bad path for Java:  Travis prepended
+#     "/usr/bin:/opt/pyenv/libexec:/opt/pyenv/plugins/python-build/bin:/"
+# to os.environ['PATH'] on July 31, 2019.  It does this, even though
+# "/usr/bin" occurs later in the path.  /usr/bin/java exists as Java-8.
+# So, it never finds java (version Java-11) in
+#   "/usr/local/lib/jvm/openjdk11/bin", which occurs later in path.
+# Unfortunately, 'Makefile' finds javac as verion Javac-11,
+# and so the 'java1' test fails, using java (Java-8).
+if os.environ.has_key('PATH') and os.environ['PATH'].startswith('/usr/bin:') \
+     and ':/usr/bin:' in os.environ['PATH']:
+  os.environ['PATH'] = os.environ['PATH'][len('/usr/bin:'):]
+
+
 #get testconfig
 # This assumes Makefile.in in main dir, but only Makefile in test dir.
 try:
@@ -65,8 +79,14 @@ RETRIES=2
 DEFAULT_S=0.3
 if uname_p[0:3] == 'arm':
   DEFAULT_S *= 2
+
+# Sleep before the first ckpt _only_.
+DEFAULT_POST_LAUNCH_SLEEP = 0.0
+POST_LAUNCH_SLEEP = 0.0
+
+uname_m = uname_m.strip() # strip off any whitespace characters
 #Allow extra time for slower CPUs
-if uname_m in ["i386", "i486", "i586", "i686", "armv7", "armv71"]:
+if uname_m in ["i386", "i486", "i586", "i686", "armv7", "armv7l", "aarch64"]:
   DEFAULT_S *= 4
 
 S=DEFAULT_S
@@ -114,7 +134,8 @@ for i in sys.argv:
   if i=="--stress":
     CYCLES=100000
   if i=="--slow":
-    SLOW=5
+    SLOW *= 5  # SLOW was initialized to 1 ; --slow --slow => SLOW==25
+    TIMEOUT *= SLOW
   if i=="--retry-once":
     RETRY_ONCE = True
   #TODO:  Install SIGSEGV handler with infinite loop, and add to LD_PRELOAD
@@ -217,10 +238,10 @@ def master_read(fd):
   os.read(fd, 4096)
   return ''
 
-#launch a child process
+#run a child process
 # NOTE:  Can eventually migrate to Python 2.7:  subprocess.check_output
 devnullFd = os.open(os.devnull, os.O_WRONLY)
-def launch(cmd):
+def runCmd(cmd):
   global devnullFd
   global master_read
   if VERBOSE:
@@ -232,7 +253,8 @@ def launch(cmd):
     # Checkpoint image can be emacs23_x, or whatever emacs is a link to.
     # vim can be vim.gnome, etc.
     if re.search("(_|/|^)(screen|script|vim.*|emacs.*|pty)(_|$)", str):
-      ptyMode = True
+      # ptyMode = True
+      pass # Using ptyMode creates bug in restart for vim, etc.
   try:
     os.stat(cmd[0])
   except:
@@ -279,8 +301,8 @@ def launch(cmd):
 #randomize port and dir, so multiple processes works
 ckptDir="dmtcp-autotest-%d" % randint(100000000,999999999)
 os.mkdir(ckptDir);
-os.environ['DMTCP_HOST'] = "localhost"
-os.environ['DMTCP_PORT'] = str(randint(2000,10000))
+os.environ['DMTCP_COORD_HOST'] = "localhost"
+os.environ['DMTCP_COORD_PORT'] = str(randint(2000,10000))
 os.environ['DMTCP_CHECKPOINT_DIR'] = os.path.abspath(ckptDir)
 #Use default SIGCKPT for test suite.
 os.unsetenv('DMTCP_SIGCKPT')
@@ -290,7 +312,7 @@ os.unsetenv('MTCP_SIGCKPT')
 if not VERBOSE:
   os.environ['JALIB_STDERR_PATH'] = os.devnull
 if VERBOSE:
-  print "coordinator port:  " + os.environ['DMTCP_PORT']
+  print "coordinator port:  " + os.environ['DMTCP_COORD_PORT']
 
 # We'll copy ckptdir to DMTCP_TMPDIR in case of error.
 def dmtcp_tmpdir():
@@ -332,8 +354,8 @@ if os.getenv('LD_LIBRARY_PATH'):
 else:
     os.environ['LD_LIBRARY_PATH'] = os.getenv("PWD")+"/lib"
 
-#launch the coordinator
-coordinator = launch(BIN+"dmtcp_coordinator")
+#run the coordinator
+coordinator = runCmd(BIN+"dmtcp_coordinator")
 
 #send a command to the coordinator process
 def coordinatorCmd(cmd):
@@ -384,16 +406,21 @@ def getStatus():
 
   while True:
     try:
-      line=coordinator.stdout.readline()
+      line=coordinator.stdout.readline().strip()
       if not line:  # Immediate empty string on stdout means EOF
         CHECK(False, "coordinator died unexpectedly")
         return (-1, False)
-      line=line.strip()
-      if line=="Status...":
-        break;
-      if VERBOSE:
-        print "Ignoring line from coordinator: ", line
-        sleep(1)
+
+      m = re.search('NUM_PEERS=(\d+)', line)
+      if m != None:
+        peers = int(m.group(1))
+        continue
+
+      m = re.search('RUNNING=(\w+)', line)
+      if m != None:
+        running = m.group(1)
+        break
+
     except IOError, (errno, strerror):
       if coordinator.poll():
         if coordinator.poll() < 0:
@@ -404,14 +431,10 @@ def getStatus():
         continue
       raise CheckFailed("I/O error(%s): %s" % (errno, strerror))
 
-  x,peers=coordinator.stdout.readline().strip().split("=")
-  CHECK(x=="NUM_PEERS", "reading coordinator status")
-  x,running=coordinator.stdout.readline().strip().split("=")
-  CHECK(x=="RUNNING", "reading coordinator status")
-
   if VERBOSE:
-    print "STATUS: peers=%s, running=%s" % (peers,running)
-  return (int(peers), (running=="yes"))
+    print "STATUS: peers=%d, running=%s" % (peers,running)
+
+  return (peers, (running=="yes"))
 
 #delete all files in ckptDir
 def clearCkptDir():
@@ -473,8 +496,8 @@ def runTestRaw(name, numProcs, cmds):
       global coordinator
       coordinatorCmd('q')
       os.system("kill -9 %d" % coordinator.pid)
-      print "Trying to kill old coordinator, and launch new one on same port"
-      coordinator = launch(BIN+"dmtcp_coordinator")
+      print "Trying to kill old coordinator, and run new one on same port"
+      coordinator = runCmd(BIN+"dmtcp_coordinator")
     for x in procs:
       #cleanup proc
       try:
@@ -505,8 +528,14 @@ def runTestRaw(name, numProcs, cmds):
 
     #wait for files to appear and status to return to original
     WAITFOR(lambda: getNumCkptFiles(ckptDir)>0 and \
-                    (CKPT_CMD == 'xc' or doesStatusSatisfy(getStatus(), status)),
+                   (CKPT_CMD == 'xc' or doesStatusSatisfy(getStatus(), status)),
             wfMsg("checkpoint error"))
+    #we now know there was at least one checkpoint file, and the correct number
+    #  of processes have restarted;  but they may fail quickly after restert
+
+    if SLOW > 1:
+      #wait and give the processes time to write all of the checkpoint files
+      sleep(S*SLOW)
 
     #make sure the right files are there
     numFiles=getNumCkptFiles(ckptDir) # len(os.listdir(ckptDir))
@@ -514,6 +543,12 @@ def runTestRaw(name, numProcs, cmds):
           "unexpected number of checkpoint files, %s procs, %d files"
           % (str(status[0]), numFiles))
 
+    if SLOW > 1:
+      #wait and see if some processes will die shortly after checkpointing
+      sleep(S*SLOW)
+      CHECK(doesStatusSatisfy(getStatus(), status),
+            "error: processes checkpointed, but died upon resume")
+
   def testRestart():
     #build restart command
     cmd=BIN+"dmtcp_restart --quiet"
@@ -521,9 +556,14 @@ def runTestRaw(name, numProcs, cmds):
       if i.endswith(".dmtcp"):
         cmd+= " "+ckptDir+"/"+i
     #run restart and test if it worked
-    procs.append(launch(cmd))
+    procs.append(runCmd(cmd))
     WAITFOR(lambda: doesStatusSatisfy(getStatus(), status),
             wfMsg("restart error"))
+    if SLOW > 1:
+      #wait and see if process will die shortly after restart
+      sleep(S*SLOW)
+      CHECK(doesStatusSatisfy(getStatus(), status),
+            "error:  processes restarted and then died")
     if HBICT_DELTACOMP == "no":
       clearCkptDir()
 
@@ -539,12 +579,17 @@ def runTestRaw(name, numProcs, cmds):
 
     #start user programs
     for cmd in cmds:
-      procs.append(launch(BIN+"dmtcp_launch "+cmd))
-      sleep(S*SLOW)
+      procs.append(runCmd(BIN+"dmtcp_launch "+cmd))
 
+    #TIMEOUT in WAITFOR has also been multiplied by SLOW
     WAITFOR(lambda: doesStatusSatisfy(getStatus(), status),
             wfMsg("user program startup error"))
 
+    # Additional sleep to allow the test to boot.
+    sleep(POST_LAUNCH_SLEEP)
+
+    #Will sleep(S*SLOW) in the following for loop.
+
     for i in range(CYCLES):
       if i!=0 and i%2==0:
         print #newline
@@ -552,6 +597,8 @@ def runTestRaw(name, numProcs, cmds):
       printFixed("ckpt:")
       # NOTE:  If this faile, it will throw an exception to CheckFailed
       #  of this function:  testRestart
+      #wait for launched processes to settle down, before we try to checkpoint
+      sleep(S*SLOW)
       testCheckpoint()
       printFixed("PASSED ")
       testKill()
@@ -589,7 +636,9 @@ def runTestRaw(name, numProcs, cmds):
                     printFixed(" (" + coredump + " copied to DMTCP_TMPDIR:" +
                                dmtcp_tmpdir() + "/)")
             else:
-              printFixed("(first process didn't die)")
+              printFixed("(Either first process didn't die, or sle this long" +
+                         " delay has been observed due to a slow" +
+                         " NFS-based filesystem.)")
             printFixed(" retry:")
             testKill()
       if i != CYCLES - 1:
@@ -680,6 +729,10 @@ runTest("dmtcp3",        1, ["./test/dmtcp3"])
 
 runTest("dmtcp4",        1, ["./test/dmtcp4"])
 
+runTest("alarm",        1, ["./test/alarm"])
+
+runTest("sched_test",    2, ["./test/sched_test"])
+
 # In 32-bit Ubuntu 9.10, the default small stacksize (8 MB) forces
 # legacy_va_layout, which places vdso in low memory.  This collides with text
 # in low memory (0x110000) in the statically linked mtcp_restart executable.
@@ -712,7 +765,19 @@ CKPT_CMD = old_ckpt_cmd
 # Test for files opened with WRONLY mode and later unlinked.
 runTest("file1",         1, ["./test/file1"])
 
-runTest("dmtcpaware1",   1, ["./test/dmtcpaware1"])
+# Test for files and their directories opened and unlinked
+# PREV. NOTE (now fixed?):
+#   Currently, we re-create deleted subdirectories when file
+#   is mmap'ed, but not yet when file is referenced by open fd.
+S=10*DEFAULT_S
+runTest("file2",         1, ["./test/file2"])
+S=DEFAULT_S
+
+# Test for normal file, /dev/tty, proc file, and illegal pathname
+runTest("stat",         1, ["./test/stat"])
+
+# Test for stack grow works on restart
+runTest("stack-growsdown",         1, ["./test/stack-growsdown"])
 
 PWD=os.getcwd()
 runTest("plugin-sleep2", 1, ["--with-plugin "+
@@ -729,50 +794,117 @@ runTest("plugin-example-db", 2, ["--with-plugin "+
                              "env EXAMPLE_DB_KEY=2 EXAMPLE_DB_KEY_OTHER=1 "+
                              "./test/dmtcp1"])
 
+runTest("plugin-init", 1, ["--with-plugin "+
+                             PWD+"/test/libdmtcp_plugin-init.so "+
+                             "./test/dmtcp1"])
+
 # Test special case:  gettimeofday can be handled within VDSO segment.
 runTest("gettimeofday",  1, ["./test/gettimeofday"])
 
 runTest("sigchild",      1, ["./test/sigchild"])
 
-runTest("shared-fd",     2, ["./test/shared-fd"])
+runTest("shared-fd1",     2, ["./test/shared-fd1"])
+
+runTest("shared-fd2",     2, ["./test/shared-fd2"])
 
 runTest("stale-fd",      2, ["./test/stale-fd"])
 
+runTest("rlimit-restore",         1, ["./test/rlimit-restore"])
+
+runTest("rlimit-nofile",      2, ["./test/rlimit-nofile"])
+
 # Disable procfd1 until we fix readlink
 #runTest("procfd1",       2, ["./test/procfd1"])
 
 # popen1 can have more than one processes
-#runTest("popen1",          1, ["./test/popen1"])
+runTest("popen1",          [1,2], ["./test/popen1"])
 
 runTest("poll",          1, ["./test/poll"])
 
+runTest("epoll1",        2, ["./test/epoll1"])
+
+if HAS_EPOLL_CREATE1 == "yes":
+  runTest("epoll2",        2, ["./test/epoll1 --use-epoll-create1"])
+
+runTest("environ",       1, ["./test/environ"])
+
 runTest("forkexec",      2, ["./test/forkexec"])
 
-if HAS_SSH == "yes":
-  S=5*DEFAULT_S
-  runTest("sshtest",     4, ["./test/sshtest"])
+runTest("realpath",      1, ["./test/realpath"])
+runTest("pthread1",      1, ["./test/pthread1"])
+runTest("pthread2",      1, ["./test/pthread2"])
+if HAS_MUTEX_WRAPPERS == "no":
+  # This is failing when --enable-mutex-wrappers is configured.
+  S=10*DEFAULT_S
+  runTest("pthread3",      1, ["./test/pthread2 80"])
   S=DEFAULT_S
+runTest("pthread4",      1, ["./test/pthread4"])
+runTest("pthread5",      1, ["./test/pthread5"])
 
-S=2*DEFAULT_S
-runTest("waitpid",      2, ["./test/waitpid"])
-S=DEFAULT_S
+if HAS_MUTEX_WRAPPERS == "yes":
+  runTest("mutex1",        1, ["./test/mutex1"])
+  runTest("mutex2",        1, ["./test/mutex2"])
+  runTest("mutex3",        1, ["./test/mutex3"])
+  # Comment out the test until pthread bug is fixed.
+  #runTest("mutex4",        1, ["./test/mutex4"])
+
+
+# FIXME:  pthread_atfork doesn't compile on some architectures.
+#         If we add a configure test for pthread_atfork, we can
+#           set a Python variable in autotest_config.py.in
+if uname_m != "armv7" and uname_m != "armv7l" and uname_m != "aarch64":
+  if os.getenv("LD_LIBRARY_PATH"):
+    os.environ["LD_LIBRARY_PATH"] += ":./test"
+  else:
+    os.environ["LD_LIBRARY_PATH"] = "./test"
+  runTest("pthread_atfork1",      2, ["./test/pthread_atfork1"])
+  runTest("pthread_atfork2",      2, ["./test/pthread_atfork2"])
+  if os.environ["LD_LIBRARY_PATH"] == "./test":
+    del os.environ["LD_LIBRARY_PATH"]
+  else:
+    os.environ["LD_LIBRARY_PATH"] = \
+      os.getenv("LD_LIBRARY_PATH")[:-len(":./test")]
+else:
+  print "Skipping pthread_atfork test; doesn't build on ARM/aarch64/glibc/Linux"
+
+if not USE_M32:  # ssh (a 64-bit child process) is forked
+  if HAS_SSH_LOCALHOST == "yes":
+    S=5*DEFAULT_S
+    runTest("ssh1",     4, ["./test/ssh1"])
+    S=DEFAULT_S
+
+if not USE_M32:  # waitpid forks a 64-bit child process, /bin/sleep
+  S=2*DEFAULT_S
+  runTest("waitpid",      2, ["./test/waitpid"])
+  S=DEFAULT_S
 
 runTest("client-server", 2, ["./test/client-server"])
 
 # frisbee creates three processes, each with 14 MB, if no gzip is used
 os.environ['DMTCP_GZIP'] = "1"
+POST_LAUNCH_SLEEP=2
 runTest("frisbee",       3, ["./test/frisbee "+p1+" localhost "+p2,
                              "./test/frisbee "+p2+" localhost "+p3,
                              "./test/frisbee "+p3+" localhost "+p1+" starter"])
-os.environ['DMTCP_GZIP'] = "0"
+POST_LAUNCH_SLEEP=DEFAULT_POST_LAUNCH_SLEEP
+os.environ['DMTCP_GZIP'] = GZIP
 
-runTest("shared-memory", 2, ["./test/shared-memory"])
+# On an NFS filesystem, a race can manifest late on the second restart,
+# due to a slow coordinator.
+S=10*DEFAULT_S
+runTest("shared-memory1", 2, ["./test/shared-memory1"])
+runTest("shared-memory2", 2, ["./test/shared-memory2"])
+S=DEFAULT_S
 
 runTest("sysv-shm1",     2, ["./test/sysv-shm1"])
 runTest("sysv-shm2",     2, ["./test/sysv-shm2"])
 runTest("sysv-sem",      2, ["./test/sysv-sem"])
 runTest("sysv-msg",      2, ["./test/sysv-msg"])
 
+# Makefile compiles cma only for Linux 3.2 and higher.
+if HAS_CMA == "yes":
+  runTest("cma",         2, ["./test/cma"])
+
 # ARM glibc 2.16 with Linux kernel 3.0 doesn't support mq_send, etc.
 if uname_p[0:3] == 'arm':
   print "Skipping posix-mq1/mq2 tests; ARM/glibc/Linux does not support mq_send"
@@ -781,11 +913,12 @@ elif TEST_POSIX_MQ == "yes":
   runTest("posix-mq2",     2, ["./test/posix-mq2"])
 
 #Invoke this test when we drain/restore data in pty at checkpoint time.
-# runTest("pty1",   2, ["./test/pty1"])
+runTest("pty1",   2, ["./test/pty1"])
 runTest("pty2",   2, ["./test/pty2"])
 
 #Invoke this test when support for timers is added to DMTCP.
-runTest("timer",   1, ["./test/timer"])
+runTest("timer1",   1, ["./test/timer1"])
+runTest("timer2",   1, ["./test/timer2"])
 runTest("clock",   1, ["./test/clock"])
 
 old_ld_library_path = os.getenv("LD_LIBRARY_PATH")
@@ -797,6 +930,7 @@ else:
 runTest("dlopen1",        1, ["./test/dlopen1"])
 # Disable the dlopen2 test until we can figure out a way to handle calls to
 # fork/exec/wait during library intialization with dlopen().
+# This seems to affect Travis CI of github, but not Ubuntu-12.04
 #if not USE_M32:
 #  runTest("dlopen2",        1, ["./test/dlopen2"])
 if old_ld_library_path:
@@ -804,13 +938,9 @@ if old_ld_library_path:
 else:
   del os.environ['LD_LIBRARY_PATH']
 
-runTest("pthread1",      1, ["./test/pthread1"])
-runTest("pthread2",      1, ["./test/pthread2"])
-S=10*DEFAULT_S
-runTest("pthread3",      1, ["./test/pthread2 80"])
-S=DEFAULT_S
-runTest("pthread4",      1, ["./test/pthread4"])
-runTest("pthread5",      1, ["./test/pthread5"])
+# Most of the remaining tests are on 64-bit processes.
+if USE_M32:
+  sys.exit()
 
 os.environ['DMTCP_GZIP'] = "1"
 runTest("gzip",          1, ["./test/dmtcp1"])
@@ -824,7 +954,7 @@ runTest("perl",          1, ["/usr/bin/perl"])
 if HAS_PYTHON == "yes":
   runTest("python",      1, ["/usr/bin/python"])
 
-os.environ['DMTCP_GZIP'] = "0"
+os.environ['DMTCP_GZIP'] = "1"
 runTest("bash",        2, ["/bin/bash --norc -c 'ls; sleep 30; ls'"])
 os.environ['DMTCP_GZIP'] = GZIP
 
@@ -835,7 +965,7 @@ if HAS_DASH == "yes":
   os.environ['DMTCP_GZIP'] = GZIP
 
 if HAS_TCSH == "yes":
-  os.environ['DMTCP_GZIP'] = "0"
+  os.environ['DMTCP_GZIP'] = "1"
   runTest("tcsh",        2, ["/bin/tcsh -f -c 'ls; sleep 30; ls'"])
   os.environ['DMTCP_GZIP'] = GZIP
 
@@ -848,10 +978,10 @@ if HAS_ZSH == "yes":
 
 if HAS_VIM == "yes":
   # Wait to checkpoint until vim finishes reading its initialization files
-  S=10*DEFAULT_S
+  S=5*DEFAULT_S
   if sys.version_info[0:2] >= (2,6):
     # Delete previous vim processes.  Vim behaves poorly with stale processes.
-    vimCommand = VIM + " /etc/passwd +3" # +3 makes cmd line unique
+    vimCommand = VIM + " -X /etc/passwd +3" # +3 makes cmd line unique
     def killCommand(cmdToKill):
       if os.getenv('USER') == None or HAS_PS == 'no':
         return
@@ -863,15 +993,17 @@ if HAS_VIM == "yes":
         if len(cmd) > 1 and cmd[1] == cmdToKill:
           os.kill(int(cmd[0]), signal.SIGKILL)
     killCommand(vimCommand)
-    runTest("vim",       1,  ["env TERM=vt100 " + vimCommand])
+    ### Commented out due to CentOS 7.5 failure.
+    ### See comment in src/dmtcpworker.cpp:dmtcp_initialize()
+    ###  runTest("vim",       1,  ["env TERM=vt100 " + vimCommand])
     killCommand(vimCommand)
   S=DEFAULT_S
 
 if sys.version_info[0:2] >= (2,6):
-  #On some systems, "emacs -nw" launches dbus-daemon processes in
+  #On some systems, "emacs -nw" runs dbus-daemon processes in
   #background throwing off the number of processes in the computation. The
-  #test thus fails. The fix is to launch emacs-nox, if found. emacs-nox
-  #doesn't launch any background processes.
+  #test thus fails. The fix is to run emacs-nox, if found. emacs-nox
+  #doesn't run any background processes.
   S=15*DEFAULT_S
   if HAS_EMACS_NOX == "yes":
     # Wait to checkpoint until emacs finishes reading its initialization files
@@ -885,8 +1017,11 @@ if sys.version_info[0:2] >= (2,6):
     # Under emacs23, it opens /dev/tty directly in a new fd.
     # To avoid this, consider using emacs --batch -l EMACS-LISTP-CODE ...
     # ... or else a better pty wrapper to capture emacs output to /dev/tty.
-    runTest("emacs",     1,  ["env TERM=vt100 /usr/bin/emacs -nw" +
-                                " --no-init-file /etc/passwd"])
+    ### Commented out due to CentOS 7.5 failure.
+    ### See comment in src/dmtcpworker.cpp:dmtcp_initialize()
+    ###  runTest("emacs",     1,  ["env TERM=vt100 /usr/bin/emacs -nw" +
+    ###                            " --no-init-file /etc/passwd"])
+    pass
   S=DEFAULT_S
 
 if HAS_SCRIPT == "yes":
@@ -898,7 +1033,9 @@ if HAS_SCRIPT == "yes":
     #  /usr/lib/locale/locale-archive is 100 MB, and yet 'locale -a |wc' shows
     #  only 8KB of content in ASCII.  The 100 MB of locale-archive condenses
     #  to 25 MB _per process_ under gzip, but this can be slow at ckpt time.
-    runTest("script",    4,  ["/usr/bin/script -f" +
+    # On some systems, the script test has two `script` processes, while on some
+    # other systems, there is only a single `script` process.
+    runTest("script",    [3,4],  ["/usr/bin/script -f" +
                               " -c 'bash -c \"ls; sleep 30\"'" +
                               " dmtcp-test-typescript.tmp"])
   os.system("rm -f dmtcp-test-typescript.tmp")
@@ -907,7 +1044,9 @@ if HAS_SCRIPT == "yes":
 # SHOULD HAVE screen RUN SOMETHING LIKE:  bash -c ./test/dmtcp1
 # FIXME: Currently fails on dekaksi due to DMTCP not honoring
 #        "Async-signal-safe functions" in signal handlers (see man 7 signal)
-if HAS_SCREEN == "yes":
+# **** Pty SUPPORT IN 2.5 BRANCH IS FLAKY.
+# **** screen NEEDS PTY; DON'T RUN SCREEN TEST UNTIL PTY FIXED.
+if HAS_SCREEN == "yes" and False:
   S=3*DEFAULT_S
   if sys.version_info[0:2] >= (2,6):
     runTest("screen",    3,  ["env TERM=vt100 " + SCREEN +
@@ -1028,6 +1167,9 @@ if HAS_OPENMPI == "yes":
   if oldPath == None:
     del os.environ['PATH']
 
+# Test DMTCP utilities:
+runTest("nocheckpoint",        1, ["./test/nocheckpoint"])
+
 print "== Summary =="
 print "%s: %d of %d tests passed" % (socket.gethostname(), stats[0], stats[1])
 
